//
// main.C
//
// $Header: /fs/nene/uysal/mambo/cholesky/fact/RCS/main.C,v 1.2 1997/03/01 03:56:29 uysal Exp uysal $
//
// Parallel Outofcore Sparse Cholesky Factorization
//
// Input  : 
//   Partitioned sparse matrix generated by 'part'
//
// Output : 
//   Cholesky Factor of the sparse matrix, rewriting the old space
//
// Data Formats :
//
//    Input Matrix :    
//    1  int   Number of Supernodes (value: N)
//    N  int   Data size (in bytes) for each supernode in local file
//    N  long  Data offset (in bytes) for each supernode in local file
//   N+1 int   Column Index for each supernode
//    N  snod  Data for each supernode (see below)    
//  
//    Supernode Format :
//    1  int   Count of Updaters (value: U)
//    U  int   Index of supernode to update current
//    1  int   No. of rows stored locally (value: R)
//    R  int   Row Indices, one for each row
//    -  elem  Matrix Data (if any) stored locally in row major format  
//
// Author : Mustafa Uysal <uysal@cs.umd.edu>
//
//
//
#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <math.h>

#include "sp.h"
#include "vector.h"
#include "buffer.h"
#include "datatype.h"
#include "mapping.h"
#include "fbuf.h"

#define SHIFT1  100
#define SHIFT2  200
#define SHIFT3  300
#define SHIFT4  400

// Function Prototypes
void Usage(char *prog);
char *SupernodeRead(int ofid, int sno, int size, long offset,
		    int &nUpdate, int &nRows, int &ndxOffset, int &dataOffset);
char *UpdateRead(int ofid, int sno, int size, long offset, int rowcnt, int gapsize, int &nRows);
void SupernodeWrite(int ofid, int sno, int nRows, elem_t *data, long offset, int size);

void BlockUpdate(int mRows, int *mIndex, elem_t *mData, int uWidth,
		 int uRows, int *uIndex, elem_t *uData, int dgIndex,
		 int dRows, int *dIndex, elem_t *dData, int dWidth);
void DiagFactor(int nRows, int *index, elem_t *data);
void PanelFactor(int nRows, elem_t *data, int dWidth, elem_t *diag);

// Global Variables 
ibmsp  mpp;          // will run for SP-2
double totRead = 0;
double totWrite = 0;
fbuf   *fdata = NULL;

#if defined(TRACE)
extern "C" trace_init(char *name);
extern "C" trace_start(void);
extern "C" trace_end(void);
#endif

elem_t PONE = 1.0;

int main(int argc, char *argv[])
{
    // argument control 
    if (argc<2) Usage(argv[0]);

    // argument extraction
    int  argndx = 1;
    char *matfile = argv[argndx++];
#if defined(TRACE)
    if (argndx < argc) trace_init(argv[argndx]);
    else Usage(argv[0]);
#endif

    // Output Logging
    mpp.init();
    mpp.printOnZero("\nParallel OutOfCore Sparse Cholesky Factorizaion");
    mpp.printOnZero("\tMatrix File          : ", matfile);

    // Trace collection starts here
#if defined(TRACE)  
    mpp.printOnZero("\tTrace File           : ", argv[argndx]);
    trace_start();
#endif

    int ofid = open(matfile, O_RDWR);
    if (ofid<0)
      {
	cerr << "Error in opening file: " << matfile << "\n";
	exit(1);
      }

    int numsnode;
    int bread = read(ofid, &numsnode, sizeof(int));

    int *snodeSize = new int[numsnode];
    bread = read(ofid, snodeSize, sizeof(int)*numsnode);

    long *snodeOffs = new long[numsnode];
    bread = read(ofid, snodeOffs, sizeof(long)*numsnode);

    int *snodeIndex = new int[numsnode+1];
    bread = read(ofid, snodeIndex, sizeof(int)*(numsnode+1));

    totRead = (numsnode+1)*2*sizeof(int) + numsnode * sizeof(long);
    mpp.printOnZero("File Header processed.");

    // create and initialize file buffer
    fbuf mdata(numsnode, snodeOffs, snodeSize, ofid);
    fdata = &mdata;

    // Following object manages of mapping supernodes onto processors
    TorusMap smap((mpapi&)mpp);
    if (mpp.myProc()==0) smap.integrityCheck();

    // Following structures contain the gap between index and data (ie. unused data)
    int *gapSize  = new int[numsnode];

    double startTime = mpp.clock();

    // Main Factorization Loop:
    for (int k = 0; k < numsnode; k++)
      {
	// progress indicator...
	if (/*(((k%100)==0) || (k==(numsnode-1))) &&*/ mpp.myProc()==0)
	  {
	    cout << ">processing : " << k << ", ";
	    cout << snodeIndex[k+1] - snodeIndex[k] << " columns of ";
	    cout << snodeIndex[k] << ", ";
	    cout << totRead << " rd, ";
	    cout << totWrite << " wr, ";
	    cout << mpp.clock() - startTime << " secs.\n";
	    cout.flush();
	  }

	// supernode read...
	int   nUpdate, nRows, ndxOffset, dataOffset;
	char *buffer = SupernodeRead(ofid, k, snodeSize[k], snodeOffs[k],
				     nUpdate, nRows, ndxOffset, dataOffset);

	// make sure dependencies are not read again
	snodeSize[k] -= ((nUpdate+2)*sizeof(int));
	snodeOffs[k] += ((nUpdate+2)*sizeof(int));

	// setup working space...
	elem_t *data  = NULL, *diag = NULL;  // data of the panel(k)
	int    *index = NULL;  // index of the panel(k)
	int    dWidth = snodeIndex[k+1] - snodeIndex[k];
	int    *uNdx  = (int*)(buffer + sizeof(int));
	int     local = smap.isColMember(k);

	if (ndxOffset>0)
	  {
	    index  = (int*)(buffer+ndxOffset);
	    data   = (elem_t*)(buffer+dataOffset);
	  }

	// disclose update info
	fdata->fbaccess(nUpdate, uNdx);
	
	// 
	// This is the main update loop:
	//
	for (int j = 0; j < nUpdate; j++)
	  {
	    char *uBuf = NULL, *mBuf = NULL;
	    int   uRows = 0, mRows = 0, width = 0;
	    int   tid  = uNdx[j], jj;
	    int   info[2] = {0,0};
	    int   haveUpdate = smap.isColMember(tid);

	    int    *uIndex = NULL, *mIndex = NULL;
	    elem_t *uData  = NULL, *mData  = NULL;

	    // if update data exists - read into memory:
	    if (haveUpdate)
	      if (snodeIndex[tid]>0)
		{
		  uBuf     = UpdateRead(ofid, tid, snodeSize[tid], snodeOffs[tid],
					snodeIndex[tid], gapSize[tid], uRows);
		  uIndex   = (int*) (uBuf+sizeof(int));
		  uData    = (elem_t*) (uBuf + (uRows+1)*sizeof(int));
		  info[0]  = uRows;
		  info[1]  = snodeSize[tid]-gapSize[tid];
		  width    = ((info[1] - uRows*sizeof(int))/sizeof(elem_t))/uRows;
		  
		  // send all 'base' data to my companion processor
		  int dest = smap.rowPeer(tid, k);
		  if (!local)
		    {
		      mpp.bsend(dest, tid+SHIFT1, 2*sizeof(int), info);
		      mpp.bsend(dest, tid+SHIFT2, info[1], uIndex);
		    }
		  
		  // if hosting the multiplier: multicast {panel(k)_owners - dest}
		  if (dest == smap.map(k,k))
		    {
		      int multip, kk;

		      // compute multiplier size
		      for (multip = 0; multip < uRows; multip++)
			if (uIndex[multip] >= snodeIndex[k+1]) break;
		      
		      // compute message size
		      int msize = (multip+1)*sizeof(int) + multip*width*sizeof(elem_t);
		      mBuf   = new char[msize];
		      mIndex = (int*) (mBuf);
		      mData  = (elem_t*)(mBuf+(multip+1)*sizeof(int));
		      mRows  = multip;

		      // bring index and data back to back
		      mIndex[0] = multip;
		      for (kk = 0; kk < multip; kk++)
			mIndex[kk+1] = uIndex[kk];
		      int lbound = width * multip;
		      for (kk = 0; kk < lbound; kk++)
			mData[kk] = uData[kk];

		      // send multiplier chunk
		      Vector *destSet = smap.procCol(k);
		      destSet->delElem(dest);
		      mpp.multicast(destSet, tid+SHIFT3, msize, mBuf);
		      delete destSet;
		      
		      // owner of multiplier adjusts the reading parameters:
		      snodeSize[tid]  -= (multip * sizeof(int));
		      snodeOffs[tid]  += (multip * sizeof(int));
		      gapSize[tid]    += (multip * width * sizeof(elem_t));
		      snodeIndex[tid] -= multip;
		      if (snodeIndex[tid]<1)
			snodeSize[tid] = width * sizeof(elem_t);
		    }
		}
	      else
		{
		  info[1]  = snodeSize[tid];
		  int dest = smap.rowPeer(tid, k);
		  if (dest != mpp.myProc())
		    {
		      mpp.bsend(dest, tid+SHIFT1, 2*sizeof(int), info);
		      continue;
		    }
		}

	    if (local)  // have target, receive source
	      {
		int msrc, mtag, msiz;

		if (!haveUpdate)
		  {
		    msrc = smap.rowPeer(k, tid);
		    mtag = tid+SHIFT1;
		    msiz = 2*sizeof(int);
		    
		    // base info index
		    mpp.brecv(mtag, msrc, msiz, info);
		    if (info[0])
		      {
			// get base data
			uBuf = new char[info[1]];
			mtag = tid + SHIFT2;
			msiz = info[1];
			mpp.brecv(mtag, msrc, msiz, uBuf);
			
			// parse base data
			uRows  = info[0];
			uIndex = (int*) uBuf;
			uData  = (elem_t*)(uBuf + info[0]*sizeof(int));
			width  = ((msiz-uRows*sizeof(int))/sizeof(elem_t))/uRows;
		      }
		  }

		// get or compute multipliers
		if (mpp.myProc() == smap.map(k,k))  // should have multiplier
		  {
		    mIndex = uIndex;
		    mData  = uData;
		    for (mRows = 0; mRows < uRows; mRows++)
		      if (uIndex[mRows] >= snodeIndex[k+1]) break;
		  }
		else                            // need to recv multiplier
		  {
		    int mSize;
		    
		    if (info[0])
		      {
			int uWidth = (info[1] - info[0]*sizeof(int)) / info[0];
			mSize  = (snodeIndex[k+1] - snodeIndex[k])*uWidth*sizeof(elem_t);
		      }
		    else
		      {
			mSize  = info[1] * (snodeIndex[k+1] - snodeIndex[k]);
		      }

		    mSize += ((snodeIndex[k+1] - snodeIndex[k] + 1)*sizeof(int));
		    mtag = tid+SHIFT3;
		    msrc = smap.map(tid,k);
		    
		    mBuf = new char[mSize];
		    mpp.brecv(mtag, msrc, mSize, mBuf);
		    
		    mRows  = *((int*)mBuf);
		    mIndex = (int*)(mBuf+sizeof(int));
		    mData  = (elem_t*)(mBuf+(mRows+1)*sizeof(int));
		  }

		// update processing:
		BlockUpdate(mRows, mIndex, mData, width,
			    uRows, uIndex, uData, snodeIndex[k],
			    nRows, index, data, dWidth);
	      }
	    
	    // housekeeping
	    delete[] uBuf;
	    if (mBuf) delete[] mBuf;
	  }
	// End of Main Update Loop.

	// factorize panel header:
	int dsize = dWidth * (dWidth + 1) / 2;
	if (index[0] == snodeIndex[k])
	  {
	    DiagFactor(dWidth, index, data);
	    Vector *destSet = smap.procCol(k);
	    mpp.multicast(destSet, k+SHIFT4, dsize * sizeof(elem_t), data);
	    delete destSet;
	  }
	else if (local)
	  {
	    int mtag = k + SHIFT4;
	    int msrc = smap.map(k,k);
	    int msiz = dsize * sizeof(elem_t);
	    diag = new elem_t[dsize];
	    mpp.brecv(mtag, msrc, msiz, diag);
	  }

	// factorize rest of the supernode:
	if (index[0] == snodeIndex[k])
	  PanelFactor(nRows - dWidth, data+dsize, dWidth, data);
	else if (local)
	  PanelFactor(nRows, data, dWidth, diag);

	// next time we read this panel, skip dependencies and read data only
	snodeIndex[k] = nRows;
	gapSize[k]    = 0;

	// Adjust reading parameters so that diagonal is not read again
	if (index)
	  {
	    SupernodeWrite(ofid, k, nRows, data, snodeOffs[k], snodeSize[k]);

	    int kk,ne = 1;
	    for (kk = 0; (kk<nRows) && (index[kk]<snodeIndex[k+1]); kk++)
	      {
		snodeIndex[k]--;
		snodeSize[k] -= sizeof(int);
		snodeOffs[k] += sizeof(int);
		gapSize[k]   += (ne * sizeof(elem_t));
		ne++;
	      }
	  }

	fdata->funlock(k);
	if (diag) delete[] diag;
      }

#if defined(TRACE)
    trace_end();
#endif
    return 0;
}



// Early exit from the program with a display of program usage
void Usage(char *progname)
{
    cerr << "\n\nUsage : " << progname << " MatrixFile traceFile\n\n";
    exit(1);
}




// Routine to Read a Supernode and decompose it into fields (first time)
char *SupernodeRead(int ofid, int sno, int size, long offset,
		    int &nUpdate, int &nRows, int &ndxOffset, int &dataOffset)
{
  char *buffer;
 
  buffer  = fdata->fbget(sno);
  nUpdate = *((int*)buffer);
  nRows   = *((int*)(buffer+(nUpdate+1)*sizeof(int)));
  fdata->flock(sno); // lock it 

  if (nRows>0)
    {
      ndxOffset  = (nUpdate+2)*sizeof(int);
      dataOffset = (nUpdate+2+nRows)*sizeof(int);
    }
  else
    {
      ndxOffset  = -1;
      dataOffset = -1;
    }

  //cout << "snode read(" << sno << ")  - nu: " << nUpdate << " nr: " << nRows << " [";
  //cout << ndxOffset << dataOffset << "] index(0) = " << *((int*)(buffer+ndxOffset)) << "\n";
  //cout.flush();

  return buffer;
}




// Read a Supernode as updater (skip unused parts)
char *UpdateRead(int ofid, int sno, int size, long offset, int rowcnt, int gapsize, int &nRows)
{
  assert( rowcnt > 0);
  int ndxsize  = rowcnt * sizeof(int);
  int datsize  = size - (ndxsize + gapsize);
  char *buffer = new char[ndxsize+datsize+sizeof(int)];
  char *ptr;

  // read and copy index 
  ptr = fdata->fbread(sno, offset, ndxsize);
  for (int k = 0; k < ndxsize; k++)
    buffer[sizeof(int)+k] = ptr[k];
  
  // read and copy data 
  ptr = fdata->fbread(sno, offset+ndxsize+gapsize, datsize);
  for (k = 0; k < datsize; k++)
    buffer[ndxsize+sizeof(int)+k] = ptr[k];

  nRows    = rowcnt;

  return buffer;
}



void SupernodeWrite(int ofid, int sno, int nRows, elem_t *data, long offset, int size)
{
  int indexSkip = nRows * sizeof(int);
  offset += (long) (indexSkip);
  size   -= indexSkip;

  fdata->fbwrite((char*)data, offset, size);
  totWrite += (long) size;
}




//
// multiplier + base -> block update operation
//
void BlockUpdate(int mRows, int *mIndex, elem_t *mData, int uWidth,
		 int uRows, int *uIndex, elem_t *uData, int dgIndex,
		 int dRows, int *dIndex, elem_t *dData, int dWidth)
{
  elem_t *dPtr, *mPtr, *uPtr;
  int m,u,i,k = 0, kBase = 0;

  //cout << " m:" << mRows << " u:" << uRows << " d:" << dRows;
  //cout << " uw:" << uWidth << " dw:" << dWidth << " di:" << dgIndex << "\n";

  // special case: panel headers
  if (dIndex[0] == dgIndex)
    {
      for (m = 0, mPtr = mData; m < mRows; m++, mPtr += uWidth)
	{
	  int dcol = mIndex[m] - dgIndex;
	  dPtr = dData + dcol*(dcol+1)/2+dcol;
	  for (u = 0, uPtr = mData; u <= m; u++, uPtr += uWidth)
	    {
	      for (i = 0; i < uWidth; i++)
		*dPtr -= (uPtr[i] * mPtr[i]);
	      dPtr++;
	    }
	}
      dData += (dWidth * (dWidth+1) / 2);
      kBase  = dWidth;
    }

  // setup for the panel body
  for (u = 0; u < uRows; u++)
    if (uIndex[u]>=(dgIndex+dWidth)) break;

  // general case: panel body
  for ( ; u < uRows; u++)
    {
      while (dIndex[k] != uIndex[u]) k++;
      uPtr = uData + u * uWidth;
      dPtr = dData + (k - kBase) * dWidth;
      assert( k >= kBase );

      for (m = 0, mPtr = mData; m < mRows; m++, mPtr += uWidth)
	{
	  int dCol = mIndex[m] - dgIndex;
	  for (i = 0; i < uWidth; i++)
	    dPtr[dCol] -= (uPtr[i] * mPtr[i]);
	}
    }
}



void DiagFactor(int nRows, int *index, elem_t *data)
{
  int   mindex;    /* Index of the first element in the multipler row */
  int   mincr;     /* Increment to find the next multiplier row */
  int   dindex;    /* Index of the current diagonal element (start of column) */
  int   cindex;
  int   base;
  int   uincr;
  int   j, mptr, bptr;
  int   size;

  size = nRows * (nRows + 1) / 2;
  for (mindex=mincr=0; mindex<size; mindex+=(++mincr))
    {
      /* Update L(j,j) */
      dindex = mindex + mincr;
      for (j=mindex; j<dindex; j++)
	data[dindex] -= data[j] * data[j];
      
#ifdef PDEF_CHECK
      if (data[dindex]<=ZERO) data[dindex] = PONE;
#endif

      /* Compute L(j,j) */
      data[dindex] = sqrt(data[dindex]);
      elem_t mult = PONE / data[dindex]; /* Equivalent to 1/diagonal */

      /* Update elements that are in the same column as dindex */
      uincr = mincr+1;
      for (base=mindex+uincr, cindex=dindex+uincr; base<size; base+=uincr, cindex+=uincr) 
	{
	  for (mptr=mindex, bptr=base; bptr<cindex; bptr++, mptr++) 
	    data[cindex] -= data[mptr] * data[bptr];
	  ++uincr;
	}
      
      /* Scale subdiagonal elements i+1:size */
      for (uincr=mincr+1, dindex+=(mincr+1); dindex<size; dindex+=(++uincr))
	data[dindex] *= mult;
    }
}




void PanelFactor(int nRows, elem_t *data, int dWidth, elem_t *diag)
{
  int   dupdndx;     /* Index for the updated destination index */
  int   dbasndx;     /* Index of the destination base elemen */
  int   i,j;         /* loop variables */

  if (nRows<=0) return;

  /* Initialization */
  elem_t *mulptr = diag;    /* Pointer to the multipler element */
  int     size   = nRows * dWidth;

  /* Allegedly efficient Compute Kernel */
  for (i=0; i<dWidth; i++)
    {
      /* Update the new column */
      for (j=0; j<i; j++)
	{
	  for (dupdndx=i,dbasndx=j; dupdndx<size; dbasndx+=dWidth,dupdndx+=dWidth)
	    data[dupdndx] -= (*mulptr) * data[dbasndx];
	  mulptr++;
	}
      
      /* Scale the subdiagonal elements */
      elem_t multip = PONE / *mulptr; /* Multiplier for the column scale in dest block */
      mulptr++;
      for (j=i; j<size; j+=dWidth)
	data[j] *= multip;
    }
}
