/*
 * comm.c
 *
 * $Source: /fs/nene/uysal/npFact/RCS/comm.c,v $
 *
 * $Header: /fs/nene/uysal/npFact/RCS/comm.c,v 1.3 1995/12/27 02:00:09 uysal Exp $
 *
 * Interprocessor communication module
 *
 * Author : Mustafa Uysal (uysal@cs.umd.edu)
 *
 */
#include <math.h>
#include <memory.h>

#include "world.h"
#include "mppworld.h"
#include "queue.h"
#include "block.h"
#include "comm.h"
#include "snode.h"
#include "mem.h"

#include "debug.h"

#define MSG_DATA_REQ  1


/*-------------- Global Variable Declarations ----------------*/

static int **pgrid;    /* Processor grid */
static int gridx;      /* x dimension size */
static int gridy;      /* y dimension size */
static int *slotMask;  /* Mask array for existence in a slot */

queue_t *sq;
queue_t *rq;

typedef struct {
  int msgid;
  int msglen;
  char *buffer;
} send_t;


int *rcv_buf;
int  rcv_id, rcv_tag, rcv_src;


/*--------------- Local Function Prototypes ---------------*/

/* Routine to compute the processor grid for communication */
void FormProcessorGrid ARGS((int np, int me));

void PostRequest(void);

/*-------------- Function Implementations ---------------------*/


/* Initialization of communication module */
int *CommInit(nslot)
    int *nslot;
{
  int i,j;

  /* initialize interprocessor communication grid */
  FormProcessorGrid(NumProcs(), MyProc());
  *nslot = gridx;

  sq = QueueNew();
  rq = QueueNew();

  slotMask = (int*) Calloc(gridx, sizeof(int));
  for (i=0; i<gridx; i++)
    {
      slotMask[i] = 0;
      for (j=0; j<gridy; j++)
	if (pgrid[j][i] == MyProc()) slotMask[i] = 1;
    }

  PostRequest();

  return slotMask;
}




/* Routine to find the mapping of the source supernode */
void SourceMapping(pid, sid, bmp, mmp, me)
    int pid, sid, *bmp, *mmp, me;
{
  int pmap, smap, i;

  pmap = pid % gridx;
  smap = sid % gridx;

  for (i=0; i<gridy; i++)
    if (pgrid[i][pmap] == me)
      {
	*bmp = pgrid[i][smap];
	*mmp = pgrid[pid%gridy][smap];
	return;
      }

  panic(("\nSourceMappin() failed."));
}



/* Routine to find the processor to which the diagonal of a pivot is mapped */
int DiagMapping(pid)
    int pid;
{
  return(pgrid[pid%gridy][pid%gridx]);
}



/* Cleanup function for the communications module */
void CommExit()
{
  QueueDestroy(sq);
  QueueDestroy(rq);
}


/* Routine to compute the processor grid for communication */
void FormProcessorGrid(np, me)
    int np;     /* number of processors */
    int me;     /* my processor id */
{
  double tmp;      /* to get the result of sqrt */
  int *gptr;       /* pointer to a row of pgrid */
  int i,j,k,cnt;   /* control vars */

  /* find out grid parameters */
  tmp = sqrt((double)np);
  gridy = (int) tmp;
  gridx = np/gridy;
  if ((gridy*gridx)<np)
    gridx++;

  /* storage allocation */
  pgrid = (int**) Calloc(gridy, sizeof(int*));

  /* fill-up the processor grid */
  for (cnt=i=0; i<gridy; i++)
    {
      gptr = pgrid[i] = (int*) Calloc(gridx, sizeof(int));
      for (j=0; j<gridx; j++, cnt++)
	gptr[j] = cnt % np;
    }

  if (!MyProc())
    {
      printf("\n-----------------------------------------------");
      printf("\n Grid - np=%d me=%d x=%d, y=%d", np, me, gridx, gridy);
      printf("\n-----------------------------------------------");
      
      for (i=0; i<gridy; i++)
	{
	  printf("\n %d : ", i);
	  for (j=0; j<gridx; j++)
	    printf("%3d ", pgrid[i][j]);
	}
    }
}



void MsgSendRec(char *buf, int size, int dest, int type)
{
  send_t *ptr;

  ptr = (send_t*) Malloc(sizeof(send_t));

  ptr->buffer = buf;
  ptr->msglen = size;
  mpc_send(buf, size, dest, type, &(ptr->msgid));

  QueueAdd(sq, ptr);
}




int FreeMsgQueue()
{
  send_t *ptr;

  while (!QueueEmpty(sq))
    {
      ptr = (send_t*) QueueHead(sq);
      if (mpc_status(ptr->msgid)<0)
	break;
      ptr = (send_t*) QueueDelete(sq);
      Free(ptr->buffer);
      Free(ptr);
    }

  return sq->load;
}


int NumRequests()
{
  return rq->load;
}



int *RecvRequest(void)
{
  while (mpc_status(rcv_id)>0)
    {
      QueueAdd(rq, rcv_buf);
      PostRequest();
    }

  if (QueueEmpty(rq)) return (int*) NULL;
  return (int*) QueueDelete(rq);
}





/* Rutine Post Server request message slot */
void PostRequest()
{
  rcv_buf = (int*) Calloc(6, sizeof(int));
  rcv_tag = MSG_DATA_REQ;
  rcv_src = DONTCARE;
  if (mpc_recv(rcv_buf, 6*sizeof(int), &rcv_src, &rcv_tag, &rcv_id))
    {
      panic(("PostRequest() failed"));
    }
}



void QueueRequest(int *buf)
{
  QueueAdd(rq, buf);
}




void PrintRequests()
{
  int *ptr;
  int i;

  for (i=0; !QueueEmpty(rq); i++)
    {
      ptr = (int*) QueueDelete(rq);
      printf("\nReq[%d]: %d %d %d %d %d %d", i, ptr[0], ptr[1], ptr[2], ptr[3],ptr[4],ptr[5]);
    }
}
