/*
 * disk.c
 *
 * $Source: /fs/nene/uysal/npFact/RCS/disk.c,v $
 *
 * $Header: /fs/nene/uysal/npFact/RCS/disk.c,v 1.2 1995/12/27 02:00:09 uysal Exp $
 *
 * Module to coordinate the accesses to the disk
 *
 * Applicable Compiler Directives :
 * DUMP_METADATA   The metadata of the matrix is dumped in a human
 *                 readble form to stdout
 *
 *
 * Author : Mustafa Uysal (uysal@cs.umd.edu)
 *
 */
#include <aio.h>
#include <fcntl.h>
#include <unistd.h>

#include "world.h"
#include "block.h"
#include "disk.h"
#include "snode.h"
#include "mppworld.h"
#include "queue.h"
#include "io.h"

#include "debug.h"


/*--------------- Static Global Variables ---------------*/

static int dataFile;    /* dataFile handler */


/*--------------- Externally Defined Globals ---------------*/

extern int errno;

extern double bytes_read;
extern double read_time;

extern double bytes_written;
extern double write_time;


/*--------------- Local Function Prototypes ---------------*/

/* routine to setup local Value Pointers to the correct location */
void SetUpValuePtrs ARGS((snode_t *sptr));

/* routine to attach to strings together */
char *stratt ARGS((char *s1, char *s2));


/*--------------- Module Implementation ---------------*/

/*
 * Routine to read the configuration file
 */
void DiskInit(fn, ns, ofptr, szptr)
    char *fn;
    int  *ns;
    int  **ofptr, **szptr;
{
  int confFile;
  char *str;
  int *ofset, *size, cnt;
  
  str = stratt(fn, ".blk");
  dataFile = open(str, O_RDWR);
  Free(str);

  str = stratt(fn, ".atr");
  confFile = open(str, O_RDONLY);
  Free(str);

  exit_if((dataFile<0)||(confFile<0),
	  ("Error in file opens [%d %d] [%s]", dataFile, confFile, fn));

  if (read(confFile, &cnt, sizeof(int)) != sizeof(int))
    panic(("Error in reading configuration file [1]."));

  ofset = (int*) Calloc(cnt, sizeof(int));
  size  = (int*) Calloc(cnt, sizeof(int));
  
  if (read(confFile, ofset, cnt*sizeof(int)) != cnt*sizeof(int))
    panic(("Error in reading configuration file [2]."));

  if (read(confFile, size, cnt*sizeof(int)) != cnt*sizeof(int))
    panic(("Error in reading configuration file [3]."));
  close(confFile);

  *ns = cnt;
  *ofptr = ofset;
  *szptr = size;

  print_if(!MyProc(), ("\n%d Supernodes", cnt));

#if defined(DUMP_METADATA)
  print_if(!MyProc(), ("\nMetadata will be printed"));
#else
  print_if(!MyProc(), ("\nMetadata will NOT be printed"));
#endif
}




/* Housekeeping and cleanup routine for disk module */
void DiskExit()
{
}




/* Blocking read of local supernode data */
snode_t *DiskReadLocal(sid)
    int sid;
{
  block_t *bptr;
  snode_t *sptr;
  int size, offset, j, cnt;
  int mode;

  /* Supernode request, if it exists in the memory */
  sptr = SnodeRequest(sid);
  if (sptr) return sptr;

  /* Find out where this supernode is */
  offset = SnodeOffset(sid);
  size   = SnodeSize(sid);

  if (size < 0) { mode = DSK_FIRST; size = abs(size); }
  else mode = DSK_NEXT;
  
  /* Space allocation and initialization */
  sptr = SnodeAlloc(sid, offset, size);
  
  /* blocking read */
  TimedSeek(dataFile, offset, SEEK_SET, &read_time);
  TimedRead(dataFile, sptr->buffer, size, &read_time, &bytes_read); 
  
  /* setup block descriptors so that the value pointers are right */
  SetUpValuePtrs(sptr, mode);
  
  /*if (mode == DSK_FIRST) SnodeStructPrint(sptr);*/

  return sptr;
}




/* Blocking write of local supernode data */
void DiskWriteLocal(ptr)
    snode_t *ptr;
{
  int *metaptr;
  int size, offset, depcnt;

  /* dooze off if its not my business */
  if (!ptr->blkcnt) return;
  
  /* Find out where this supernode is */
  size   = SnodeSize(ptr->sid);
  assert(size<0);

  metaptr = (int*)ptr->buffer;
  depcnt = metaptr[0];
  SnodeLocatorAdjust(ptr->sid, (depcnt+1)*sizeof(int));
  offset = SnodeOffset(ptr->sid);
  size   = SnodeSize(ptr->sid);

  /* blocking write */
  TimedSeek(dataFile, offset, SEEK_SET, &write_time);
  TimedWrite(dataFile, ptr->blkPtr[0].value, size, &write_time, &bytes_written);
}





/* Routine to fix [*value] fields of block descriptors, after reading the
 * supernode data from disk */
void SetUpValuePtrs(sptr, mode)
    snode_t *sptr;
    int mode;
{
  caddr_t address;
  block_t *blkptr;
  int blkcnt, depcnt;
  int i, id, size;

  /* get # of blocks */
  address = sptr->buffer + sptr->size - sizeof(int);
  blkcnt = *((int*)address);

  /* get the address of block descriptors */
  address -= (blkcnt * sizeof(block_t));
  sptr->blkPtr = (block_t*) address;
  sptr->blkcnt = blkcnt;
  
  if (mode == DSK_FIRST)
    {
      depcnt = *((int*)sptr->buffer);
      address = sptr->buffer + (depcnt+1)*sizeof(int); 
    }
  else
    {
      address = sptr->buffer;
    }
  
  /* set-up value pointers of block descriptors */
  for (i=0; i<blkcnt; i++)
    {
      sptr->blkPtr[i].value = (nonzero_t*) address;
      address += BlockSizeByte(sptr->blkPtr+i);
    }
}





/*
 * STRing ATTach for attaching two strings together and making a 
 * new string 
 */
char *stratt(s1, s2)
     char *s1, *s2;
{
  char *new;

  new = (char *) Malloc(strlen(s1)+strlen(s2)+1);

  strcpy(new, s1);
  strcat(new, s2);

  return(new);
}




