/*
 * factor.c
 *
 * $Source: /fs/nene/uysal/npFact/RCS/factor.c,v $
 *
 * $Header: /fs/nene/uysal/npFact/RCS/factor.c,v 1.3 1995/12/27 02:00:09 uysal Exp $
 *
 * High level algorithm and driver routines for parallel
 * out of core sparse factorization. Supernodal Fan-in style of
 * computations.
 *
 * Author : Mustafa Uysal (uysal@cs.umd.edu)
 *
 */
#include <stdlib.h>
#include <mpproto.h>

#include "mem.h"
#include "world.h"
#include "mppworld.h"
#include "kernel.h"
#include "snode.h"
#include "block.h"
#include "disk.h"
#include "comm.h"

#include "debug.h"


int rcv_msgid = 20;
int cur_snode = -1;
int s_lim     = 0;
int mem_lock[2] = {-1, -1};
int num_service = 0;
int term_cnt = 0;

extern double mem_inuse;


#define TERM_REQ    0
#define BASE_REQ    1
#define MULT_REQ    2
#define DIAG_REQ    3

#define MSG_DATA_REQ  1
#define RETRY_OFFSET 10
#define MAX_IDLE    10000000
#define TOO_MANY_MSGS  50

#define DEFAULT_SN_SIZE 0x20000   /* 128 K */


/*--------------- Statistics Variables ---------------*/
/* IO performance */
double bytes_read, read_time, iowait;
double bytes_written, write_time, ioservice;

/* kernel performance */
double k_df, k_du, k_bm, k_dm, k_dr;
double p_up, p_fa;

/* High level performance */
double etime;                     /* execution time */
double ovhd;



/*--------------- Function Prototypes ---------------*/

/* high level out-of-core sparse parallel Cholesky factorization */
void OutOfCoreFactor ARGS(( void ));

/* routine to update one supernode with another */
void PanelUpdate ARGS((block_t *mult, int cnt, snode_t *src, snode_t *dst));

/* routine to factor blocks of the pivot supernode */
void PanelFactor ARGS((block_t *diag, snode_t *pivot));

/* routine to report the performance results */
void ReportResults ARGS((void));

/* Routine to initialize all timers */
void TimerInit ARGS((void));

snode_t *GetSourceAndMult();
block_t *DiagRecv(int, char **, int*);
int DiagMapping(int);

int ServiceRemoteRequest(int *);
int *RecvRequest(void);

void TerminationLoop(void);

int atoi();

/*--------------- Implementation ---------------*/

main(argc, argv)
    int   argc;
    char *argv[];
{
  int snCount;     /* Supernode count */
  int pLevel;      /* # of parallel supernodes under update */
  int *pSlots;     /* Mask array of size pLevel for parallel slots */
  int *snOfset, *snSize;
  
  /* argument checking */
  if (argc<2) panic(("Usage : pf matrix"));
  if (argc>2) s_lim = atoi(argv[2]);
  MemInit(0);

  if (!MyProc())
    {
      message(("Parallel Out-of core Sparse Factorization\n"));
      message(("Input File : %s", argv[1]));
      print_if(s_lim>0, ("\n** EXECUTE upto %d SUPERNODES **", s_lim));
    }

  /* System initialization (for each module) */
  MPPinit();
  DiskInit(argv[1], &snCount, &snOfset, &snSize);
  SnodeInit(snCount, snOfset, snSize);
  pSlots = CommInit(&pLevel);
  TimerInit();
  KernelInit(!MyProc());

  /* Flush out I/O buffers on nodes */
  Sync();  Sync();  Sync();
  fflush(stdout);
  Sync();  Sync();  Sync();

  if (s_lim>0) snCount = s_lim;

  /* Do some useful number crunching */
  etime = CLOCK();
  OutOfCoreFactor(snCount, pSlots, pLevel);
  etime = CLOCK() - etime;
  
  /* Housekeeping */
  CommExit();
  DiskExit();
  SnodeExit();
  MemExit();

  /* Performance Report */
  ReportResults();
}





/*
 * High level algorithm for the out-of-core factorization.
 */
void OutOfCoreFactor(snCount, pMask, pLevel)
    int snCount, *pMask, pLevel;
{
  int snStrip;        /* Current strip of supernodes being updated in parallel */
  int snNext;
  int snCurrent;      /* Current supernode to be factored next */
  int count, *list;   /* List of supernode numbers that will update the current */
  int mcnt;           /* # of multiplier blocks */
  block_t *mult;      /* Array of multiplier blocks */
  snode_t *pivot;     /* pivot supernode being factored */
  snode_t *source;    /* source supernode used in the update */
  double t0, t1;
  int i, j, k;
  char *ptr1;
  int sz1, sz2;

  for (snStrip=0; snStrip<snCount; snStrip+=pLevel)
    {
      print_if(!MyProc() && !(snStrip%100), 
	       ("\n[%5d wc:%4.3lf secs cp:%4.3lf secs wait:%4.3lf secs serv:%4.3lf secs]", 
		snStrip, CLOCK()-etime, p_up+p_fa, iowait, ioservice));
      
      snNext = MIN(snStrip+pLevel, snCount);
      pLevel = snNext - snStrip;
      
      /* Parallel Pivot Update Loop - no prefetching now! */
      for (j=0; j<pLevel; cur_snode=snStrip+j++)
	{
	  /* I should have nothing to update, so carry-on */
	  if (!pMask[j]) continue;

	  /* Read in the pivot supernode */
	  t0 = CLOCK();
	  pivot = DiskReadLocal(j+snStrip);
	  t1 = CLOCK();
	  ioservice += (t1-t0);

	  /* check if I have something to update on this */
	  if (!pivot->blkcnt) continue;

	  list = SnodeUpdaters(pivot, &count);

	  /* go through the update loop */
	  for (i=0; i<count; i++)
	    {
	      mem_lock[1] = list[i];

	      /* Obtain source and multipliers */
	      source = GetSourceAndMult(pivot->sid, list[i], &mult, &mcnt, &ptr1, &sz1);

	      /* Carry out the update */
	      t0 = CLOCK();
	      PanelUpdate(mult, mcnt, source, pivot);
	      t1 = CLOCK();
	      p_up += (t1-t0);

	      SnodeFree(source);
	      if (ptr1) Free(ptr1);
	      mem_lock[1] = -1;

	      /* Perform Server Responssibilities */
	      while(ServiceRemoteRequest(RecvRequest()))
		;
	    }

	  /* factorize the diagonal if it is local */	  
	  if (pivot->blkPtr[0].xcoor == pivot->blkPtr[0].ycoor)
	    {
	      t0 = CLOCK();
	      dgfact(pivot->blkPtr);
	      t1 = CLOCK();
	      p_fa += (t1-t0);

	      mult = pivot->blkPtr;
	      sz2 = 0;
	    }
	  else
	    { /* recv if it is not */
	      mult = DiagRecv(pivot->sid, &ptr1, &sz2);
	    }

	  /* Factorize my local portions */
	  t0 = CLOCK();
	  PanelFactor(mult, pivot);
	  p_fa += (CLOCK()-t0);

	  if (sz2) Free(ptr1);

	  /*
	  printf("\n\tFinish sn:%d", pivot->sid);
	  fflush(stdout);
	  */

	  /* Write to disk */
	  DiskWriteLocal(pivot);
	  SnodeFree(pivot);
	  pivot = NULL;
	}
    }

  TerminationLoop();
}






/* Routine to factorize the current pivot */
void PanelFactor(diag, pivot)
    block_t *diag;
    snode_t *pivot;
{
  double t0,t1;
  block_t *rect;
  int i;

  if (!pivot->blkcnt) return;

  for (i=((diag==&(pivot->blkPtr[0])) ? 1 : 0); i<pivot->blkcnt; i++)
    {
      rect = &(pivot->blkPtr[i]);
      dbg_if(rect->xcoor>=rect->ycoor, 
	     panic(("error x=%d y=%d pvt=%d", rect->xcoor, rect->ycoor, pivot->sid)));
      t0 = CLOCK();
      drmod(rect, diag);
      t1 = CLOCK();
      k_dr += (t1-t0);
    }
}





/* Routine to perform the supernodal update of the pivot */
void PanelUpdate(mult, cnt, src, pvt)
    block_t *mult;
    int cnt;
    snode_t *src, *pvt;
{
  double t0,t1;
  block_t *diag;
  int i,j,k,ndx; 

  /* abort if there is nothing to update */
  if (!pvt->blkcnt || !mult) return;

  /* find out the index of the 1st base block in src */
  for (ndx=0; ndx<src->blkcnt; ndx++)
    if (src->blkPtr[ndx].ycoor > pvt->sid) break;

  /* branch out for alternative update sequences : one with digonal and one without */
  diag = &(pvt->blkPtr[0]);
  if (diag->xcoor == diag->ycoor)
    {
      for (i=0; i<cnt; i++)
        {
          /* Perform the modifications to diagonal */
	  t0 = CLOCK();
          dgupdt(diag, &mult[i]);
	  t1 = CLOCK();
	  k_du += (t1-t0);
          for (j=i+1; j<cnt; j++)
            dmod(diag, &mult[j], &mult[i]);
	  t0 = CLOCK();
	  k_dm += (t0-t1);
	  
	  for (j=ndx; j<src->blkcnt; j++)
            {
              /* match up the destination index set (which is sorted) */
              for (k=1; (k<pvt->blkcnt)&&(pvt->blkPtr[k].ycoor<src->blkPtr[j].ycoor); k++)
                ;

              /* if a match is found, fire off bmod() operations */
	      t0 = CLOCK();
              for ( ; (k<pvt->blkcnt)&&(pvt->blkPtr[k].ycoor==src->blkPtr[j].ycoor); k++)
		bmod(&(pvt->blkPtr[k]), &(src->blkPtr[j]), &mult[i]);
	      t1 = CLOCK();
	      k_bm += (t1-t0);
            }
	}
    }
  else
    {
      for (i=0; i<cnt; i++)
	for (j=ndx; j<src->blkcnt; j++)
	  {
	    /* match up the destination index set (which is sorted) */
	    for (k=0; (k<pvt->blkcnt)&&(pvt->blkPtr[k].ycoor<src->blkPtr[j].ycoor); k++)
	      ;
	    
	    /* if a match is found, fire off bmod() operations */
	    t0 = CLOCK();
	    for ( ; (k<pvt->blkcnt)&&(pvt->blkPtr[k].ycoor==src->blkPtr[j].ycoor); k++)
	      bmod(&(pvt->blkPtr[k]), &(src->blkPtr[j]), &mult[i]);
	    t1 = CLOCK();
	    k_bm += (t1-t0);
	  }
    }
}






/* Routine to report all performance results */
void ReportResults()
{
  double t_kernel, t_panel, t_diff;
  double t_comm;

  t_kernel = k_df + k_du + k_bm + k_dm + k_dr;
  t_panel  = p_fa + p_up;
  Sync(); Sync();

  IOreport();
  Sync(); Sync();

  message(("\nTOT exec:[%.3lf secs] panl:[%.3lf secs]", etime, t_panel));
  message(("\nTOT krnl:[%.3lf secs] ovhd:[%.3lf secs]", t_kernel, ovhd));
  message(("\nPNL pfct:[%.3lf secs] pupt:[%.3lf secs]", p_fa, p_up));

  Sync();  Sync();
  printf("\n");
}




/* Routine to initialize all timers */
void TimerInit()
{
  /* I/O performance */
  ioservice = iowait = bytes_read = read_time = bytes_written = write_time = 0.0;

  /* kernel performance */
  k_df = k_du = k_bm = k_dm = k_dr = p_up = p_fa = 0.0;

  /* High level performance */
  ovhd = etime = 0.0;
}




/* Routine to fetch source and multiplier */
snode_t *GetSourceAndMult(pid, sid, mptr, mcnt, p1, s1)
    int pid, sid, *mcnt;
    block_t **mptr;
    char **p1;
    int *s1;
{
  double t0, t1, t2, t3, svc_tot;
  snode_t *source;
  int Me, bMap, mMap, tmp;
  int bmsg[6], mmsg[6];
  int sbMsgId, smMsgId;
  int rbMsgId, rmMsgId;
  int btag, mtag, bsize, msize;
  int i, *metaptr;
  char *address;
  int idle, nidle;
  
  svc_tot = 0.0;
  t0 = CLOCK();

  source = NULL;
  Me = MyProc();
  *s1 = mtag = btag = 0;
  msize = bsize = DEFAULT_SN_SIZE;
  *p1 = NULL;
  
  /* Find the source supernode */
  SourceMapping(pid, sid, &bMap, &mMap, Me);

  /* Handle request for base */
  if (bMap == Me)
    {
      source = DiskReadLocal(sid);
      /*if (Me==2) printf("\nLocal Base pid:%d sid:%d got:%d", pid, sid, source->sid);*/

      if (bMap==mMap)
	{
	  for (i=0;i<source->blkcnt && source->blkPtr[i].ycoor < pid; i++)
	    ;
	  *mptr = source->blkPtr+i;
	  for (*mcnt=0;i<source->blkcnt && source->blkPtr[i].ycoor == pid; (*mcnt)++, i++)
	    ;
	  assert(*mcnt>0);
	  goto end_timer;
	}
    }
  else
    {
      source = SnodeAlloc(sid, -1, bsize);
      
      /* Make Data Request */
      bmsg[0] = sid;
      bmsg[1] = btag = rcv_msgid;
      bmsg[2] = bsize;
      bmsg[3] = Me;
      bmsg[4] = pid;
      bmsg[5] = BASE_REQ;
      mpc_send(bmsg, 6*sizeof(int), bMap, MSG_DATA_REQ, &sbMsgId);
      
      /* Post for Reply */
      mpc_recv(source->buffer, bsize, &bMap, &btag, &rbMsgId);
      if (++rcv_msgid > 60000) rcv_msgid = 20;
    }
  
  /* Handle request for multiplier */
  if ((Me!=mMap)&&(mMap!=bMap))
    {
      *p1 = (char *) Malloc(msize);
      *s1 = msize;
      *mptr = (block_t*) *p1;
      
      /* Make Data Request */
      mmsg[0] = sid;
      mmsg[1] = mtag = rcv_msgid;
      mmsg[2] = msize;
      mmsg[3] = Me;
      mmsg[4] = pid;
      mmsg[5] = MULT_REQ;
      mpc_send(mmsg, 6*sizeof(int), mMap, MSG_DATA_REQ, &smMsgId);
      
      /* Post for Reply */
      mpc_recv(*mptr, msize, &mMap, &mtag, &rmMsgId);
      if (++rcv_msgid > 60000) rcv_msgid = 20;
      /*printf("\nMultReq(sid:%d pid:%d to:%d)", sid, pid, mMap);*/
      /*printf("\n(A)pid:%d sid:%d got:%d", pid, sid, source->sid);*/
    }

  /* Wait for replys come back */
  for (nidle=idle=0; mtag || btag; idle++)
    {
      t2 = CLOCK();
      /* Perform Server Responssibilities */
      while(ServiceRemoteRequest(RecvRequest()))
	;
      t3 = CLOCK();
      svc_tot += (t3 - t2);

      if (idle>MAX_IDLE)
	{
	  message(("\nBaseMult Loop - [#rq:%d #ns:%d cur:%d b[%d@%d], m[%d@%d]", 
		   NumRequests(), FreeMsgQueue(), cur_snode, btag, bMap, mtag, mMap));
	  idle = 0;
	  nidle++;
	}

      if (nidle > 5)
	{
	  PrintRequests();
	  panic(("Exiting from idleness"));
	}

      if (mtag)
	{
	  tmp = mpc_status(rmMsgId);

	  if (tmp>0)
	    {
	      metaptr = (int*)*mptr;
	      
	      if (metaptr[0] > msize)
		{
		  i = msize;
		  *s1 = msize = metaptr[0];
		  Free(*mptr);
		  *mptr = (block_t*) Malloc(msize);
		  *p1 = (char *) *mptr;
		  mtag -= RETRY_OFFSET;
		  mpc_recv(*mptr, msize, &mMap, &mtag, &rmMsgId);
		}
	      else
		{
		  /*printf("\n(B1)pid:%d sid:%d got:%d", pid, sid, source->sid);*/
		  mtag = 0;
		  assert(mpc_status(smMsgId)>0);
		  *mcnt = metaptr[1];
		  address = *p1;
		  address += (2*sizeof(int));
		  *mptr = (block_t*) address;
		  address += ((*mcnt)*sizeof(block_t));
		  for (i=0; i<*mcnt; i++)
		    {
		      (*mptr)[i].value = (nonzero_t*) address;
		      address += BlockSizeByte((*mptr)+i);
		    }
		  /*printf("\n(B2)pid:%d sid:%d got:%d mc:%d", pid, sid, source->sid, *mcnt);*/
		}
	    }
	}

      if (btag)
	{
	  tmp = mpc_status(rbMsgId);
	  if (tmp>0)
	    {
	      metaptr = (int*)source->buffer;
	      /*printf("\n\tBaseResp(sid:%d stat:%d size:%d msize:%d)", sid, tmp, metaptr[0], bsize);*/

	      if (metaptr[0] > bsize)
		{
		  bsize = metaptr[0];
		  SnodeFree(source);
		  source = SnodeAlloc(sid, -1, bsize);
		  btag -= RETRY_OFFSET;
		  mpc_recv(source->buffer, bsize, &bMap, &btag, &rbMsgId);
		}
	      else
		{
		  assert(mpc_status(sbMsgId)>0);
		  btag = 0;
		  
		  /* set-up value pointers of block descriptors */
		  source->blkcnt = metaptr[1];
		  address = (char *) source->buffer;
		  address += (2*sizeof(int));
		  source->blkPtr = (block_t*) address;
		  address += (source->blkcnt*sizeof(block_t));
		  for (i=0; i<source->blkcnt; i++)
		    {
		      source->blkPtr[i].value = (nonzero_t*) address;
		      address += BlockSizeByte(source->blkPtr+i);
		    }

		  /*printf("\n(C)pid:%d sid:%d got:%d", pid, sid, source->sid);*/
		  if (bMap==mMap)
		    {
		      assert(source->blkPtr[0].ycoor == pid);
		      assert(mtag == 0);
		      *mptr = source->blkPtr;
		      for (*mcnt=0; source->blkPtr[*mcnt].ycoor == pid; (*mcnt)++)
			;
		      assert(*mcnt>0);
		    }
		  /*printf("\n(D)pid:%d sid:%d got:%d", pid, sid, source->sid);*/
		}
	    }
	}
    }

  /*printf("\n(D)pid:%d sid:%d got:%d", pid, sid, source->sid);*/
     
end_timer: 
  t1 = CLOCK();
  iowait += ((t1 - t0) - svc_tot);
  exit_if(source->sid!=sid, ("\nIntegrity check %d!=%d, bm:%d mm:%d pid:%d", 
			     source->sid, sid, bMap, mMap, pid));
  
  return source;
}





/* Routine to recv the diagonal */
block_t *DiagRecv(int pid, char **rptr, int *sz)
{
  int dmsg[6], dtag, dmap, dsize, *metaptr, tmp;
  char *address;
  double t0, t1, t2, t3, svc_tot;
  int sndid, rcvid, flag;
  block_t *diag;
  int idle, nidle;

  svc_tot = 0.0;
  t0 = CLOCK();

  dmap = DiagMapping(pid);
  assert(dmap != MyProc());
  
  *sz = dsize = DEFAULT_SN_SIZE;
  diag = (block_t*) Malloc(dsize);
      
  /* Make Data Request */
  dmsg[0] = pid;
  dmsg[1] = dtag = rcv_msgid;
  dmsg[2] = dsize;
  dmsg[3] = MyProc();
  dmsg[4] = pid;
  dmsg[5] = DIAG_REQ;
  if (mpc_send(dmsg, 6*sizeof(int), dmap, MSG_DATA_REQ, &sndid))
    {
      panic(("mpc_send() failed."));
    }

  /* Post for Reply */
  if (mpc_recv(diag, (size_t)dsize, &dmap, &dtag, &rcvid))
    {
      panic(("mpc_recv() failed."));
    }
  if (++rcv_msgid > 60000) rcv_msgid = 20;

  nidle = idle = flag = 1;
  while (flag)
    {
      /* Perform Server Responssibilities */
      t2 = CLOCK();
      while(ServiceRemoteRequest(RecvRequest()))
	;
      t3 = CLOCK();
      svc_tot += (t3 - t2);

      if (++idle>MAX_IDLE)
	{
	  message(("\nDiag Loop did not finish...dm:%d", dmap));
	  idle = 0;
	  nidle++;
	}

      if (nidle>6)
	{
	  PrintRequests();
	  panic(("\nExiting from idleness..."));
	}

      /* check if the service has been performed */
      tmp = mpc_status(rcvid);

      if (tmp>0)
	{
	  metaptr = (int*) diag;
	  if (metaptr[0]>dsize)
	    {
	      tmp = dsize;
	      *sz = dsize = metaptr[0];
	      Free(diag);
	      diag = (block_t*) Malloc(dsize);
	      dtag -= RETRY_OFFSET;
	      mpc_recv(diag, dsize, &dmap, &dtag, &rcvid);
	    }
	  else
	    {
	      assert(mpc_status(sndid)>0);
	      *rptr = address = (char*) diag;
	      address += sizeof(int);
	      diag = (block_t*) address;
	      address += sizeof(block_t);
	      diag->value = (nonzero_t*) address;
	      flag = 0;
	    }
	}
    }

  t1 = CLOCK();
  iowait += ((t1 - t0) - svc_tot);

  return diag;
}





/* Routine to service a nonlocal data request */
int ServiceRemoteRequest(int *req)
{
  int op, sid, pid, size, tag, to;
  snode_t *sptr;
  int lsize, i, j, k, tmp;
  char *buffer, *buf;
  double t0, t1;

  FreeMsgQueue();

  if (!req) return 0;

  sid = req[0];
  if ((cur_snode < sid))
    {
      QueueRequest(req);
      return 0;
    }
  
  if (!(++num_service % 10000))
    message(("\n\tServiced %dk requests...(%.0lf)", num_service/10000, mem_inuse));
  
  t0 = CLOCK();
  
  /* decode request */
  tag = req[1];
  size = req[2];
  to = req[3];
  pid = req[4];
  op = req[5];
  Free(req);
  if (op == TERM_REQ)
    {
      term_cnt++;
      return 1;
    }

  sptr = DiskReadLocal(sid);

  switch(op)
    {
    case DIAG_REQ:
      lsize = BlockSizeByte(&sptr->blkPtr[0]) + sizeof(int) + sizeof(block_t);
      break;

    case MULT_REQ:
      for (i=0; i<sptr->blkcnt && sptr->blkPtr[i].ycoor < pid; i++)
	;
      for (j=lsize=0; i<sptr->blkcnt && sptr->blkPtr[i].ycoor==pid; i++, j++)
	lsize += BlockSizeByte(sptr->blkPtr+i);

      lsize += (2*sizeof(int) + j*sizeof(block_t));
      break;

    case BASE_REQ:
      for (i=0; i<sptr->blkcnt && sptr->blkPtr[i].ycoor < pid; i++)
	;
      for (j=lsize=0; i<sptr->blkcnt; i++, j++)
	lsize += BlockSizeByte(sptr->blkPtr+i);

      lsize += (2*sizeof(int) + j*sizeof(block_t));      
      break;

    default:
      panic(("Invalid service request"));
    }

  if (lsize > size)
    {
      buffer = (char*) Malloc(sizeof(int));
      memcpy(buffer, &lsize, sizeof(int));
      MsgSendRec(buffer, sizeof(int), to, tag);
      tag -= RETRY_OFFSET;
    }

  buffer = (char*) Malloc(lsize);
  memcpy(buffer, &lsize, sizeof(int));  
  buf = buffer + sizeof(int);

  switch(op)
    {
    case DIAG_REQ:
      memcpy(buf, &sptr->blkPtr[0], sizeof(block_t));
      buf += sizeof(block_t);
      memcpy(buf, sptr->blkPtr[0].value, BlockSizeByte(&sptr->blkPtr[0]));
      break;

    case MULT_REQ:
    case BASE_REQ:
      memcpy(buf, &j, sizeof(int));
      buf += sizeof(int);
      for (i-=j, k=0; k<j; i++, k++)
	{
	  memcpy(buf, sptr->blkPtr+i, sizeof(block_t));
	  buf += sizeof(block_t);
	}
      for (i-=j, k=0; k<j; i++, k++)
	{
	  tmp = BlockSizeByte(sptr->blkPtr+i);
	  memcpy(buf, sptr->blkPtr[i].value, tmp);
	  buf += tmp;
	}
      break;

    default:
      panic(("Invalid service request"));
    }

  MsgSendRec(buffer, lsize, to, tag);
  /*printf("\nServiceDone(to:%d, op:%d, tag:%d, sid:%d, size:%d)", to, op, tag, sid, lsize);*/

  if (mem_lock[1]!=sptr->sid) SnodeFree(sptr);
  t1 = CLOCK();
  ioservice += (t1 - t0);
  
  return 1;
}





void MsgRecvDiscardAll()
{
  char *tbuf;
  int llen, lsrc, ltag, cnt;
  int newlen;
  
  ltag = lsrc = DONTCARE;
  while(mpc_probe(&lsrc, &ltag, &newlen)==0)
    {
      if (newlen<0) break;
      message(("\nMsgDiscard tg:%d fm:%d ln:%d", ltag, lsrc, newlen));
      tbuf = (char*) Malloc(newlen+1);
      mpc_brecv(tbuf, newlen, &lsrc, &ltag, &llen);
      Free(tbuf);
      ltag = lsrc = DONTCARE;
    }
}




void TerminationLoop(void)
{
  int obuf[6], dest;
  int *sndid, flag;
  int idle;

  sndid = (int*) Calloc(NumProcs(), sizeof(int));

  /* Make Termination Request */
  obuf[0] = -1;
  obuf[1] = -1;
  obuf[2] = -1;
  obuf[3] = -1;
  obuf[4] = -1;
  obuf[5] = TERM_REQ;
  for (dest=0; dest<NumProcs(); dest++)
    if (dest != MyProc())
      if (mpc_send(obuf, 6*sizeof(int), dest, MSG_DATA_REQ, &sndid[dest]))
	{
	  panic(("mpc_send() failed."));
	}

  for (term_cnt++, idle = 0; term_cnt < NumProcs(); )
    {
      /* Perform Server Responssibilities */
      while(ServiceRemoteRequest(RecvRequest()))
	;

      if (++idle>MAX_IDLE)
	{
	  message(("\nTermination Loop Continues: %d", term_cnt));
	  idle = 0;
	}
    }
}
