/*
 * mem.c
 *
 * $Source: /fs/nene/uysal/npFact/RCS/mem.c,v $
 *
 * $Header: /fs/nene/uysal/npFact/RCS/mem.c,v 1.1 1995/10/16 03:42:25 uysal Exp $
 *
 * Memory management module
 *
 * Author : Mustafa Uysal (uysal@cs.umd.edu)
 *
 */
#include <malloc.h>
#include "mem.h"
#include "world.h"
#include "list.h"

#define WARNING_CUTOFF    0x80000  /* 512 K */

double mem_inuse   = 0.0;
static double mem_alloced = 0.0;
static double mem_freed   = 0.0;
static list_t *mem_list = NULL;

typedef struct {
  caddr_t mref;
  double  msiz;
} memref_t;



/* Local Function Prototypes */
memref_t *AllocMemRef(caddr_t ptr, double size);
void FreeMemRef(memref_t *mptr);
void PrintMemRef(memref_t *ptr);
int CompMemRef(caddr_t ptr1, memref_t *ptr2);


/*-------------------- Implementation --------------------*/


void MemInit(int verbose)
{
  mem_list = ListNew();
}



void MemExit(void)
{
  ListDestroy(mem_list);
}



caddr_t safe_malloc(size, fn, ln)
    int size;
    char *fn;
    int ln;
{
  caddr_t ptr;

  if (!(ptr = malloc(size)))
    panic(("malloc(%d) in %s at line %d failed.", size, fn, ln));

  mem_alloced += (double) size;
  mem_inuse   += (double) size;

  /*
  if (size>WARNING_CUTOFF)
    printf("\nMemory Warning: malloc(%d, File:%s Line:%d) use:%.0lf", size, fn, ln, mem_inuse);
  */

  ListInsertTail(mem_list, AllocMemRef(ptr, (double)size));

  return ptr;
}





caddr_t safe_calloc(cnt, size, fn, ln)
    int cnt;
    int size;
    char *fn;
    int ln;
{
  caddr_t ptr;

  if (!(ptr = calloc(cnt, size)))
    panic(("calloc(%d,%d) in %s at line %d failed.", cnt, size, fn, ln));

  mem_alloced += (double) (size*cnt);
  mem_inuse   += (double) (size*cnt);

  if ((cnt*size)>WARNING_CUTOFF)
    printf("\nMemory Warning: calloc(%d,%d, File:%s Line:%d) use:%.0lf", cnt, size, fn, ln, mem_inuse);

  ListInsertTail(mem_list, AllocMemRef(ptr, (double)(size*cnt)));

  return ptr;
}




void  safe_free(ptr, fn, ln)
    caddr_t ptr;
    char *fn;
    int ln;
{
  memref_t *mptr;

  exit_if(!ptr, ("Trying to free a null pointer [Line %d of %s]", ln, fn));

  mptr = (memref_t*) ListFindDelete(mem_list, ptr, CompMemRef);

  if (mptr)
    {
      mem_freed += mptr->msiz;
      mem_inuse -= mptr->msiz;
    }
  else
    {
      printf("\nPointer %x is not in the list", ptr);
    }

  FreeMemRef(mptr);  
  free(ptr);
}





memref_t *AllocMemRef(caddr_t ptr, double size)
{
  memref_t *new;

  new = malloc(sizeof(memref_t));
  if (!new) panic(("malloc(%d) failed", size));

  mem_alloced += (double) (sizeof(memref_t));
  mem_inuse   += (double) (sizeof(memref_t));

  new->mref = ptr;
  new->msiz = size;

  return new;
}





void FreeMemRef(memref_t *mptr)
{
  if (!mptr)
    {
      printf("\nTrying to free a null pointer\n");
      ListPrintAll(mem_list, PrintMemRef, "MemList");
      panic(("\nTrying to free a null pointer\n"));
    }

  free(mptr);

  mem_freed += (double) (sizeof(memref_t));
  mem_inuse -= (double) (sizeof(memref_t));
}




int CompMemRef(caddr_t ptr1, memref_t *ptr2)
{
  if (ptr1 == ptr2->mref) return 0;
  else return 1; 
}



void PrintMemRef(memref_t *ptr)
{
  printf("\nMemRef[p:%x, s:%.0lf]", ptr->mref, ptr->msiz);
}
