/*
 * kernel.c
 *
 * $Source: /fs/nene/uysal/npFact/RCS/zkernel.c,v $
 *
 * $Header: /fs/nene/uysal/npFact/RCS/zkernel.c,v 1.1 1995/10/16 03:42:25 uysal Exp $
 *
 * Numerical computation kernels for the Cholesky factorization. 
 * Complex Arithmetic
 * 
 * Note :
 * These kernels are stripped down versions of the in-core kernels and there
 * are more restrictions on the use of them all of which should be handled
 * in the driver routine.
 * 
 * Author : Mustafa Uysal <uysal@cs.umd.edu>
 *
 */
#include <stdio.h>
#include <math.h>

#include "world.h"
#include "block.h"
#include "kernel.h"
#include "blas.h"
#include "complex.h"
#include "debug.h"

#define LEVEL3

#define CX_SQRTFP   23  /* FLOP values for different basic OPS (Complex kernels) */
#define CX_DIVFP    17
#define CX_RECIPFP  17
#define CX_MULTFP   6
#define CX_ADDFP    2
#define CX_MUADDFP  8


#if defined(LEVEL3)
nonzero_t DOT();
#endif defined(LEVEL3)


#if   !defined(COMPLEX)   /* Code for Complex Arithmetic */
#error "zkernel.c contains only COMPLEX Arithmetic kernels"
#endif 

static int iONE;
static fp_t FPZERO, FPONE;
static nonzero_t MONE, PONE, ZERO;




/*-------------------- Implementation --------------------*/


void KernelInit(verbose)
    int verbose;
{
#if defined(LEVEL0)
  print_if(verbose, ("\nKernel - User-Level subroutines"));
#elif defined(LEVEL1)
  print_if(verbose, ("\nKernel - Level-1 BLAS subroutines"));
#elif defined(LEVEL3)
  print_if(verbose, ("\nKernel - Level-3 BLAS subroutines"));
#endif

  FPZERO = (fp_t) 0.0;  
  FPONE  = (fp_t) 1.0;
  iONE  = 1;

  CX_INIT(ZERO, FPZERO, FPZERO);
  CX_INIT(PONE, FPONE, FPZERO);
  CX_INIT(MONE, -FPONE, FPZERO);
}




/*------------------------------------------------------------*/
/*            User Level Numerical Kernels                    */
/*------------------------------------------------------------*/
#if defined(LEVEL0)  /* Level-0 BLAS Kernels */
void dgfact(dblk)
    block_t *dblk;
{
  nonzero_t mult;  /* Equivalent to 1/diagonal */
  int   mindex;    /* Index of the first element in the multipler row */
  int   mincr;     /* Increment to find the next multiplier row */
  int   dindex;    /* Index of the current diagonal element (start of column) */
  int   cindex;
  int   base;
  int   uincr;
  int   j, mptr, bptr;
  int   size;

  size = BlockSize(dblk);
  for (mindex=mincr=0; mindex<size; mindex+=(++mincr))
    {
      /* Update L(j,j) */
      dindex = mindex + mincr;
      for (j=mindex; j<dindex; j++)
        CX_MULTSUB(dblk->value[dindex], dblk->value[j], dblk->value[j]);
      
#ifdef PDEF_CHECK
      if (CX_LEQ(dblk->value[dindex], ZERO))
	CX_ASSIGN(dblk->value[dindex], PONE);
#endif

      /* Compute L(j,j) */
      CX_SQRT(dblk->value[dindex]);
      CX_DIV(mult, PONE, dblk->value[dindex]);

      /* Update elements that are in the same column as dindex */
      uincr = mincr+1;
      for (base=mindex+uincr, cindex=dindex+uincr; base<size; base+=uincr, cindex+=uincr) 
	{
	  for (mptr=mindex, bptr=base; bptr<cindex; bptr++, mptr++) 
            CX_MULTSUB(dblk->value[cindex], dblk->value[mptr], dblk->value[bptr]);
	  ++uincr;
	}
      
      /* Scale subdiagonal elements i+1:size */
      for (uincr=mincr+1, dindex+=(mincr+1); dindex<size; dindex+=(++uincr))
        CX_MULTFCT(dblk->value[dindex], mult);
    }
}





/*
 * dgupdt()
 *
 * Computes the update of a rectangular dense block to a diagonal.
 */
void dgupdt(diag, base)
     block_t *diag;     /* Diagonal being updated */
     block_t *base;     /* Updating block */
{
  nonzero_t *basptr; /* Pointer to the base row */
  nonzero_t *mulptr; /* Pointer to the multiplier row */
  int   bhsize;      /* horizontal size of the updating block */
  int   bvsize;      /* vertical size of the updating block */
  int   dindex;      /* Index of the diagonal block coordinate being updated */
  int   mndxst;      /* Saved value of dest index for next multiplier */
  int   mincr;       /* Saved value of increment for next multipler */
  int   dincr;       /* increment of the dindex for next iteration */
  int   colno;       /* Difference between the first updated column and the leftmost
		      * border of the diagonal block */
  int   i,j,k;
  int   dgsize;
  
  dgsize = BlockSize(diag);
  mulptr = base->value;
  bvsize = base->lrow - base->urow;
  bhsize = base->rcol - base->lcol;
  colno  = base->urow - diag->urow;
  mndxst = colno*(colno+1)/2+colno;
  mincr  = colno;

  for (i=0; i<bvsize; i++, mulptr+=bhsize, mndxst+=(++mincr+1))
    {
      dindex = mndxst;
      dincr  = mincr;
      basptr = mulptr;
      for (j=i; j<bvsize; j++, basptr+=bhsize, dindex+=(++dincr))
	for (k=0; k<bhsize; k++)
          CX_MULTSUB(diag->value[dindex], mulptr[k], basptr[k]);
    }
}






/*
 * bmod()
 *
 * Computes base-multiplier updates to non-diagonal dense blocks.
 *
 */
void bmod(dest, base, mult)
     block_t *dest, *base, *mult;
{
  nonzero_t *mulptr;    /* Pointer to the current multiplier row */
  nonzero_t *basptr;    /* Pointer to the current base row */
  nonzero_t *mulend;    /* Pointer to the last element in the multiplier */
  nonzero_t *bassta;    /* Pointer to the first element in the base */
  int    bhsize;    /* Horizontal size of base block */
  int    bfrstrow;  /* First row in the update from the base */
  int    blastrow;  /* Last row in the update from the base */
  int    dhsize;    /* Horizontal size of the destination block */
  int    dfrstcol;  /* First column in the update from destination */
  int    dfrstrow;  /* First row in the update from destination */
  int    dstndxst;  /* Storage for the destination index */
  int    dstndx;    /* Index of the destination element being updated */
  int    k, j;      /* Innermost loop variable */

  /* return if alignment mismatch */
  if (base->lrow<=dest->urow) return;
  if (base->urow>=dest->lrow) return;

  /* Initialization of Multiplier variables */
  mulend = mult->value + BlockSize(mult);

  /* Initialization of Base variables */
  bhsize   = base->rcol - base->lcol;
  bfrstrow = MAX(dest->urow, base->urow) - base->urow;
  blastrow = MIN(dest->lrow, base->lrow) - base->urow;
  bassta   = base->value + bfrstrow * bhsize;
  
  /* Initialization of Destination variables */
  dhsize   = dest->rcol - dest->lcol;
  dfrstcol = mult->urow - dest->lcol;
  dfrstrow = MAX(dest->urow, base->urow) - dest->urow;
  dstndxst = dfrstrow * dhsize + dfrstcol;
  
  /* allegedly efficient compute kernel */
  for (mulptr = mult->value; mulptr<mulend; mulptr+=bhsize)
    {
      dstndx = dstndxst++;
      for (j=bfrstrow; j<blastrow; dstndx+=dhsize, j++)
	{
	  basptr = base->value + j*bhsize;
	  for (k=0; k<bhsize; k++)
	    CX_MULTSUB(dest->value[dstndx], mulptr[k], basptr[k]);
	}
    }
}





/*
 * dmod()
 *
 * Computes the update of two Rectangles on a diagonal block.
 *
 */
void dmod(dest, base, mult)
     block_t *dest, *base, *mult;
{
  nonzero_t *mulptr;    /* Pointer to the current multiplier row */
  nonzero_t *basptr;    /* Pointer to the current base row */
  nonzero_t *mulend;    /* Pointer to the last element in the multiplier */
  int    bvsize;    /* Vertical size of base block */
  int    bhsize;    /* Horizontal size of base block */
  int    dfrstcol;  /* First column in the update from destination */
  int    dfrstrow;  /* First row in the update from destination */
  int    dstndx;    /* Index of the destination element being updated */
  int    dstndxst;  /* Stored value of dstndx */
  int    dincr;     /* Increment variable for the destination index */
  int    k,j;       /* Innermost loop variable */

#if defined(SNUPD_LOG)
  if (dest->xcoor == PVT_WATCH)
    printf("\n@DMOD[%d,%d] <- [%d,%d] * [%d, %d]", dest->urow, dest->lcol,
	   base->urow, base->lcol, mult->urow, mult->lcol);
#endif

  /* Initialization of Multiplier variables */
  mulend = mult->value + BlockSize(mult);
  
  /* Initialization of Base variables */
  bhsize = base->rcol - base->lcol;
  bvsize = base->lrow - base->urow;
  
  /* Initialization of Destination variables */
  dfrstcol = mult->urow - dest->lcol;
  dfrstrow = base->urow - dest->urow;
  dstndxst = dfrstrow*(dfrstrow+1)/2+dfrstcol;

  /* Allegedly efficient compute kernel for this op */
  for (mulptr=mult->value; mulptr<mulend; mulptr+=bhsize)
    {
      dstndx = dstndxst++;
      dincr  = dfrstrow;
      for (basptr=base->value, j=0; j<bvsize; j++, basptr+=bhsize, dstndx+=(++dincr))
	for (k=0; k<bhsize; k++)
          CX_MULTSUB(dest->value[dstndx], mulptr[k], basptr[k]);
    }
}






/*
 * drmod()
 *
 * Computes the Subdiagonal rect Update of a diagonal block
 */
void drmod(blk, diag)
    block_t *blk;     /* block being factored */
    block_t *diag;    /* diagonal block */
{
  nonzero_t *mulptr;     /* Pointer to the multipler element */
  nonzero_t multip;      /* Multiplier for the column scale in dest block */
  int   dhsize;      /* Width of the current block panel */
  int   dupdndx;     /* Index for the updated destination index */
  int   dbasndx;     /* Index of the destination base elemen */
  int   i,j;         /* loop variables */
  int   size;

  /* Initialization */
  mulptr = diag->value;
  dhsize = blk->rcol - blk->lcol;
  size   = BlockSize(blk);

  /* Allegedly efficient Compute Kernel */
  for (i=0; i<dhsize; i++)
    {
      /* Update the new column */
      for (j=0; j<i; j++)
	{
	  for (dupdndx=i,dbasndx=j; dupdndx<size; dbasndx+=dhsize,dupdndx+=dhsize)
	    CX_MULTSUB(blk->value[dupdndx], (*mulptr), blk->value[dbasndx]);
	  mulptr++;
	}
      
      /* Scale the subdiagonal elements */
      CX_DIV(multip, PONE, (*mulptr));
      mulptr++;
      for (j=i; j<size; j+=dhsize)
        CX_MULTFCT(blk->value[j], multip);
    }
}

#endif defined(LEVEL0)






/*------------------------------------------------------------*/
/*------------------------------------------------------------*/
/*                                                            */
/*          Numerical Kernels with Level 1 BLAS               */
/*                                                            */
/*------------------------------------------------------------*/
/*------------------------------------------------------------*/

#if defined(LEVEL1)
#error "Level 1 BLAS Implementation is not ready for COMPLEX Arithmetic"
#endif defined(LEVEL1)








/*------------------------------------------------------------*/
/*------------------------------------------------------------*/
/*                                                            */
/*          Numerical Kernels with Level 3 BLAS               */
/*                                                            */
/*------------------------------------------------------------*/
/*------------------------------------------------------------*/

#if defined(LEVEL3)
void dgfact(dblk)
    block_t *dblk;
{
  nonzero_t mult;  /* Equivalent to 1/diagonal */
  nonzero_t dres;  /* Result of dot product */
  int   mindex;    /* Index of the first element in the multipler row */
  int   mincr;     /* Increment to find the next multiplier row */
  int   dindex;    /* Index of the current diagonal element (start of column) */
  int   cindex;
  int   base;
  int   uincr;
  int   j, mptr, bptr;
  int   size;

  size = BlockSize(dblk);
  for (mindex=mincr=0; mindex<size; mindex+=(++mincr))
    {
      /* Update L(j,j) */
      dindex = mindex + mincr;
      dres = DOT(&mincr, dblk->value+mindex, &iONE,dblk->value+mindex, &iONE);
      CX_DEC(dblk->value[dindex], dres);

#ifdef PDEF_CHECK
      if (CX_LEQ(dblk->value[dindex], ZERO))
	CX_ASSIGN(dblk->value[dindex], PONE);
#endif

      /* Compute L(j,j) */
      CX_SQRT(dblk->value[dindex]);
      CX_DIV(mult, PONE, dblk->value[dindex]);

      /* Update elements that are in the same column as dindex */
      uincr = mincr+1;
      for (base=mindex+uincr, cindex=dindex+uincr; base<size; base+=uincr, cindex+=uincr) 
	{
	  dres = DOT(&mincr, dblk->value+mindex, &iONE,dblk->value+base, &iONE);
	  CX_DEC(dblk->value[dindex], dres);
	  ++uincr;
	}
      
      /* Scale subdiagonal elements i+1:size */
      for (uincr=mincr+1, dindex+=(mincr+1); dindex<size; dindex+=(++uincr))
        CX_MULTFCT(dblk->value[dindex], mult);
    }
}





/*
 * dgupdt()
 *
 * Computes the update of a rectangular dense block to a diagonal.
 */
void dgupdt(diag, base)
     block_t *diag;     /* Diagonal being updated */
     block_t *base;     /* Updating block */
{
  nonzero_t *basptr; /* Pointer to the base row */
  nonzero_t *mulptr; /* Pointer to the multiplier row */
  nonzero_t dres;    /* Result of DOT product */
  int   bhsize;      /* horizontal size of the updating block */
  int   bvsize;      /* vertical size of the updating block */
  int   dindex;      /* Index of the diagonal block coordinate being updated */
  int   mndxst;      /* Saved value of dest index for next multiplier */
  int   mincr;       /* Saved value of increment for next multipler */
  int   dincr;       /* increment of the dindex for next iteration */
  int   colno;       /* Difference between the first updated column and the leftmost
		      * border of the diagonal block */
  int   i,j,k;
  int   dgsize;
  
  dgsize = BlockSize(diag);
  mulptr = base->value;
  bvsize = base->lrow - base->urow;
  bhsize = base->rcol - base->lcol;
  colno  = base->urow - diag->urow;
  mndxst = colno*(colno+1)/2+colno;
  mincr  = colno;

  for (i=0; i<bvsize; i++, mulptr+=bhsize, mndxst+=(++mincr+1))
    {
      dindex = mndxst;
      dincr  = mincr;
      basptr = mulptr;
      for (j=i; j<bvsize; j++, basptr+=bhsize, dindex+=(++dincr))
	{
	  dres = DOT(&bhsize, mulptr, &iONE, basptr, &iONE);
	  CX_DEC(diag->value[dindex], dres);
	}
    }
}






/*
 * bmod()
 *
 * Computes base-multiplier updates to non-diagonal dense blocks.
 *
 */
void bmod(dest, base, mult)
     block_t *dest, *base, *mult;
{
  nonzero_t *mulptr;    /* Pointer to the current multiplier row */
  nonzero_t *basptr;    /* Pointer to the current base row */
  nonzero_t *mulend;    /* Pointer to the last element in the multiplier */
  nonzero_t *bassta;    /* Pointer to the first element in the base */
  nonzero_t *deststa;   /* Pointer to the first element in the dest */
  int    bhsize;    /* Horizontal size of base block */
  int    bfrstrow;  /* First row in the update from the base */
  int    blastrow;  /* Last row in the update from the base */
  int    dhsize;    /* Horizontal size of the destination block */
  int    dfrstcol;  /* First column in the update from destination */
  int    dfrstrow;  /* First row in the update from destination */
  int    dstndxst;  /* Storage for the destination index */
  int    dstndx;    /* Index of the destination element being updated */
  int    bvsize  ;  /* Vertical size of the base block */
  int    mvsize;    /* Vertical size of the multiplier block */
  int    k, j;      /* Innermost loop variable */
  char   yt,nt;

  /* return if alignment mismatch */
  if (base->lrow<=dest->urow) return;
  if (base->urow>=dest->lrow) return;

  /* Initialization of Multiplier variables */
  mulend = mult->value + BlockSize(mult);
  mvsize = mult->lrow - mult->urow;

  /* Initialization of Base variables */
  bhsize   = base->rcol - base->lcol;
  bfrstrow = MAX(dest->urow, base->urow);
  bvsize   = MIN(dest->lrow, base->lrow) - bfrstrow;
  blastrow = MIN(dest->lrow, base->lrow) - base->urow;
  bassta   = base->value + (bfrstrow - base->urow) * bhsize;
  
  /* Initialization of Destination variables */
  dhsize   = dest->rcol - dest->lcol;
  dfrstcol = mult->urow - dest->lcol;
  dfrstrow = MAX(dest->urow, base->urow) - dest->urow;
  dstndxst = dfrstrow * dhsize + dfrstcol;
  deststa  = dest->value + dstndxst;

  /* BLAS specific variables */
  nt = 'N'; yt = 'T';
  
  GEMM(&yt,&nt,&mvsize,&bvsize,&bhsize,&MONE,mult->value,&bhsize,
       bassta, &bhsize, &PONE, deststa, &dhsize);
}





/*
 * dmod()
 *
 * Computes the update of two Rectangles on a diagonal block.
 *
 */
void dmod(dest, base, mult)
     block_t *dest, *base, *mult;
{
  nonzero_t *mulptr;    /* Pointer to the current multiplier row */
  nonzero_t *basptr;    /* Pointer to the current base row */
  nonzero_t *mulend;    /* Pointer to the last element in the multiplier */
  nonzero_t dres;       /* Result of DOT product */
  int    bvsize;    /* Vertical size of base block */
  int    bhsize;    /* Horizontal size of base block */
  int    dfrstcol;  /* First column in the update from destination */
  int    dfrstrow;  /* First row in the update from destination */
  int    dstndx;    /* Index of the destination element being updated */
  int    dstndxst;  /* Stored value of dstndx */
  int    dincr;     /* Increment variable for the destination index */
  int    k,j;       /* Innermost loop variable */

  /* Initialization of Multiplier variables */
  mulend = mult->value + BlockSize(mult);
  
  /* Initialization of Base variables */
  bhsize = base->rcol - base->lcol;
  bvsize = base->lrow - base->urow;
  
  /* Initialization of Destination variables */
  dfrstcol = mult->urow - dest->lcol;
  dfrstrow = base->urow - dest->urow;
  dstndxst = dfrstrow*(dfrstrow+1)/2+dfrstcol;

  /* Allegedly efficient compute kernel for this op */
  for (mulptr=mult->value; mulptr<mulend; mulptr+=bhsize)
    {
      dstndx = dstndxst++;
      dincr  = dfrstrow;
      for (basptr=base->value, j=0; j<bvsize; j++, basptr+=bhsize, dstndx+=(++dincr))
	{
	  dres = DOT(&bhsize, mulptr, &iONE, basptr, &iONE);
	  CX_DEC(dest->value[dstndx], dres);
	}
    }
}






/*
 * drmod()
 *
 * Computes the Subdiagonal rect Update of a diagonal block
 */
void drmod(blk, diag)
    block_t *blk;     /* block being factored */
    block_t *diag;    /* diagonal block */
{
  nonzero_t *mulptr;     /* Pointer to the multipler element */
  nonzero_t multip;      /* Multiplier for the column scale in dest block */
  int   dhsize;      /* Width of the current block panel */
  int   dvsize;      /* Height of the block */
  int   dupdndx;     /* Index for the updated destination index */
  int   dbasndx;     /* Index of the destination base elemen */
  int   i,j;         /* loop variables */
  int   size;

  /* Initialization */
  mulptr = diag->value;
  dhsize = blk->rcol - blk->lcol;
  dvsize = blk->lrow - blk->urow;
  size   = BlockSize(blk);

  /* Allegedly efficient Compute Kernel */
  for (i=0; i<dhsize; i++)
    {
      /* Update the new column */
      for (j=0; j<i; j++)
	{
          CX_ASSIGN(multip, *mulptr);
          CX_NEGATE(multip);
          AXPY(&dvsize, &multip, blk->value+j, &dhsize, blk->value+i, &dhsize);
	  mulptr++;

	  /*
	  for (dupdndx=i,dbasndx=j; dupdndx<size; dbasndx+=dhsize,dupdndx+=dhsize)
	    blk->value[dupdndx] -= (*mulptr) * blk->value[dbasndx];
	  */
	}
      
      /* Scale the subdiagonal elements */
      CX_DIV(multip, PONE, (*mulptr));
      mulptr++;

      SCAL(&dvsize, &multip, blk->value+i, &dhsize);
    }
}





nonzero_t DOT(N, X, INCX, Y, INCY)
    nonzero_t *X, *Y;
    int  *N, *INCX, *INCY; 
{
  nonzero_t acc;
  int xi,yi;

  CX_INIT(acc, FPZERO, FPZERO);
  
  for (xi=0; xi<*N; X+=*INCX, Y+=*INCY, xi++)
    CX_MULTADD(acc, (*X), (*Y));
  
  return(acc);
}

#endif defined(LEVEL3)
