/*
 * comm.c
 *
 * $Source: /fs/nene/uysal/pf/newFact/RCS/comm.c,v $
 *
 * $Header: /fs/nene/uysal/pf/newFact/RCS/comm.c,v 1.2 1995/10/16 04:24:45 uysal Exp $
 *
 * Interprocessor communication module
 *
 * Author : Mustafa Uysal (uysal@cs.umd.edu)
 *
 */
#include <math.h>
#include <memory.h>

#include "world.h"
#include "mppworld.h"
#include "queue.h"
#include "block.h"
#include "comm.h"
#include "snode.h"
#include "mem.h"

#include "debug.h"



/*--------------- Symbolic Definitions ---------------*/

#define MSG_SNSIZE       1
#define MSG_SNDATA   11001
#define MSG_SNREQ    22001

#define MSG_MULCNT    5001
#define MSG_MULBLK   15001
#define MSG_MULDAT   25001

#define MSG_DIAG     30001

#define MUL_ASYNC


/*-------------- Global Variable Declarations ----------------*/

static int **pgrid;    /* Processor grid */
static int gridx;      /* x dimension size */
static int gridy;      /* y dimension size */
static int *slotMask;  /* Mask array for existence in a slot */
static int mone = -1;

extern double k_df, c_mula, c_mulb;
extern int c_vola, c_volb;

/*--------------- Local Function Prototypes ---------------*/

/* Routine to send a supernode to another processor */
void CommSnodeSend ARGS((snode_t *sptr, int from, int to));

/* Routine to compute the processor grid for communication */
void FormProcessorGrid ARGS((int np, int me));

block_t *CommMultSend ARGS(());




/*-------------- Function Implementations ---------------------*/


/* Initialization of communication module */
int *CommInit(nslot)
    int *nslot;
{
  int i,j;

  /* initialize interprocessor communication grid */
  FormProcessorGrid(NumProcs(), MyProc());
  *nslot = gridx;

  slotMask = (int*) Calloc(gridx, sizeof(int));
  for (i=0; i<gridx; i++)
    {
      slotMask[i] = 0;
      for (j=0; j<gridy; j++)
	if (pgrid[j][i] == MyProc()) slotMask[i] = 1;
    }

  return slotMask;
}



/* Cleanup function for the communications module */
void CommExit()
{
}


void CommBlockSend(pid, src, me)
    int pid;       /* ID of the pivot supernode */
    snode_t *src;  /* source supernode struct */
{
  double t0,t1;
  int pmap, smap;
  int i;
  
  pmap = pid % gridx;
  smap = src->sid % gridx;
  t0 = CLOCK();
  for (i=0; i<gridy; i++)
    if (pgrid[i][smap]==me && pgrid[i][pmap]!=me)
      CommSnodeSend(src, me, pgrid[i][pmap], pid);
  t1 = CLOCK();
  c_mula += (t1-t0);
}



/* Routine to shuffle blocks to where they needed */
block_t *CommBlockReceive(pid, src, me, count, slab_lead)
    int pid;       /* ID of the pivot supernode */
    snode_t *src;  /* source supernode struct */
    int me;
    int *count, slab_lead;
{
  double t0,t1;
  block_t *ret;
  caddr_t buf;
  int pmap, smap;
  int sz, fm, tg;
  int i, mrcv_flg, msnd_flg;
  
  pmap = pid % gridx;
  smap = src->sid % gridx;
  ret  = (block_t*) NULL;

  msnd_flg = mrcv_flg = 1;

  t0 = CLOCK();
  for (i=0; i<gridy; i++)
    {
      if (pgrid[i][pmap]==me && pgrid[i][smap]!=me)
	{
 	  sz = -1;
	  tg = MSG_SNDATA+src->sid;
	  fm = pgrid[i][smap];
	  buf = RecvQueuePoll(&tg, &fm, &sz);

	  dbg_if(!buf, MPPstatus());
	  dbg_msg(5, ("\nRecvd Updates to:%d by:%d size:%d", pid, src->sid, sz));
  
	  /* incorporate all from the network into source */
	  if (sz)
	    SnodeBlockMerge(src, buf, sz, pid);
	  else
	    Free(buf); /* of 1 byte */
	}
      if (pgrid[i][smap]==me && msnd_flg)
	{
	  msnd_flg = 0;
	  if (ret = CommMultSend(src, pid, me, pmap, count))
	    mrcv_flg = 0;
	}
      if (pgrid[i][pmap]==me && mrcv_flg)
	{
	  mrcv_flg = 0;

	  CommAsyncRecv(NULL, pgrid[pid%gridy][smap], MSG_MULBLK+src->sid, -1);
	  CommAsyncRecv(NULL, pgrid[pid%gridy][smap], MSG_MULDAT+src->sid, -1);
	}
    }
  
  
  t1 = CLOCK();
  c_mula += (t1-t0);

  if (!mrcv_flg && !ret)
    {
      tg = MSG_MULBLK+src->sid;
      fm = pgrid[pid%gridy][smap];
      sz = -1;
      ret = (block_t *) RecvQueuePoll(&tg, &fm, &sz);
      *count = sz / sizeof(block_t);

      tg = MSG_MULDAT+src->sid;
      fm = pgrid[pid%gridy][smap];
      sz = -1;
      buf = (caddr_t) RecvQueuePoll(&tg, &fm, &sz);
      
      /* add the buffer space to the queue of reclaimables */
      QueueAdd(src->bufq, ret);
      QueueAdd(src->bufq, buf);

      /* setup the pointers in the multipliers correct */
      for (i=0; i<*count; i++)
	{
	  ret[i].value = (nonzero_t*) buf;
	  buf += BlockSizeByte(&ret[i]);
	}
    }
  
  t0 = CLOCK();
  c_mulb += (t0-t1);

  return ret;
}





/* Routine to send a supernode to another processor */
void CommSnodeSend(sptr, from, to, pvt)
    snode_t *sptr;
    int from, to;
    int pvt;
{
  int size, ndx, enable;

  if (to == from)   return;

  for (ndx = 0; ndx<sptr->blkcnt && sptr->blkPtr[ndx].ycoor<pvt; ndx++)
    ;
  enable = (ndx < sptr->blkcnt) ? 1 : 0;

  size = sptr->size - ((caddr_t)(sptr->blkPtr[ndx].value) - (caddr_t)sptr->buffer);

  CommAsyncSend(enable ? (caddr_t) sptr->blkPtr[ndx].value : (caddr_t) sptr->buffer, 
		to, MSG_SNDATA+sptr->sid, enable*size);

  dbg_msg(5, ("\nUpdates to:%d by:%d sz:%d", pvt, sptr->sid, enable*size));

  c_vola += sptr->size;
}






block_t *CommMultSend(src, id, me, map, count)
    snode_t *src;
    int id, me, map;
    int *count;
{
  block_t *bptr, *mult;
  int bcnt;
  int i,j,k;
  int lo, hi, cnt;

  bcnt = src->blkcnt;
  bptr = src->blkPtr;
  mult = (block_t*) NULL;
  *count = 0;

  for (i=0;(i<bcnt) && (bptr[i].ycoor<id); i++)
    ;
  for ( hi=lo=i; (hi<bcnt) && (bptr[hi].ycoor == id); hi++)
    ;
  if (hi>lo) /* sender */
    {
       /* compute send parameters */
      for (cnt=0, i=lo; i<hi; i++)
	cnt += BlockSizeByte(&(bptr[i]));
      
      /* initiate the send operations */
      for (i=0; i<gridy; i++)
	{
	  /* don't send to myself */
	  if (pgrid[i][map] == me) 
	    {
	      mult = &bptr[lo];
	      *count = hi - lo;
	      continue;
	    }
	  /* don't send twice */
	  for (j=0; j<i; j++)
	    if (pgrid[j][map] == pgrid[i][map]) break;
	  
	  /* Following Sends should complete before going out of context */
	  if (j == i)
	    {
	      CommAsyncSend(&bptr[lo], pgrid[i][map], MSG_MULBLK+src->sid,
			    (hi-lo)*sizeof(block_t));
	      CommAsyncSend(bptr[lo].value, pgrid[i][map], MSG_MULDAT+src->sid, cnt);
	      c_volb += ((hi-lo)*sizeof(block_t)+cnt);
	    }
	}
    }

  return mult;
}







/* Routine to multicast the diagonal of the pivot */
block_t *CommDiagRecv(sptr, me)
    snode_t *sptr;
    int me;
{
  double t0,t1;
  block_t *bptr;
  int i, j, map, size;
  caddr_t buffer;

  map = sptr->sid % gridx;

  /* handle the case where you don't care about the diagonal */
  if (!sptr->blkcnt)
    {
      size = -1;
      map  = DONTCARE;
      j    = MSG_DIAG+sptr->sid;
      CommAsyncRecv(NULL, map, j, size);
      buffer = RecvQueuePoll(&j, &map, &size);
      dbg_msg(5, ("\n\td-out:%d fm:%d sz:%d j=%d", sptr->sid, map, size, j));
      Free(buffer);
      return (block_t*) NULL;
    }

  /* This is the case where you do care about it */
  if (sptr->blkPtr[0].ycoor == sptr->sid)  /* sender */
    {
      /* factorize the diagonal before sending */
      t0 = CLOCK();
      dgfact(&(sptr->blkPtr[0]));
      t1 = CLOCK();
      k_df += (t1-t0);

      for (i=0; i<gridy; i++)
	{
	  if (pgrid[i][map] == me) continue;
	  for (j=0; j<i; j++)
	    if (pgrid[j][map] == pgrid[i][map]) break;

	  if (j==i)
	    {
	      size = BlockSizeByte(&sptr->blkPtr[0]);
	      CommAsyncSend(sptr->blkPtr[0].value, pgrid[i][map],MSG_DIAG+sptr->sid, size);
	      dbg_msg(5, ("\n\td-snd:%d to:%d sz:%d", sptr->sid, pgrid[i][map], size));
	    }
	}

      bptr = &sptr->blkPtr[0];
    }
  else                                     /* receiver(s) */
    {
      /* initialize the diagonal descriptor */
      bptr = (block_t*) Malloc(sizeof(block_t));
      
      bptr->xcoor = bptr->ycoor = sptr->sid;
      bptr->lcol  = bptr->urow  = sptr->blkPtr[0].lcol;
      bptr->rcol  = bptr->lrow  = sptr->blkPtr[0].rcol;
      
      size = BlockSizeByte(bptr);
      bptr->value = (nonzero_t*) Malloc(size);
      MsgRecv(MSG_DIAG+sptr->sid, bptr->value, size);

      dbg_msg(5, ("\n\td-rcv:%d sz:%d", sptr->sid, size));
      
      QueueAdd(sptr->bufq, bptr);
      QueueAdd(sptr->bufq, bptr->value);
    }

  return bptr;
}







/* Routine to compute the processor grid for communication */
void FormProcessorGrid(np, me)
    int np;     /* number of processors */
    int me;     /* my processor id */
{
  double tmp;      /* to get the result of sqrt */
  int *gptr;       /* pointer to a row of pgrid */
  int i,j,k,cnt;   /* control vars */

  /* find out grid parameters */
  tmp = sqrt((double)np);
  gridy = (int) tmp;
  gridx = np/gridy;
  if ((gridy*gridx)<np)
    gridx++;

  /* storage allocation */
  pgrid = (int**) Calloc(gridy, sizeof(int*));

  /* fill-up the processor grid */
  for (cnt=i=0; i<gridy; i++)
    {
      gptr = pgrid[i] = (int*) Calloc(gridx, sizeof(int));
      for (j=0; j<gridx; j++, cnt++)
	gptr[j] = cnt % np;
    }

  if (!MyProc())
    {
      printf("\n-----------------------------------------------");
      printf("\n Grid - np=%d me=%d x=%d, y=%d", np, me, gridx, gridy);
      printf("\n-----------------------------------------------");
      
      for (i=0; i<gridy; i++)
	{
	  printf("\n %d : ", i);
	  for (j=0; j<gridx; j++)
	    printf("%3d ", pgrid[i][j]);
	}
    }
}




/* Routine to inform Network Interface that we are waiting a msg */
void NetPrefetchBase(sid, pid)
    int sid, pid;  /* source and pivot IDs, resp. */
{
  int pmap, smap;
  int i, me;
  
  pmap = pid % gridx;
  smap = sid % gridx;
  me = MyProc();

  if (pmap != smap)
    for (i=0; i<gridy; i++)
      if (pgrid[i][pmap]==me && pgrid[i][smap]!=me)
	CommAsyncRecv(NULL, pgrid[i][smap], MSG_SNDATA+sid, -1);
}




/* Routine to enable BASE block communications */
void EnableBASEcomm(sid, pid)
    int sid, pid;  /* source and pivot IDs, resp. */
{
  int pmap, smap;
  int i, me;
  
  pmap = pid % gridx;
  smap = sid % gridx;
  me = MyProc();

  dbg_msg(0, ("\nEnable(%d,%d)", sid, pid));
  
  if (pmap != smap)
    for (i=0; i<gridy; i++)
      {
	if (pgrid[i][pmap]==me && pgrid[i][smap]!=me)
	  {
	    dbg_msg(0, ("\n\teRqS+:%d to:%d", MSG_SNREQ+sid, pgrid[i][smap]));
	    dbg_msg(0, ("\n\teSzR:%d fm:%d", MSG_SNDATA+sid, pgrid[i][smap]));
	    CommAsyncSend(&mone, pgrid[i][smap], MSG_SNREQ+sid, sizeof(int));
	    CommAsyncRecv(NULL, pgrid[i][smap], MSG_SNDATA+sid, -1);
	    c_vola += sizeof(int);
	  }
	if (pgrid[i][smap]==me && pgrid[i][pmap]!=me)
	  {
	    dbg_msg(0, ("\n\teRqR:%d fm:%d", MSG_SNREQ+sid, pgrid[i][pmap]));
	    CommAsyncRecv(NULL, pgrid[i][pmap], MSG_SNREQ+sid, sizeof(int));
	  }
      }
}





/* Routine to disable BASE block communications */
void DisableBASEcomm(ptr, pvt)
    snode_t *ptr;
    int pvt;
{
  int pmap, smap;
  int i, me;
  
  pmap = pvt % gridx;
  smap = ptr->sid % gridx;
  me = MyProc();

  dbg_msg(0, ("\nDisable(%d,%d)", ptr->sid, pvt));

  if (pmap != smap)
    for (i=0; i<gridy; i++)
      {
	if (pgrid[i][pmap]==me && pgrid[i][smap]!=me)
	  {
	    dbg_msg(0, ("\n\tdRqS-:%d to:%d", MSG_SNREQ+ptr->sid, pgrid[i][smap]));
	    CommAsyncSend(&ptr->sid, pgrid[i][smap], MSG_SNREQ+ptr->sid, sizeof(int));
	    c_vola += sizeof(int);
	  }
	if (pgrid[i][smap]==me && pgrid[i][pmap]!=me)
	  {
	    dbg_msg(0, ("\n\tdRqR:%d fm:%d", MSG_SNREQ+ptr->sid, pgrid[i][pmap]));
	    CommAsyncRecv(NULL, pgrid[i][pmap], MSG_SNREQ+ptr->sid, sizeof(int));
	  }
      }
}




