/*
 * complex.h
 *
 * Definitions for Complex Arithmetic Kernels 
 *
 * Author : Mustafa Uysal (uysal@cs.umd.edu)
 *
 */
#ifndef __COMPLEX_H__
#define __COMPLEX_H__


/* Complex variable initialization */
#define  CX_INIT(to,rp,ip)\
{\
  (to).RealPart = (fp_t) rp;\
  (to).ImagPart = (fp_t) ip;\
}

/* Complex variable assignment */    
#define  CX_ASSIGN(to,from)\
{\
   (to).RealPart = (from).RealPart;\
   (to).ImagPart = (from).ImagPart;\
}

/* Complex conjugate assignment */
#define  CX_CJASSIGN(to,from)\
{\
    (to).RealPart = (from).RealPart;\
    (to).ImagPart = -(from).ImagPart;\
}

/* Complex negate assignment */
#define  CX_NGASSIGN(to,from)\
{\
    (to).RealPart = -(from).RealPart;\
    (to).ImagPart = -(from).ImagPart;\
}

/* Complex conjugate negate assignment */
#define  CX_CJNGASSIGN(to,from)\
{\
    (to).RealPart = -(from).RealPart;\
    (to).ImagPart = (from).ImagPart;\
}

/* Complex conjugate */
#define  CX_CONJ(a)\
{\
   (a).ImagPart = -(a).ImagPart;\
}

/* Complex negation */
#define  CX_NEGATE(a)\
{\
    (a).RealPart = -(a).RealPart;\
    (a).ImagPart = -(a).ImagPart;\
}

/* Complex subtraction */
#define  CX_DEC(a,b)\
{\
    (a).RealPart -= (b).RealPart;\
    (a).ImagPart -= (b).ImagPart;\
}

/* Complex Multiply Subtract : [to -= a * b] */
#define CX_MULTSUB(to,a,b)\
{\
   (to).RealPart -= (a).RealPart * (b).RealPart - (a).ImagPart * (b).ImagPart;\
   (to).ImagPart -= (a).RealPart * (b).ImagPart + (a).ImagPart * (b).RealPart;\
}

/* Complex Multiply Add : [to += a * b] */
#define CX_MULTADD(to,a,b)\
{\
   (to).RealPart += (a).RealPart * (b).RealPart - (a).ImagPart * (b).ImagPart;\
   (to).ImagPart += (a).RealPart * (b).ImagPart + (a).ImagPart * (b).RealPart;\
}

/* Complex Multiplication with a Factor : [to *= with] */
#define CX_MULTFCT(to,with)\
{\
   fp_t tmp_ = (to).RealPart;\
   (to).RealPart = tmp_ * (with).RealPart - (to).ImagPart * (with).ImagPart;\
   (to).ImagPart = tmp_ * (with).ImagPart + (to).ImagPart * (with).RealPart;\
}

/* Complex Division : [to = n / d] */
#define CX_DIV(to,n,d)\
{\
   fp_t  r_, s_;\
   if (((d).RealPart >= (d).ImagPart && (d).RealPart >  -(d).ImagPart)||\
       ((d).RealPart  < (d).ImagPart && (d).RealPart <= -(d).ImagPart))\
   {\
      r_ = (d).ImagPart / (d).RealPart;\
      s_ = (d).RealPart + r_*(d).ImagPart;\
      (to).RealPart = ((n).RealPart + r_*(n).ImagPart)/s_;\
      (to).ImagPart = ((n).ImagPart - r_*(n).RealPart)/s_;\
   }\
   else\
   {\
      r_ = (d).RealPart / (d).ImagPart;\
      s_ = (d).ImagPart + r_*(d).RealPart;\
      (to).RealPart = (r_*(n).RealPart + (n).ImagPart)/s_;\
      (to).ImagPart = (r_*(n).ImagPart - (n).RealPart)/s_;\
   }\
}

/* Complex SquareRoot */
#ifdef DOUBLE
#define CX_SQRT(x)    zsqrtf(&(x))  /* Call to Fortran Library */
#else
#define CX_SQRT(x)    csqrtf(&(x))  /* Call to Fortran Library */
#endif

/* Complex Comparison (Less than or equal to) : [a <= b] */
#define CX_LEQ(a,b)\
((((a).RealPart <= (b).RealPart) && ((a).ImagPart <= (b).ImagPart)) ? 1 : 0)

#endif __COMPLEX_H__

