/*
 * disk.c
 *
 * $Source: /fs/nene/uysal/pf/newFact/RCS/disk.c,v $
 *
 * $Header: /fs/nene/uysal/pf/newFact/RCS/disk.c,v 1.1 1995/10/16 03:42:25 uysal Exp $
 *
 * Module to coordinate the accesses to the disk
 *
 * Applicable Compiler Directives :
 * DUMP_METADATA   The metadata of the matrix is dumped in a human
 *                 readble form to stdout
 *
 *
 * Author : Mustafa Uysal (uysal@cs.umd.edu)
 *
 */
#include <aio.h>
#include <fcntl.h>
#include <unistd.h>

#include "world.h"
#include "block.h"
#include "disk.h"
#include "snode.h"
#include "mppworld.h"
#include "queue.h"
#include "io.h"

#include "debug.h"

/*--------------- Local Data Types ---------------*/

typedef struct {
  aio_t   *handle;
  snode_t *dataptr;
} disk_wrt_t;



/*--------------- Static Global Variables ---------------*/

static int dataFile;    /* dataFile handler */
static int swritq;      /* queue of SNs scheduled to be written back to disk */


/*--------------- Externally Defined Globals ---------------*/

extern int errno;

extern double bytes_read;
extern double read_time;
extern double iowait_time;

extern double bytes_written;
extern double write_time;


/*--------------- Local Function Prototypes ---------------*/

/* routine to setup local Value Pointers to the correct location */
void SetUpValuePtrs ARGS((snode_t *sptr));

/* routine to attach to strings together */
char *stratt ARGS((char *s1, char *s2));

/*--------------- Module Implementation ---------------*/

/*
 * Routine to read the configuration file
 */
void DiskInit(fn, ns, ofptr, szptr)
    char *fn;
    int  *ns;
    int  **ofptr, **szptr;
{
  int confFile;
  char *str;
  int *ofset, *size, cnt;
  
  str = stratt(fn, ".blk");
  dataFile = open(str, O_RDWR);
  Free(str);

  str = stratt(fn, ".atr");
  confFile = open(str, O_RDONLY);
  Free(str);

  exit_if((dataFile<0)||(confFile<0),
	  ("Error in file opens [%d %d] [%s]", dataFile, confFile, fn));

  if (read(confFile, &cnt, sizeof(int)) != sizeof(int))
    panic(("Error in reading configuration file [1]."));

  ofset = (int*) Calloc(cnt, sizeof(int));
  size  = (int*) Calloc(cnt, sizeof(int));
  
  if (read(confFile, ofset, cnt*sizeof(int)) != cnt*sizeof(int))
    panic(("Error in reading configuration file [2]."));

  if (read(confFile, size, cnt*sizeof(int)) != cnt*sizeof(int))
    panic(("Error in reading configuration file [3]."));
  close(confFile);

  *ns = cnt;
  *ofptr = ofset;
  *szptr = size;

  swritq = QueueNew();
  print_if(!MyProc(), ("\n%d Supernodes", cnt));

#if defined(DUMP_METADATA)
  print_if(!MyProc(), ("\nMetadata will be printed"));
#else
  print_if(!MyProc(), ("\nMetadata will NOT be printed"));
#endif
}




/* Housekeeping and cleanup routine for disk module */
void DiskExit()
{
  disk_wrt_t *rec;

  /* Discard the completed write requests */
  while (!QueueEmpty(swritq))
    {
      rec = (disk_wrt_t*) QueueDelete(swritq);
      while (AIOstatus(rec->handle) != AIO_DONE)
	;
      
      rec->dataptr->status = SN_INUSE;
      SnodeFree(rec->dataptr);
      Free(rec->handle);
      Free(rec);
    }
  QueueDestroy(swritq);
}





#if    defined(BLOCKING_READ)

/* Blocking read of local supernode data */
snode_t *DiskReadLocal(sid, pvt, mode)
    int sid;
    int pvt;
    int mode;
{
  block_t *bptr;
  snode_t *sptr;
  int size, offset, j, cnt;
  double ts, te;
  
  /* Find out where this supernode is */
  offset = SnodeOffset(sid);
  size   = SnodeSize(sid);
  exit_if(size<0, ("Size adjustment problem at SN#%d [%d]", sid, size));

  /* Space allocation and initialization */
  sptr = (snode_t*) SnodeAlloc(sid, offset, size);

  /* blocking read */
  ts = CLOCK();
  lseek(dataFile, offset, SEEK_SET);
  assert(read(dataFile, sptr->buffer, size) == size);
  te = CLOCK();

  /* statistics */
  read_time += (te - ts);
  bytes_read += ((double) size);
  
  /* setup block descriptors so that the value pointers are right */
  SetUpValuePtrs(sptr, pvt, mode);

#if    defined(DUMP_METADATA)
  if (mode != DSK_NEXT)
    {
      for (j=0; j<sptr->blkcnt; j++)
	{
	  bptr = &(sptr->blkPtr[j]);
	  message(("\tBlk[%d of %d] is [%d %d @ %d %d %d %d]", j, pvt,
		   bptr->ycoor, bptr->xcoor, bptr->lcol, bptr->rcol,
		   bptr->urow, bptr->lrow));
	}
    }
#endif

  
  return sptr;
}



/* Blocking write of local supernode data */
void DiskWriteLocal(ptr)
    snode_t *ptr;
{
  int size, offset;
  double ts, te;
  
  /* dooze off if its not my business */
  if (!ptr->blkcnt) return;
  
  /* Find out where this supernode is */
  offset = SnodeOffset(ptr->sid);
  size   = SnodeSize(ptr->sid);

  /* blocking write */
  ts = CLOCK();
  lseek(dataFile, offset, SEEK_SET);
  assert(write(dataFile, ptr->blkPtr[0].value, size) == size);
  te = CLOCK();

  /* statistics */
  write_time += (te - ts);
  bytes_written += ((double) size);
  
  /* Adjust the offset so that diagonal won't be read again */
  if (ptr->blkPtr[0].ycoor == ptr->blkPtr[0].xcoor)
    SnodeLocatorAdjust(ptr->sid, BlockSizeByte(&(ptr->blkPtr[0])));
}



#elif defined(ASYNC)

/* Asynchronous reading of supernode data */
snode_t *DiskReadLocal(sid, pvt, mode, slabs, adjPtr)
    int sid;
    int pvt;
    int mode;
    int slabs;
    int *adjPtr;
{
  block_t *bptr;
  snode_t *sptr;
  int size, offset, j, cnt;
  aio_t *handle;
  double ts, te;

  *adjPtr = 0;
  
  /* Supernode request, if it exists in the memory */
  if ((sptr = SnodeRequest(sid)) == NULL)
    {
      /* On Disk, do a blocking Read */

      /* Find out where this supernode is */
      offset = SnodeOffset(sid);
      size   = SnodeSize(sid);

      /* Space allocation and initialization */
      sptr = SnodeAlloc(sid, offset, size);
      exit_if(!sptr, ("DiskReadLocal(%d,%d,%d) failed, not enough memory",
		      sid, pvt, mode));

      /* blocking read */
      TimedSeek(dataFile, offset, SEEK_SET, &read_time);
      TimedRead(dataFile, sptr->buffer, size, &read_time, &bytes_read); 
      printf("bread(%d)\n", size);

      /* setup block descriptors so that the value pointers are right */
      SetUpValuePtrs(sptr, pvt, mode);
      *adjPtr = SnodeAdjustSize(sptr, pvt, slabs);

      goto out;
    }

  /* Wait for the READ to complete */
  if (sptr->status == SN_READING)
    {
      bytes_read += (double) sptr->size;
      ts = CLOCK();
      handle = (aio_t *) sptr->blkPtr;
      AIOwait(handle, IO_READ);
      Free(handle);
      te = CLOCK();
      iowait_time += (te - ts);

      /* setup block descriptors so that the value pointers are right */
      SetUpValuePtrs(sptr, pvt, mode);
      *adjPtr = SnodeAdjustSize(sptr, pvt, slabs);

      goto out;
    }

 out:  
#if defined(DUMP_METADATA)
  if (mode == DSK_FIRST) SnodeStructPrint(sptr);
#endif

  if (sptr->status != SN_LOCKWT)
    sptr->status = SN_INUSE;
  return sptr;
}




/* Asynchronous prefetch of a supernode data */
int PrefetchSN(id, pvt)
    int id, pvt;
{
  int size, offset, bcnt;
  snode_t *sptr;
  aio_t *handle;
  caddr_t adr;
  
  /* make communication system aware of the data transfer */
  NetPrefetchBase(id, pvt);

  /* Find out where this supernode is */
  sptr = SnodeRequest(id);

  /* Not in Memory */
  if (!sptr)
    {
      /* try allocating mem-space */
      offset = SnodeOffset(id);
      size   = SnodeSize(id);
      sptr   = SnodeAlloc(id, offset, size);
      if (!sptr) return 0;
      
      sptr->status = SN_READING;
      printf("AIOread(%d)\n", size);
      handle = IssueAIOread(dataFile, sptr->buffer, size, offset, SEEK_SET);
      assert(!sptr->blkPtr);
      sptr->blkPtr = (block_t*) handle; /* Save handle here */
    }
  else if (sptr->status == SN_WRITING)
    sptr->status = SN_LOCKWT;
  else if (sptr->status == SN_INUSE || sptr->status == SN_READING)
    ; /* I know about these cases */
  else 
    message(("\n\tCase not handled [id:%d, pv:%d, st:%d]", id, pvt, sptr->status));

  return 1;
}





/* Asynchronous write for local supernode data */
void DiskWriteLocal(ptr)
    snode_t *ptr;
{
  int size, offset;
  disk_wrt_t *rec;

  /* dooze off if its not my business */
  if (!ptr->blkcnt)
    return;

  /* Update status */
  ptr->status = SN_WRITING;

  /* Find out where this supernode is */
  offset = SnodeOffset(ptr->sid);
  size   = SnodeSize(ptr->sid);

  /* async  write */
  rec = (disk_wrt_t *) Malloc(sizeof(disk_wrt_t));
  rec->handle = IssueAIOwrite(dataFile, ptr->blkPtr[0].value, size, offset, SEEK_SET);
  rec->dataptr = ptr;
  QueueAdd(swritq, rec);
  bytes_written += (double)size;
}



/* Routine to discard the completed write requests */
void  WriteQueueDrain()
{
  disk_wrt_t *rec;
  
  while (!QueueEmpty(swritq))
    {
      rec = (disk_wrt_t*) QueueHead(swritq);
      if (AIOstatus(rec->handle) == AIO_DONE)
	{
	  rec = (disk_wrt_t*) QueueDelete(swritq);
	  if (rec->dataptr->status == SN_WRITING) 
	    SnodeFree(rec->dataptr);
	  else rec->dataptr->status = SN_INUSE;
	  Free(rec->handle);
	  Free(rec);
	}
    }
}


#else
#error "One of BLOCKING  or ASYNC should be specified"
#endif





/* Routine to fix [*value] fields of block descriptors, after reading the
 * supernode data from disk */
void SetUpValuePtrs(sptr, sid, mode)
    snode_t *sptr;
    int sid;
    int mode;
{
  caddr_t address;
  block_t *blkptr;
  int blkcnt, depcnt;
  int i, id, size;

  /* get # of blocks */
  dbg_if(TRUE, { size = sptr->size; id = sptr->sid; });

  address = sptr->buffer + sptr->size - sizeof(int);
  blkcnt = *((int*)address);

  /* get the address of block descriptors */
  address -= (blkcnt * sizeof(block_t));
  sptr->blkPtr = (block_t*) address;
  sptr->blkcnt = blkcnt;
  
  /* skip non-updating block descriptors */
  for (i=0; i<blkcnt && sptr->blkPtr[i].ycoor < sid; i++) 
    ;

  if (mode == DSK_FIRST)
    {
      depcnt = *((int*)sptr->buffer);
      address = sptr->buffer + (depcnt+1)*sizeof(int); 
      SnodeLocatorAdjust(sptr->sid, (depcnt+1)*sizeof(int));
    }
  else
    {
      address = sptr->buffer;
    }
  
  /* set-up value pointers of remaining block descriptors */
  dbg_if(TRUE, size=0);
  for (; i<blkcnt; i++)
    {
      sptr->blkPtr[i].value = (nonzero_t*) address;
      address += BlockSizeByte(&sptr->blkPtr[i]);
      dbg_if(TRUE, size += BlockSizeByte(&sptr->blkPtr[i]));
    }

  dbg_if(mode==DSK_NEXT&&(size+blkcnt*sizeof(block_t)+sizeof(int))!=sptr->size,
	 {
	   SnodeStructPrint(sptr);
	   panic(("\nWrong, sz:%d bc:%d[%d] os:%d id:%d to:%d", size, blkcnt, 
		  blkcnt*sizeof(block_t), sptr->size, sptr->sid, sid));
	 });
}




/*
 * STRing ATTach for attaching two strings together and making a 
 * new string 
 */
char *stratt(s1, s2)
     char *s1, *s2;
{
  char *new;

  new = (char *) Malloc(strlen(s1)+strlen(s2)+1);

  strcpy(new, s1);
  strcat(new, s2);

  return(new);
}




