/*
 * factor.c
 *
 * $Source: /fs/nene/uysal/pf/newFact/RCS/factor.c,v $
 *
 * $Header: /fs/nene/uysal/pf/newFact/RCS/factor.c,v 1.2 1995/10/16 04:24:45 uysal Exp $
 *
 * High level algorithm and driver routines for parallel
 * out of core sparse factorization. Supernodal Fan-in style of
 * computations.
 *
 * Author : Mustafa Uysal (uysal@cs.umd.edu)
 *
 */
#include <stdlib.h>

#include "world.h"
#include "mppworld.h"
#include "queue.h"
#include "kernel.h"
#include "snode.h"
#include "block.h"
#include "disk.h"
#include "comm.h"

#include "debug.h"


/*--------------- Statistics Variables ---------------*/
/* IO performance */
double bytes_read, read_time, iowait_time;
double bytes_written, write_time;
double io_t, io_w, io_r;

/* kernel performance */
double k_df, k_du, k_bm, k_dm, k_dr;
double p_up, p_fa;

/* communication performance */
double c_mul, c_dia, c_tot;
double c_mula, c_mulb;
long   c_vola, c_volb;

/* High level performance */
double etime;                     /* execution time */
double ovhd;

/* temp stuff */
/*int p_cnt = 0, p_zro = 0, p_mlt = 0, p_tot=0;*/

int s_lim = 0;

/*--------------- Function Prototypes ---------------*/

/* high level out-of-core sparse parallel Cholesky factorization */
void OutOfCoreFactor ARGS(( void ));

/* routine to update one supernode with another */
void PanelUpdate ARGS((block_t *mult, int cnt, snode_t *src, snode_t *dst));

/* routine to factor blocks of the pivot supernode */
void PanelFactor ARGS((block_t *diag, snode_t *pivot));

/* routine to report the performance results */
void ReportResults ARGS((void));

/* Routine to initialize all timers */
void TimerInit ARGS((void));

int atoi();

/*--------------- Implementation ---------------*/

main(argc, argv)
    int   argc;
    char *argv[];
{
  int snCount;     /* Supernode count */
  int pLevel;      /* # of parallel supernodes under update */
  int *pSlots;     /* Mask array of size pLevel for parallel slots */
  int *snOfset, *snSize;
  
  /* argument checking */
  if (argc<2) panic(("Usage : pf matrix"));
  if (argc>2) s_lim = atoi(argv[2]);

  if (!MyProc())
    {
      message(("Parallel Out-of core Sparse Factorization\n"));
      message(("Input File : %s", argv[1]));
      print_if(s_lim>0, ("\n** EXECUTE upto %d SUPERNODES **", s_lim));
    }

  /* System initialization (for each module) */
  MPPinit();
  MemInit(0);
  DiskInit(argv[1], &snCount, &snOfset, &snSize);
  SnodeInit(snCount, snOfset, snSize);
  pSlots = CommInit(&pLevel);
  TimerInit();
  KernelInit(!MyProc());

  /* Flush out I/O buffers on nodes */
  Sync();  Sync();  Sync();
  fflush(stdout);
  Sync();  Sync();  Sync();

  if (s_lim>0) snCount = s_lim;

  /* Do some useful number crunching */
  etime = CLOCK();
  OutOfCoreFactor(snCount, pSlots, pLevel);
  etime = CLOCK() - etime;
  
  /* Housekeeping */
  CommExit();
  DiskExit();
  SnodeExit();
  MemExit();

  /* Performance Report */
  ReportResults();
}





/*
 * High level algorithm for the out-of-core factorization.
 */
void OutOfCoreFactor(snCount, pMask, pLevel)
    int snCount, *pMask, pLevel;
{
  int snStrip;        /* Current strip of supernodes being updated in parallel */
  int snNext;
  int snCurrent;      /* Current supernode to be factored next */
  int count, *list;   /* List of supernode numbers that will update the current */
  int mcnt;           /* # of multiplier blocks */
  block_t *mult;      /* Array of multiplier blocks */
  snode_t *pivot;     /* pivot supernode being factored */
  snode_t *source;    /* source supernode used in the update */
  snode_t **parr;     /* Array of pivots of the current slab */
  snode_t **sarr;     /* Array of sources of the current slab */
  int *carr, **larr;  /* Array of lists and counts for the current slab */
  int *farr;          /* Array of flags for intra-slab dependences */
  double t0, t1;
  int sid, size_adjust;
  int snFreeFlag;
  int i, j, k;

  parr = (snode_t**) Calloc(pLevel, sizeof(snode_t*));
  sarr = (snode_t**) Calloc(pLevel, sizeof(snode_t*));
  carr = (int*) Calloc(pLevel, sizeof(int));
  farr = (int*) Calloc(pLevel, sizeof(int));
  larr = (int**) Calloc(pLevel, sizeof(int*));
  list = (int*) MemGet(snCount*sizeof(int));
  for (i=0; i<snCount; i++)
    list[i] = 0;

  /* initialize */
  for (i=0; i<pLevel; i++)
    farr[i] = 0;

  /* Set out prefetches for the current blocks */
  for (i=0; i<MIN(snCount, pLevel); i++)
    PrefetchSN(i,i);

  for (snStrip=0; snStrip<snCount; snStrip+=pLevel)
    {
      print_if(!MyProc() && !(snStrip%100), 
	       ("\n[%5d wc:%4.3lf secs cp:%4.3lf secs io:%4.3lf secs co:%4.3lf secs]", 
		snStrip, CLOCK()-etime, p_up+p_fa, io_r+io_w, c_mul+c_dia));
      snNext = MIN(snStrip+pLevel, snCount);
      pLevel = snNext - snStrip;
      
      /* Each node prefetches the data in the current stripe */
      t0 = CLOCK();
      for (count=0, i=snStrip; i<snNext; i++)
	{
	  j = i - snStrip;
	  parr[j] = DiskReadLocal(i, i, DSK_FIRST, pLevel, &k);
	  larr[j] = SnodeUpdaters(parr[j], &carr[j]);
	  if (carr[j]) PrefetchSN(larr[j][0], i);
	  count = MAX(carr[j], count);
	}
      t1 = CLOCK();
      io_r += (t1-t0);

      /* Parallel Pivot Update Loop */
      for (i=0; i<count; i++)
	{
	  /* First get data and start shipments */
	  for (j=0; j<pLevel; j++)
	    {
	      if (i<carr[j] && larr[j][i]<snStrip)
		{
		  /* Read in the updaters data */
		  t0 = CLOCK();
		  sarr[j] = DiskReadLocal(larr[j][i], snStrip, DSK_NEXT, pLevel, &k);
		  if (!list[sarr[j]->sid])  list[sarr[j]->sid] = k;
		  t1 = CLOCK();
		  io_r += (t1-t0);
		  
		  CommBlockSend(snStrip+j, sarr[j], MyProc());
		}

	      /* Enable disk prefetch for next round */
	      if ((i+1)<carr[j] && larr[j][i+1]<snStrip)
		{
		  t0 = CLOCK();
		  PrefetchSN(larr[j][i+1], j+snStrip);
		  t1 = CLOCK();
		  io_r += (t1-t0);
		}
	    }

	  /* Perform updates for this round */
	  for (j=0; j<pLevel; j++)
	    if (i<carr[j])
	      {
		if (larr[j][i]<snStrip)
		  {
		    /* Receive off-processor blocks necessary for the updates */
		    t1 = CLOCK();
		    mult = CommBlockReceive(snStrip+j, sarr[j], MyProc(), &mcnt, snStrip);
		    t0 = CLOCK();
		    c_mul += (t0-t1);
		    
		    /* Perform the supernodal update */
		    PanelUpdate(mult, mcnt, sarr[j], parr[j]);
		    t1 = CLOCK();
		    p_up += (t1-t0);
		  }
		else
		  farr[j] = 1;
	      }

	  /* Space reclaimer of the past round */
	  SendQueueFlush();

	  t0 = CLOCK();
	  for (j=0; j<pLevel; j++)
	    if (i<carr[j] && larr[j][i]<snStrip)
	      {
		/* check if this SN is in the prefetch pipeline */
		sid = larr[j][i];
		snFreeFlag = 1;
		for (k=0; k<pLevel; k++)
		  if ((i+1)<carr[k])
		    if (larr[k][i+1] == sid)
		      {
			snFreeFlag = 0;
			break;
		      }
		if (snFreeFlag)
		  {
		    source = SnodeRequest(larr[j][i]);
		    SnodeFree(source);
		  }
	      }
	  ovhd += (CLOCK() - t0);
	}
      
      /* Prefetch for the next round */
      t0 = CLOCK();
      for (j=snNext; j<MIN(snCount, snNext+pLevel); j++)
	PrefetchSN(j,j);
      io_r += (CLOCK() - t0);

      /* Factorize current slab */
      for (j=0; j<pLevel; j++)
	{
	  /* Intra-slab dependence exists */
	  for (i=carr[j]-1; farr[j] && i>=0 && larr[j][i] >= snStrip; i--)
	    {
	      t0 = CLOCK();
	      NetPrefetchBase(larr[j][i], snStrip+j);
	      source = parr[larr[j][i] - snStrip];
	      CommBlockSend(snStrip+j, source, MyProc());
	      t1 = CLOCK();
	      ovhd += (t1-t0);

	      /* Receive off-processor blocks necessary for the updates */
	      mult = CommBlockReceive(snStrip+j, source, MyProc(), &mcnt, snStrip);
	      t0 = CLOCK();
	      c_mul += (t0-t1);

	      /* Perform the supernodal update */
	      PanelUpdate(mult, mcnt, source, parr[j]);
	      t1 = CLOCK();
	      p_up += (t1-t0);
	    }

	  farr[j] = 0;
	  if (pMask[j])
	    {
	      /* Receive the diagonal for local factorization */
	      t0 = CLOCK();
	      mult = CommDiagRecv(parr[j], MyProc());
	      t1 = CLOCK();
	      c_dia += (t1-t0);
	      
	      /* Factorize my local portions */
	      PanelFactor(mult, parr[j]);
	      t0 = CLOCK();
	      p_fa += (t0-t1);
	    }
	  
	  t0 = CLOCK();
	  DiskWriteLocal(parr[j]);
	  SnodeLocatorAdjust(j+snStrip, SnodeAdjustSize(parr[j], snStrip, pLevel));
	  t1 = CLOCK();
	  io_w += (t1-t0);
	}

      /* Reclaim memory of old pivots */
      SendQueueFlush();
      t0 = CLOCK();
      WriteQueueDrain();
      t1 = CLOCK();
      io_w += (t1-t0);

      /* Adjust supernode offset & sizes for the next strip */
      for (j=0; j<pLevel; j++)
	for (k=0; k<carr[j]; k++)
	  {
	    sid = larr[j][k];
	    if (list[sid])
	      {
		SnodeLocatorAdjust(sid, list[sid]);
		list[sid] = 0;
	      }
	  }
      ovhd += (CLOCK() - t1);
    }
  
  MemPut(list, snCount*sizeof(int));
}






/* Routine to factorize the current pivot */
void PanelFactor(diag, pivot)
    block_t *diag;
    snode_t *pivot;
{
  double t0,t1;
  block_t *rect;
  int i;

  if (!pivot->blkcnt) return;

  for (i=((diag==&(pivot->blkPtr[0])) ? 1 : 0); i<pivot->blkcnt; i++)
    {
      rect = &(pivot->blkPtr[i]);
      dbg_if(rect->xcoor>=rect->ycoor, 
	     panic(("error x=%d y=%d pvt=%d", rect->xcoor, rect->ycoor, pivot->sid)));
      t0 = CLOCK();
      drmod(rect, diag);
      t1 = CLOCK();
      k_dr += (t1-t0);
    }
}





/* Routine to perform the supernodal update of the pivot */
void PanelUpdate(mult, cnt, src, pvt)
    block_t *mult;
    int cnt;
    snode_t *src, *pvt;
{
  double t0,t1;
  block_t *diag;
  int i,j,k,ndx; 

  /* abort if there is nothing to update */
  if (!pvt->blkcnt || !mult) return;

  /* find out the index of the 1st base block in src */
  for (ndx=0; ndx<src->blkcnt; ndx++)
    if (src->blkPtr[ndx].ycoor > pvt->sid) break;

  /* branch out for alternative update sequences : one with digonal and one without */
  diag = &(pvt->blkPtr[0]);
  if (diag->xcoor == diag->ycoor)
    {
      for (i=0; i<cnt; i++)
        {
          /* Perform the modifications to diagonal */
	  t0 = CLOCK();
          dgupdt(diag, &mult[i]);
	  t1 = CLOCK();
	  k_du += (t1-t0);
          for (j=i+1; j<cnt; j++)
            dmod(diag, &mult[j], &mult[i]);
	  t0 = CLOCK();
	  k_dm += (t0-t1);
	  
	  for (j=ndx; j<src->blkcnt; j++)
            {
              /* match up the destination index set (which is sorted) */
              for (k=1; (k<pvt->blkcnt)&&(pvt->blkPtr[k].ycoor<src->blkPtr[j].ycoor); k++)
                ;

              /* if a match is found, fire off bmod() operations */
	      t0 = CLOCK();
              for ( ; (k<pvt->blkcnt)&&(pvt->blkPtr[k].ycoor==src->blkPtr[j].ycoor); k++)
		bmod(&(pvt->blkPtr[k]), &(src->blkPtr[j]), &mult[i]);
	      t1 = CLOCK();
	      k_bm += (t1-t0);
            }
	}
    }
  else
    {
      for (i=0; i<cnt; i++)
	for (j=ndx; j<src->blkcnt; j++)
	  {
	    /* match up the destination index set (which is sorted) */
	    for (k=0; (k<pvt->blkcnt)&&(pvt->blkPtr[k].ycoor<src->blkPtr[j].ycoor); k++)
	      ;
	    
	    /* if a match is found, fire off bmod() operations */
	    t0 = CLOCK();
	    for ( ; (k<pvt->blkcnt)&&(pvt->blkPtr[k].ycoor==src->blkPtr[j].ycoor); k++)
	      bmod(&(pvt->blkPtr[k]), &(src->blkPtr[j]), &mult[i]);
	    t1 = CLOCK();
	    k_bm += (t1-t0);
	  }
    }
}






/* Routine to report all performance results */
void ReportResults()
{
  double t_kernel, t_panel, t_diff;
  double t_comm;

  io_t = io_w + io_r;
  t_kernel = k_df + k_du + k_bm + k_dm + k_dr;
  t_panel  = p_fa + p_up;
  t_comm   = c_mul + c_dia;
  t_diff   = etime - (t_panel + io_t + ovhd + t_comm);

  message(("\nTOT exec:[%.3lf secs] panl:[%.3lf secs]", etime, t_panel));
  message(("\nTOT comm:[%.3lf secs] io:[%.3lf secs]", t_comm, io_t));
  message(("\nTOT krnl:[%.3lf secs] ovhd:[%.3lf secs] diff:[%.3lf secs]", t_kernel, ovhd, t_diff));
  message(("\nPNL pfct:[%.3lf secs] pupt:[%.3lf secs]", p_fa, p_up));
  message(("\nKER dfct:[%.3lf secs] dupt:[%.3lf secs]", k_df, k_du));
  message(("\nKER bmod:[%.3lf secs] dmod:[%.3lf secs] drmd:[%.3lf secs]", k_bm, k_dm, k_dr));
  message(("\nIO  read:[%.3lf secs] writ:[%.3lf secs]", io_r, io_w));
  message(("\nCOM bmul:[%.3lf secs] diag:[%.3lf secs]", c_mul, c_dia));
  message(("\nCOM ma-b:[%.3lf secs] bmlb:[%.3lf secs]", c_mula, c_mulb));
  message(("\nCOM vola:[%d Bytes] volb:[%d Bytes]", c_vola, c_volb));
  message(("\nIO  rsiz:[%.2lf MB] wsiz:[%.2lf MB]", (double)bytes_read/(double)MEGABYTE,
	   (double)bytes_written/(double)MEGABYTE));
  message(("\nIO  Wait:[%.3lf secs] Read:[%.3lf secs] Size:[%.3lf MB] Rate:[%.3lf MB/secs]",
	   iowait_time, read_time, (bytes_read/MEGABYTE), 
	   (bytes_read+iowait_time)/(MEGABYTE*read_time)));

  /*
  message(("IO Writ:[%.3lf secs] Size:[%.3lf MB] Rate:[%.3lf MB/secs]",
           write_time, (bytes_written/MEGABYTE), bytes_written/(MEGABYTE*write_time))); 
  */


  MPPreport();
  IOreport();

  Sync();  Sync(); Sync();
  printf("\n");
}




/* Routine to initialize all timers */
void TimerInit()
{
  /* I/O performance */
  iowait_time = bytes_read = read_time = bytes_written = write_time = io_t = 0.0;

  /* kernel performance */
  k_df = k_du = k_bm = k_dm = k_dr = p_up = p_fa = 0.0;

  /* communication performance */
  c_mula = c_mulb = c_dia = c_mul = c_tot = 0.0;
  c_vola = c_volb = 0;
  
  /* High level performance */
  ovhd = etime = 0.0;
}

