/*
 * io.c
 *
 * $Source: /fs/nene/uysal/pf/newFact/RCS/io.c,v $
 *
 * $Header: /fs/nene/uysal/pf/newFact/RCS/io.c,v 1.1 1995/10/16 03:42:25 uysal Exp $
 *
 * Disk I/O module for the parallel block partitioner. This module
 * is responsible for providing support of I/O operations to
 * various components of the block partitioner.
 *
 * Author : Mustafa Uysal <uysal@cs.umd.edu>
 *
 */
#include <stdio.h>
#include <fcntl.h>
#include <malloc.h>

#include "world.h"
#include "queue.h"
#include "io.h"

#define LARGE_VOLUME  (1000*MEGABYTE)

extern int errno;

static queue_t *preadq;  /* queue of pending async read requests */
static queue_t *pwritq;  /* queue of pending async write requests */
static queue_t *outioq;  /* queue of outstanding AIO requests */

static long  r_max, r_min, r_tot, r_cnt;
static long  w_max, w_min, w_tot, w_cnt;
static long  s_cnt;

static double r_tim, w_tim, s_tim, a_tim; 



/*--------------- Implementation ---------------*/					
	

void IOinit()
{
  preadq = QueueNew();
  pwritq = QueueNew();
  outioq = QueueNew();
  
  r_max = r_tot = r_cnt = 0;
  w_max = w_tot = w_cnt = 0;
  r_tim = w_tim = s_tim = a_tim = 0.0;
  r_min = w_min = LARGE_VOLUME;
}




void IOexit()
{
  QueueDestroy(preadq);
  QueueDestroy(pwritq);
  QueueDestroy(outioq);
}




void IOreport()
{
  message(("\nIO-READ-SIZE max:[%7ld B] min:[%7ld B] avg:[%7ld B] tot:[%10ld MB]",
	   r_max, r_min, r_tot/r_cnt, r_tot/MEGABYTE));
  message(("\nIO-WRIT-SIZE max:[%7ld B] min:[%7ld B] avg:[%7ld B] tot:[%10ld MB]",
	   w_max, w_min, w_tot/w_cnt, w_tot/MEGABYTE));
  message(("\nIO-TIME read:[%.3lf secs] writ:[%.3lf secs]", r_tim, w_tim));
  message(("\nIO-TIME asyn:[%.3lf secs] seek:[%.3lf secs]", a_tim, s_tim));
  message(("\nIO-RATE read:[%.3lf MB/s] writ:[%.3lf MB/s]", 
	   (double)r_tot/(r_tim*(double)MEGABYTE), 
	   (double)w_tot/(w_tim*(double)MEGABYTE)));
}


										

/* routine to yield the time taken by a read operation */
int TimedRead(fd, buf, size, timer, ctr)
    int fd;
    char *buf;
    int size;
    double *timer, *ctr;
{
  double t0, t1;
  int nb;
  
  t0 = CLOCK();
  nb = read(fd,buf,size);
  t1 = CLOCK();

  exit_if(nb<0||nb!=size, 
	  ("File Read Error for size:%d nb:%d en:%d", size, nb, errno));

  *timer += (t1 - t0);
  *ctr   += (double) nb;
  
  r_max = MAX(nb, r_max);
  r_min = MIN(nb, r_min);
  r_cnt++;
  r_tim += (t1-t0);
  r_tot += nb;

  return nb;
}



/* routine to yield the time taken by a seek operation */
void TimedSeek(fd, offset, whence, timer)
    int fd;
    int offset;
    int whence;
    double *timer;
{
  double t0, t1;

  t0 = CLOCK();
  lseek(fd, offset, whence);
  t1 = CLOCK();

  *timer += (t1 - t0);
  s_tim  += (t1-t0);
  s_cnt++;
}





/* routine to yield the time taken by a write operation */
void TimedWrite(fd, buf, size, timer, ctr)
    int fd;
    char *buf;
    int size;
    double *timer, *ctr;
{
  double t0, t1;

  t0 = CLOCK();
  assert(write(fd,buf,size)==size);
  t1 = CLOCK();

  *timer += (t1 - t0);
  *ctr   += (double) size;

  w_max = MAX(w_max, size);
  w_min = MIN(w_min, size);
  w_tim += (t1-t0);
  w_tot += size;
  w_cnt++;
}





/* routine to issue an async write request */
aio_t *IssueAIOwrite(fd, buf, size, offset, whence)
    int fd;
    char *buf;
    int size, offset, whence;
{
  double t0, t1;
  aio_t *new;

  t0 = CLOCK();
  new = (aio_t*) Malloc(sizeof(aio_t));

  /* fill in OS control info */
  new->handle.aio_whence = whence;
  new->handle.aio_offset = offset;
  new->handle.aio_buf    = buf;
  new->handle.aio_nbytes = size;
  new->handle.aio_flag   = 0;

  /* put the request into OS queue */
  if (aio_write(fd, &(new->handle))<0)
    panic(("aio_write(%d,%d,%d) failed with error:%d", size, offset, whence, errno));
  t1 = CLOCK();

  w_max = MAX(w_max, size);
  w_min = MIN(w_min, size);
  w_tim += (t1-t0);
  w_tot += size;
  w_cnt++;

  return(new);
}




/* routine to issue an async read request */
aio_t *IssueAIOread(fd, buf, size, offset, whence)
    int fd;
    char *buf;
    int size, offset, whence;
{
  double t0, t1;
  aio_t *new;

  t0 = CLOCK();
  new = (aio_t*) Malloc(sizeof(aio_t));

  /* fill in OS control info */
  new->handle.aio_whence = whence;
  new->handle.aio_offset = offset;
  new->handle.aio_buf    = buf;
  new->handle.aio_nbytes = size;
  new->handle.aio_flag   = 0;

  /* put the request into OS queue */
  if (aio_read(fd, &(new->handle))<0)
    panic(("aio_read(%d,%d,%d) failed with error:%d", size, offset, whence, errno));
  t1 = CLOCK();

  r_max = MAX(size, r_max);
  r_min = MIN(size, r_min);
  r_cnt++;
  r_tim += (t1-t0);
  r_tot += size;

  return(new);
}




/* routine to check the status of an asynchronous I/O operation */
int AIOstatus(req)
    aio_t *req;
{
  int ret;

  ret = aio_error(req->handle.aio_handle);
  if (!ret) return AIO_DONE;

  return AIO_CONT;
}





int AIOwait(req, mode)
    aio_t *req;
    int mode;  
{
  double t0, t1;

  t0 = CLOCK();
  while (AIOstatus(req) != AIO_DONE)
    ;
  t1 = CLOCK();

  a_tim += (t1 - t0);
  if (mode == IO_READ)
    r_tim += (t1-t0);
  else if (mode == IO_WRIT)
    w_tim += (t1-t0);
}
