/*
 * kernel.c
 *
 * $Source: /fs/nene/uysal/pf/newFact/RCS/kernel.c,v $
 *
 * $Header: /fs/nene/uysal/pf/newFact/RCS/kernel.c,v 1.1 1995/10/16 03:42:25 uysal Exp $
 *
 * Numerical computation kernels for the Cholesky factorization. Note that 
 * these kernels are stripped down versions of the in-core kernels and there
 * are more restrictions on the use of them all of which should be handled
 * in the driver routine.
 * 
 * Author : Mustafa Uysal <uysal@cs.umd.edu>
 *
 */
#include <stdio.h>
#include <math.h>

#include "world.h"
#include "block.h"
#include "kernel.h"
#include "blas.h"

#include "debug.h"

#if   !defined(REAL)   /* Code for Floating Point Arithmetic */
#error "rkernel.c contains only Floating-Point Arithmetic kernels"
#endif 

#define LEVEL3

#define RE_SQRTFP   4  /* FLOP values for different basic OPS (Real Kernels) */
#define RE_DIVFP    4
#define RE_RECIPFP  3
#define RE_MULTFP   1
#define RE_ADDFP    1
#define RE_MUADDFP  2



static int iONE = 1;
static fp_t MONE, PONE, ZERO;

int flop_count;



void KernelInit(verbose)
    int verbose;
{
#if defined(LEVEL0)
  print_if(verbose, ("\nKernel - User-Level subroutines"));
#elif defined(LEVEL1)
  print_if(verbose, ("\nKernel - Level-1 BLAS subroutines"));
#elif defined(LEVEL3)
  print_if(verbose, ("\nKernel - Level-3 BLAS subroutines"));
#endif


  MONE = (fp_t) -1.0;
  PONE = (fp_t) 1.0;
  ZERO = (fp_t) 0.0;
  flop_count = 0;
}




/*------------------------------------------------------------*/
/*            User Level Numerical Kernels                    */
/*------------------------------------------------------------*/
#if defined(LEVEL0)  /* Level-0 BLAS Kernels */
void dgfact(dblk)
    block_t *dblk;
{
  nonzero_t mult;  /* Equivalent to 1/diagonal */
  int   mindex;    /* Index of the first element in the multipler row */
  int   mincr;     /* Increment to find the next multiplier row */
  int   dindex;    /* Index of the current diagonal element (start of column) */
  int   cindex;
  int   base;
  int   uincr;
  int   j, mptr, bptr;
  int   size;

#if defined(SNUPD_LOG)
  if (dblk->xcoor == PVT_WATCH)
    printf("\n@DFCT[%d,%d]", dblk->urow, dblk->lcol);
#endif

  size = BlockSize(dblk);
  for (mindex=mincr=0; mindex<size; mindex+=(++mincr))
    {
      /* Update L(j,j) */
      dindex = mindex + mincr;
      for (j=mindex; j<dindex; j++)
	dblk->value[dindex] -= dblk->value[j] * dblk->value[j];
      
#ifdef PDEF_CHECK
      if (dblk->value[dindex]<=ZERO) dblk->value[dindex] = PONE;
      /*exit_if(dblk->value[dindex] <= ZERO,
	        ("\nNegative diagonal at column %d value %lf",
	       dindex-mindex+dblk->lcol, dblk->value[dindex]));*/
#endif

      /* Compute L(j,j) */
      dblk->value[dindex] = sqrt(dblk->value[dindex]);
      mult = PONE / dblk->value[dindex];

      /* Update elements that are in the same column as dindex */
      uincr = mincr+1;
      for (base=mindex+uincr, cindex=dindex+uincr; base<size; base+=uincr, cindex+=uincr) 
	{
	  for (mptr=mindex, bptr=base; bptr<cindex; bptr++, mptr++) 
	    dblk->value[cindex] -= dblk->value[mptr] * dblk->value[bptr];
	  ++uincr;
	}
      
      /* Scale subdiagonal elements i+1:size */
      for (uincr=mincr+1, dindex+=(mincr+1); dindex<size; dindex+=(++uincr))
	dblk->value[dindex] *= mult;
    }
}





/*
 * dgupdt()
 *
 * Computes the update of a rectangular dense block to a diagonal.
 */
void dgupdt(diag, base)
     block_t *diag;     /* Diagonal being updated */
     block_t *base;     /* Updating block */
{
  nonzero_t *basptr; /* Pointer to the base row */
  nonzero_t *mulptr; /* Pointer to the multiplier row */
  int   bhsize;      /* horizontal size of the updating block */
  int   bvsize;      /* vertical size of the updating block */
  int   dindex;      /* Index of the diagonal block coordinate being updated */
  int   mndxst;      /* Saved value of dest index for next multiplier */
  int   mincr;       /* Saved value of increment for next multipler */
  int   dincr;       /* increment of the dindex for next iteration */
  int   colno;       /* Difference between the first updated column and the leftmost
		      * border of the diagonal block */
  int   i,j,k;
  int   dgsize;
  
#if defined(SNUPD_LOG)
  if (diag->xcoor == PVT_WATCH)
    printf("\n@DUPD[%d,%d] <- [%d,%d, %.12lE]", diag->urow, diag->lcol,
	   base->urow, base->lcol, base->value[0]);
#endif

  dgsize = BlockSize(diag);
  mulptr = base->value;
  bvsize = base->lrow - base->urow;
  bhsize = base->rcol - base->lcol;
  colno  = base->urow - diag->urow;
  mndxst = colno*(colno+1)/2+colno;
  mincr  = colno;

  for (i=0; i<bvsize; i++, mulptr+=bhsize, mndxst+=(++mincr+1))
    {
      dindex = mndxst;
      dincr  = mincr;
      basptr = mulptr;
      for (j=i; j<bvsize; j++, basptr+=bhsize, dindex+=(++dincr))
	for (k=0; k<bhsize; k++)
	  diag->value[dindex] -= mulptr[k] * basptr[k];
    }
}






/*
 * bmod()
 *
 * Computes base-multiplier updates to non-diagonal dense blocks.
 *
 */
void bmod(dest, base, mult)
     block_t *dest, *base, *mult;
{
  nonzero_t *mulptr;    /* Pointer to the current multiplier row */
  nonzero_t *basptr;    /* Pointer to the current base row */
  nonzero_t *mulend;    /* Pointer to the last element in the multiplier */
  nonzero_t *bassta;    /* Pointer to the first element in the base */
  int    bhsize;    /* Horizontal size of base block */
  int    bfrstrow;  /* First row in the update from the base */
  int    blastrow;  /* Last row in the update from the base */
  int    dhsize;    /* Horizontal size of the destination block */
  int    dfrstcol;  /* First column in the update from destination */
  int    dfrstrow;  /* First row in the update from destination */
  int    dstndxst;  /* Storage for the destination index */
  int    dstndx;    /* Index of the destination element being updated */
  int    k, j;      /* Innermost loop variable */

  /* return if alignment mismatch */
  if (base->lrow<=dest->urow) return;
  if (base->urow>=dest->lrow) return;

#if defined(SNUPD_LOG)
  if (dest->xcoor == PVT_WATCH)
    {
      printf("\n@BMOD[%d,%d] <- [%d,%d] * [%d, %d] ->info:%d %d", dest->urow, dest->lcol, 
	     base->urow, base->lcol, mult->urow, mult->lcol, base->lrow, dest->lrow);
      printf("\n\t-= %.12lE * %.12lE", base->value[0], mult->value[0]);
    }
#endif

  /* Initialization of Multiplier variables */
  mulend = mult->value + BlockSize(mult);

  /* Initialization of Base variables */
  bhsize   = base->rcol - base->lcol;
  bfrstrow = MAX(dest->urow, base->urow) - base->urow;
  blastrow = MIN(dest->lrow, base->lrow) - base->urow;
  bassta   = base->value + bfrstrow * bhsize;
  
  /* Initialization of Destination variables */
  dhsize   = dest->rcol - dest->lcol;
  dfrstcol = mult->urow - dest->lcol;
  dfrstrow = MAX(dest->urow, base->urow) - dest->urow;
  dstndxst = dfrstrow * dhsize + dfrstcol;
  
  /* allegedly efficient compute kernel */
  for (mulptr = mult->value; mulptr<mulend; mulptr+=bhsize)
    {
      dstndx = dstndxst++;
      for (j=bfrstrow; j<blastrow; dstndx+=dhsize, j++)
	{
	  basptr = base->value + j*bhsize;
	  for (k=0; k<bhsize; k++)
	    dest->value[dstndx] -= mulptr[k] * basptr[k];
	}
    }
}





/*
 * dmod()
 *
 * Computes the update of two Rectangles on a diagonal block.
 *
 */
void dmod(dest, base, mult)
     block_t *dest, *base, *mult;
{
  nonzero_t *mulptr;    /* Pointer to the current multiplier row */
  nonzero_t *basptr;    /* Pointer to the current base row */
  nonzero_t *mulend;    /* Pointer to the last element in the multiplier */
  int    bvsize;    /* Vertical size of base block */
  int    bhsize;    /* Horizontal size of base block */
  int    dfrstcol;  /* First column in the update from destination */
  int    dfrstrow;  /* First row in the update from destination */
  int    dstndx;    /* Index of the destination element being updated */
  int    dstndxst;  /* Stored value of dstndx */
  int    dincr;     /* Increment variable for the destination index */
  int    k,j;       /* Innermost loop variable */

#if defined(SNUPD_LOG)
  if (dest->xcoor == PVT_WATCH)
    printf("\n@DMOD[%d,%d] <- [%d,%d] * [%d, %d]", dest->urow, dest->lcol,
	   base->urow, base->lcol, mult->urow, mult->lcol);
#endif

  /* Initialization of Multiplier variables */
  mulend = mult->value + BlockSize(mult);
  
  /* Initialization of Base variables */
  bhsize = base->rcol - base->lcol;
  bvsize = base->lrow - base->urow;
  
  /* Initialization of Destination variables */
  dfrstcol = mult->urow - dest->lcol;
  dfrstrow = base->urow - dest->urow;
  dstndxst = dfrstrow*(dfrstrow+1)/2+dfrstcol;

  /* Allegedly efficient compute kernel for this op */
  for (mulptr=mult->value; mulptr<mulend; mulptr+=bhsize)
    {
      dstndx = dstndxst++;
      dincr  = dfrstrow;
      for (basptr=base->value, j=0; j<bvsize; j++, basptr+=bhsize, dstndx+=(++dincr))
	for (k=0; k<bhsize; k++)
	  dest->value[dstndx] -= mulptr[k] * basptr[k];
    }
}






/*
 * drmod()
 *
 * Computes the Subdiagonal rect Update of a diagonal block
 */
void drmod(blk, diag)
    block_t *blk;     /* block being factored */
    block_t *diag;    /* diagonal block */
{
  nonzero_t *mulptr;     /* Pointer to the multipler element */
  nonzero_t multip;      /* Multiplier for the column scale in dest block */
  int   dhsize;      /* Width of the current block panel */
  int   dupdndx;     /* Index for the updated destination index */
  int   dbasndx;     /* Index of the destination base elemen */
  int   i,j;         /* loop variables */
  int   size;

#if defined(SNUPD_LOG)
  if (blk->xcoor == PVT_WATCH)
    printf("\n@DRMD[%d,%d] <- [%d,%d]", blk->urow, blk->lcol,
	   diag->urow, diag->lcol);
#endif
  
  /* Initialization */
  mulptr = diag->value;
  dhsize = blk->rcol - blk->lcol;
  size   = BlockSize(blk);

  /* Allegedly efficient Compute Kernel */
  for (i=0; i<dhsize; i++)
    {
      /* Update the new column */
      for (j=0; j<i; j++)
	{
	  for (dupdndx=i,dbasndx=j; dupdndx<size; dbasndx+=dhsize,dupdndx+=dhsize)
	    blk->value[dupdndx] -= (*mulptr) * blk->value[dbasndx];
	  mulptr++;
	}
      
      /* Scale the subdiagonal elements */
      multip = PONE / *mulptr;
      mulptr++;
      for (j=i; j<size; j+=dhsize)
	blk->value[j] *= multip;
    }
}

#endif defined(LEVEL0)








/*------------------------------------------------------------*/
/*------------------------------------------------------------*/
/*                                                            */
/*          Numerical Kernels with Level 1 BLAS               */
/*                                                            */
/*------------------------------------------------------------*/
/*------------------------------------------------------------*/

#if defined(LEVEL1)

void dgfact(dblk)
    block_t *dblk;
{
  nonzero_t mult;  /* Equivalent to 1/diagonal */
  int   mindex;    /* Index of the first element in the multipler row */
  int   mincr;     /* Increment to find the next multiplier row */
  int   dindex;    /* Index of the current diagonal element (start of column) */
  int   cindex;
  int   base;
  int   uincr;
  int   j, mptr, bptr;
  int   size;

#if defined(SNUPD_LOG)
  if (dblk->xcoor == PVT_WATCH)
    printf("\n@DFCT[%d,%d]", dblk->urow, dblk->lcol);
#endif

  size = BlockSize(dblk);
  for (mindex=mincr=0; mindex<size; mindex+=(++mincr))
    {
      /* Update L(j,j) */
      dindex = mindex + mincr;
      dblk->value[dindex] -= DOT(&mincr, dblk->value+mindex, &iONE, 
				 dblk->value+mindex, &iONE);

#ifdef PDEF_CHECK
      if (dblk->value[dindex]<=ZERO) dblk->value[dindex] = PONE;
      /*EXIT_IF(dblk->value[dindex] <= ZERO,
	      (msg, "\nNegative diagonal at column %d value %lf",
	       dindex-mindex+dblk->lcol, dblk->value[dindex]));*/
#endif

      /* Compute L(j,j) */
      dblk->value[dindex] = sqrt(dblk->value[dindex]);
      mult = PONE / dblk->value[dindex];

      /* Update elements that are in the same column as dindex */
      uincr = mincr+1;
      for (base=mindex+uincr, cindex=dindex+uincr; base<size; base+=uincr, cindex+=uincr) 
	{
	  dblk->value[cindex] -= DOT(&mincr, dblk->value+mindex, &iONE, 
					 dblk->value+base, &iONE);

	  ++uincr;
	}
      
      /* Scale subdiagonal elements i+1:size */
      for (uincr=mincr+1, dindex+=(mincr+1); dindex<size; dindex+=(++uincr))
	dblk->value[dindex] *= mult;
    }
}





/*
 * dgupdt()
 *
 * Computes the update of a rectangular dense block to a diagonal.
 */
void dgupdt(diag, base)
     block_t *diag;     /* Diagonal being updated */
     block_t *base;     /* Updating block */
{
  nonzero_t *basptr; /* Pointer to the base row */
  nonzero_t *mulptr; /* Pointer to the multiplier row */
  int   bhsize;      /* horizontal size of the updating block */
  int   bvsize;      /* vertical size of the updating block */
  int   dindex;      /* Index of the diagonal block coordinate being updated */
  int   mndxst;      /* Saved value of dest index for next multiplier */
  int   mincr;       /* Saved value of increment for next multipler */
  int   dincr;       /* increment of the dindex for next iteration */
  int   colno;       /* Difference between the first updated column and the leftmost
		      * border of the diagonal block */
  int   i,j,k;
  int   dgsize;
  
#if defined(SNUPD_LOG)
  if (diag->xcoor == PVT_WATCH)
    printf("\n@DUPD[%d,%d] <- [%d,%d, %.12lE]", diag->urow, diag->lcol,
	   base->urow, base->lcol, base->value[0]);
#endif

  dgsize = BlockSize(diag);
  mulptr = base->value;
  bvsize = base->lrow - base->urow;
  bhsize = base->rcol - base->lcol;
  colno  = base->urow - diag->urow;
  mndxst = colno*(colno+1)/2+colno;
  mincr  = colno;

  for (i=0; i<bvsize; i++, mulptr+=bhsize, mndxst+=(++mincr+1))
    {
      dindex = mndxst;
      dincr  = mincr;
      basptr = mulptr;
      for (j=i; j<bvsize; j++, basptr+=bhsize, dindex+=(++dincr))
	diag->value[dindex] -= DOT(&bhsize, mulptr, &iONE, basptr, &iONE);
    }
}






/*
 * bmod()
 *
 * Computes base-multiplier updates to non-diagonal dense blocks.
 *
 */
void bmod(dest, base, mult)
     block_t *dest, *base, *mult;
{
  nonzero_t *mulptr;    /* Pointer to the current multiplier row */
  nonzero_t *basptr;    /* Pointer to the current base row */
  nonzero_t *mulend;    /* Pointer to the last element in the multiplier */
  nonzero_t *bassta;    /* Pointer to the first element in the base */
  int    bhsize;    /* Horizontal size of base block */
  int    bfrstrow;  /* First row in the update from the base */
  int    blastrow;  /* Last row in the update from the base */
  int    dhsize;    /* Horizontal size of the destination block */
  int    dfrstcol;  /* First column in the update from destination */
  int    dfrstrow;  /* First row in the update from destination */
  int    dstndxst;  /* Storage for the destination index */
  int    dstndx;    /* Index of the destination element being updated */
  int    k, j;      /* Innermost loop variable */

  /* return if alignment mismatch */
  if (base->lrow<=dest->urow) return;
  if (base->urow>=dest->lrow) return;

#if defined(SNUPD_LOG)
  if (dest->xcoor == PVT_WATCH)
    {
      printf("\n@BMOD[%d,%d] <- [%d,%d] * [%d, %d] ->info:%d %d", dest->urow, dest->lcol, 
	     base->urow, base->lcol, mult->urow, mult->lcol, base->lrow, dest->lrow);
      printf("\n\t-= %.12lE * %.12lE", base->value[0], mult->value[0]);
    }
#endif

  /* Initialization of Multiplier variables */
  mulend = mult->value + BlockSize(mult);

  /* Initialization of Base variables */
  bhsize   = base->rcol - base->lcol;
  bfrstrow = MAX(dest->urow, base->urow) - base->urow;
  blastrow = MIN(dest->lrow, base->lrow) - base->urow;
  bassta   = base->value + bfrstrow * bhsize;
  
  /* Initialization of Destination variables */
  dhsize   = dest->rcol - dest->lcol;
  dfrstcol = mult->urow - dest->lcol;
  dfrstrow = MAX(dest->urow, base->urow) - dest->urow;
  dstndxst = dfrstrow * dhsize + dfrstcol;
  
  /* allegedly efficient compute kernel */
  for (mulptr = mult->value; mulptr<mulend; mulptr+=bhsize)
    {
      dstndx = dstndxst++;
      basptr = base->value + bfrstrow*bhsize;
      for (j=bfrstrow; j<blastrow; dstndx+=dhsize, j++, basptr += bhsize)
	dest->value[dstndx] -= DOT(&bhsize, mulptr, &iONE, basptr, &iONE);
    }
}





/*
 * dmod()
 *
 * Computes the update of two Rectangles on a diagonal block.
 *
 */
void dmod(dest, base, mult)
     block_t *dest, *base, *mult;
{
  nonzero_t *mulptr;    /* Pointer to the current multiplier row */
  nonzero_t *basptr;    /* Pointer to the current base row */
  nonzero_t *mulend;    /* Pointer to the last element in the multiplier */
  int    bvsize;    /* Vertical size of base block */
  int    bhsize;    /* Horizontal size of base block */
  int    dfrstcol;  /* First column in the update from destination */
  int    dfrstrow;  /* First row in the update from destination */
  int    dstndx;    /* Index of the destination element being updated */
  int    dstndxst;  /* Stored value of dstndx */
  int    dincr;     /* Increment variable for the destination index */
  int    k,j;       /* Innermost loop variable */

#if defined(SNUPD_LOG)
  if (dest->xcoor == PVT_WATCH)
    printf("\n@DMOD[%d,%d] <- [%d,%d] * [%d, %d]", dest->urow, dest->lcol,
	   base->urow, base->lcol, mult->urow, mult->lcol);
#endif

  /* Initialization of Multiplier variables */
  mulend = mult->value + BlockSize(mult);
  
  /* Initialization of Base variables */
  bhsize = base->rcol - base->lcol;
  bvsize = base->lrow - base->urow;
  
  /* Initialization of Destination variables */
  dfrstcol = mult->urow - dest->lcol;
  dfrstrow = base->urow - dest->urow;
  dstndxst = dfrstrow*(dfrstrow+1)/2+dfrstcol;

  /* Allegedly efficient compute kernel for this op */
  for (mulptr=mult->value; mulptr<mulend; mulptr+=bhsize)
    {
      dstndx = dstndxst++;
      dincr  = dfrstrow;
      for (basptr=base->value, j=0; j<bvsize; j++, basptr+=bhsize, dstndx+=(++dincr))
	dest->value[dstndx] -= DOT(&bhsize, mulptr, &iONE, basptr, &iONE);
    }
}






/*
 * drmod()
 *
 * Computes the Subdiagonal rect Update of a diagonal block
 */
void drmod(blk, diag)
    block_t *blk;     /* block being factored */
    block_t *diag;    /* diagonal block */
{
  nonzero_t *mulptr;     /* Pointer to the multipler element */
  nonzero_t multip;      /* Multiplier for the column scale in dest block */
  int   dhsize;      /* Width of the current block panel */
  int   dvsize;      /* Height of the block */
  int   dupdndx;     /* Index for the updated destination index */
  int   dbasndx;     /* Index of the destination base elemen */
  int   i,j;         /* loop variables */
  int   size;

#if defined(SNUPD_LOG)
  if (blk->xcoor == PVT_WATCH)
    printf("\n@DRMD[%d,%d] <- [%d,%d]", blk->urow, blk->lcol,
	   diag->urow, diag->lcol);
#endif
  
  /* Initialization */
  mulptr = diag->value;
  dhsize = blk->rcol - blk->lcol;
  dvsize = blk->lrow - blk->urow;
  size   = BlockSize(blk);

  /* Allegedly efficient Compute Kernel */
  for (i=0; i<dhsize; i++)
    {
      /* Update the new column */
      for (j=0; j<i; j++)
	{
	  for (dupdndx=i,dbasndx=j; dupdndx<size; dbasndx+=dhsize,dupdndx+=dhsize)
	    blk->value[dupdndx] -= (*mulptr) * blk->value[dbasndx];
	  mulptr++;
	}
      
      /* Scale the subdiagonal elements */
      multip = PONE / *mulptr;
      mulptr++;
      /*
      for (j=i; j<size; j+=dhsize)
	blk->value[j] *= multip;
      */
      SCAL(&dvsize, &multip, blk->value+i, &dhsize);
    }
}
#endif defined(LEVEL1)








/*------------------------------------------------------------*/
/*------------------------------------------------------------*/
/*                                                            */
/*          Numerical Kernels with Level 3 BLAS               */
/*                                                            */
/*------------------------------------------------------------*/
/*------------------------------------------------------------*/

#if defined(LEVEL3)

void dgfact(dblk)
    block_t *dblk;
{
  nonzero_t mult;  /* Equivalent to 1/diagonal */
  int   mindex;    /* Index of the first element in the multipler row */
  int   mincr;     /* Increment to find the next multiplier row */
  int   dindex;    /* Index of the current diagonal element (start of column) */
  int   cindex;
  int   base;
  int   uincr;
  int   j, mptr, bptr;
  int   size;

  size = BlockSize(dblk);
  for (mindex=mincr=0; mindex<size; mindex+=(++mincr))
    {
      /* Update L(j,j) */
      dindex = mindex + mincr;
      dblk->value[dindex] -= DOT(&mincr, dblk->value+mindex, &iONE, 
				 dblk->value+mindex, &iONE);

#ifdef PDEF_CHECK
      if (dblk->value[dindex]<=ZERO) dblk->value[dindex] = PONE;
      /*EXIT_IF(dblk->value[dindex] <= ZERO,
	      (msg, "\nNegative diagonal at column %d value %lf",
	       dindex-mindex+dblk->lcol, dblk->value[dindex]));*/
#endif

      /* Compute L(j,j) */
      dblk->value[dindex] = sqrt(dblk->value[dindex]);
      mult = PONE / dblk->value[dindex];

      /* Update elements that are in the same column as dindex */
      uincr = mincr+1;
      for (base=mindex+uincr, cindex=dindex+uincr; base<size; base+=uincr, cindex+=uincr) 
	{
	  dblk->value[cindex] -= DOT(&mincr, dblk->value+mindex, &iONE, 
					 dblk->value+base, &iONE);

	  ++uincr;
	}
      
      /* Scale subdiagonal elements i+1:size */
      for (uincr=mincr+1, dindex+=(mincr+1); dindex<size; dindex+=(++uincr))
	dblk->value[dindex] *= mult;
    }

  /*flop_count += (dblk->rcol - dblk->lcol)*(RE_SQRTFP + RE_RECIPFP)  */
}





/*
 * dgupdt()
 *
 * Computes the update of a rectangular dense block to a diagonal.
 */
void dgupdt(diag, base)
     block_t *diag;     /* Diagonal being updated */
     block_t *base;     /* Updating block */
{
  nonzero_t *basptr; /* Pointer to the base row */
  nonzero_t *mulptr; /* Pointer to the multiplier row */
  int   bhsize;      /* horizontal size of the updating block */
  int   bvsize;      /* vertical size of the updating block */
  int   dindex;      /* Index of the diagonal block coordinate being updated */
  int   mndxst;      /* Saved value of dest index for next multiplier */
  int   mincr;       /* Saved value of increment for next multipler */
  int   dincr;       /* increment of the dindex for next iteration */
  int   colno;       /* Difference between the first updated column and the leftmost
		      * border of the diagonal block */
  int   i,j,k;
  int   dgsize;
  
  dgsize = BlockSize(diag);
  mulptr = base->value;
  bvsize = base->lrow - base->urow;
  bhsize = base->rcol - base->lcol;
  colno  = base->urow - diag->urow;
  mndxst = colno*(colno+1)/2+colno;
  mincr  = colno;

  for (i=0; i<bvsize; i++, mulptr+=bhsize, mndxst+=(++mincr+1))
    {
      dindex = mndxst;
      dincr  = mincr;
      basptr = mulptr;
      for (j=i; j<bvsize; j++, basptr+=bhsize, dindex+=(++dincr))
	diag->value[dindex] -= DOT(&bhsize, mulptr, &iONE, basptr, &iONE);
    }
}






/*
 * bmod()
 *
 * Computes base-multiplier updates to non-diagonal dense blocks.
 *
 */
void bmod(dest, base, mult)
     block_t *dest, *base, *mult;
{
  nonzero_t *mulptr;    /* Pointer to the current multiplier row */
  nonzero_t *basptr;    /* Pointer to the current base row */
  nonzero_t *mulend;    /* Pointer to the last element in the multiplier */
  nonzero_t *bassta;    /* Pointer to the first element in the base */
  nonzero_t *deststa;   /* Pointer to the first element in the dest */
  int    bhsize;    /* Horizontal size of base block */
  int    bfrstrow;  /* First row in the update from the base */
  int    blastrow;  /* Last row in the update from the base */
  int    dhsize;    /* Horizontal size of the destination block */
  int    dfrstcol;  /* First column in the update from destination */
  int    dfrstrow;  /* First row in the update from destination */
  int    dstndxst;  /* Storage for the destination index */
  int    dstndx;    /* Index of the destination element being updated */
  int    bvsize  ;  /* Vertical size of the base block */
  int    mvsize;    /* Vertical size of the multiplier block */
  int    k, j;      /* Innermost loop variable */
  char   yt,nt;

  /* return if alignment mismatch */
  if (base->lrow<=dest->urow) return;
  if (base->urow>=dest->lrow) return;

  /* Initialization of Multiplier variables */
  mulend = mult->value + BlockSize(mult);
  mvsize = mult->lrow - mult->urow;

  /* Initialization of Base variables */
  bhsize   = base->rcol - base->lcol;
  bfrstrow = MAX(dest->urow, base->urow);
  bvsize   = MIN(dest->lrow, base->lrow) - bfrstrow;

  blastrow = MIN(dest->lrow, base->lrow) - base->urow;
  bassta   = base->value + (bfrstrow - base->urow) * bhsize;
  
  /* Initialization of Destination variables */
  dhsize   = dest->rcol - dest->lcol;
  dfrstcol = mult->urow - dest->lcol;
  dfrstrow = MAX(dest->urow, base->urow) - dest->urow;
  dstndxst = dfrstrow * dhsize + dfrstcol;
  deststa  = dest->value + dstndxst;

  /* BLAS specific variables */
  nt = 'N'; yt = 'T';
  GEMM(&yt,&nt,&mvsize,&bvsize,&bhsize,&MONE,mult->value,&bhsize,
       bassta, &bhsize, &PONE, deststa, &dhsize);
}





/*
 * dmod()
 *
 * Computes the update of two Rectangles on a diagonal block.
 *
 */
void dmod(dest, base, mult)
     block_t *dest, *base, *mult;
{
  nonzero_t *mulptr;    /* Pointer to the current multiplier row */
  nonzero_t *basptr;    /* Pointer to the current base row */
  nonzero_t *mulend;    /* Pointer to the last element in the multiplier */
  int    bvsize;    /* Vertical size of base block */
  int    bhsize;    /* Horizontal size of base block */
  int    dfrstcol;  /* First column in the update from destination */
  int    dfrstrow;  /* First row in the update from destination */
  int    dstndx;    /* Index of the destination element being updated */
  int    dstndxst;  /* Stored value of dstndx */
  int    dincr;     /* Increment variable for the destination index */
  int    k,j;       /* Innermost loop variable */

#if defined(SNUPD_LOG)
  if (dest->xcoor == PVT_WATCH)
    printf("\n@DMOD[%d,%d] <- [%d,%d] * [%d, %d]", dest->urow, dest->lcol,
	   base->urow, base->lcol, mult->urow, mult->lcol);
#endif

  /* Initialization of Multiplier variables */
  mulend = mult->value + BlockSize(mult);
  
  /* Initialization of Base variables */
  bhsize = base->rcol - base->lcol;
  bvsize = base->lrow - base->urow;
  
  /* Initialization of Destination variables */
  dfrstcol = mult->urow - dest->lcol;
  dfrstrow = base->urow - dest->urow;
  dstndxst = dfrstrow*(dfrstrow+1)/2+dfrstcol;

  /* Allegedly efficient compute kernel for this op */
  for (mulptr=mult->value; mulptr<mulend; mulptr+=bhsize)
    {
      dstndx = dstndxst++;
      dincr  = dfrstrow;
      for (basptr=base->value, j=0; j<bvsize; j++, basptr+=bhsize, dstndx+=(++dincr))
	dest->value[dstndx] -= DOT(&bhsize, mulptr, &iONE, basptr, &iONE);
    }
}






/*
 * drmod()
 *
 * Computes the Subdiagonal rect Update of a diagonal block
 */
void drmod(blk, diag)
    block_t *blk;     /* block being factored */
    block_t *diag;    /* diagonal block */
{
  nonzero_t *mulptr;     /* Pointer to the multipler element */
  nonzero_t multip;      /* Multiplier for the column scale in dest block */
  int   dhsize;      /* Width of the current block panel */
  int   dvsize;      /* Height of the block */
  int   dupdndx;     /* Index for the updated destination index */
  int   dbasndx;     /* Index of the destination base elemen */
  int   i,j;         /* loop variables */
  int   size;

#if defined(SNUPD_LOG)
  if (blk->xcoor == PVT_WATCH)
    printf("\n@DRMD[%d,%d] <- [%d,%d]", blk->urow, blk->lcol,
	   diag->urow, diag->lcol);
#endif
  
  /* Initialization */
  mulptr = diag->value;
  dhsize = blk->rcol - blk->lcol;
  dvsize = blk->lrow - blk->urow;
  size   = BlockSize(blk);

  /* Allegedly efficient Compute Kernel */
  for (i=0; i<dhsize; i++)
    {
      /* Update the new column */
      for (j=0; j<i; j++)
	{
	  for (dupdndx=i,dbasndx=j; dupdndx<size; dbasndx+=dhsize,dupdndx+=dhsize)
	    blk->value[dupdndx] -= (*mulptr) * blk->value[dbasndx];
	  mulptr++;
	}
      
      /* Scale the subdiagonal elements */
      multip = PONE / *mulptr;
      mulptr++;
      SCAL(&dvsize, &multip, blk->value+i, &dhsize);
    }
}

#endif defined(LEVEL3)
