/*
 * snode.c
 *
 * $Source: /fs/nene/uysal/pf/newFact/RCS/snode.c,v $
 *
 * $Header: /fs/nene/uysal/pf/newFact/RCS/snode.c,v 1.1 1995/10/16 03:42:25 uysal Exp $
 *
 * Resource management and dependency control modules for
 * supernodes
 *
 * Author : Mustafa Uysal <uysal@cs.umd.edu>
 *
 */
#include <malloc.h>

#include "mem.h"
#include "snode.h"
#include "disk.h"
#include "queue.h"

#include "debug.h"


/*--------------- Static Variables ---------------*/

static int snCount;
static int *Offset;
static int *Size;

static snode_t **snStore;


void SnodeFreeForced ARGS((snode_t *ptr));
void SnodeCompact ARGS((int id, int size));


/*--------------- Implementation ---------------*/


/*
 * Initialization of supernode structures altogether
 */
void SnodeInit(cnt, offset, size)
    int cnt;
    int *offset, *size;
{
  int i;

  snCount = cnt;
  Offset = offset;
  Size = size;

  snStore = (snode_t**) MemGet(cnt*sizeof(snode_t**));
  for (i=0; i<cnt; i++)
    snStore[i] = (snode_t*) NULL;
}




/*
 * Extracts the list of updaters for a particular supernode
 * from the supernode data just read from the disk. 
 */
int *SnodeUpdaters(sptr, cnt)
    snode_t *sptr;
    int *cnt;
{
  int *ptr = (int*)sptr->buffer;

  *cnt = *ptr;
  return(ptr+1);
}




/*
 * Allocate memory for a supernode data + metadata 
 */
snode_t *SnodeAlloc(id, offset, size)
    int id, offset, size;
{
  caddr_t memptr;
  snode_t *sptr;

  /* Allocate buffer space if available */
  memptr = MemGet(sizeof(snode_t) + size);
  exit_if(!memptr, ("No Memory - SnodeAlloc(%d,%d,%d)", id, offset, size));

  snStore[id] = sptr = (snode_t*) memptr;
  sptr->buffer = memptr + sizeof(snode_t);
  sptr->bufq = QueueNew();

  sptr->sid = id;
  sptr->offset = offset;
  sptr->size = size;

  sptr->blkcnt = 0;
  sptr->blkPtr = NULL;
  sptr->status = SN_NULL;

  return sptr;
}




/* Routine to reclaim memory allocated by a supernode */
void SnodeFree( ptr )
    snode_t *ptr;
{
  if (!ptr) return;
  if (ptr->status == SN_LOCKWT)
    {
      ptr->status = SN_WRITING;
      return;
    }
  SnodeFreeForced(ptr);
}





/*
 * Routine to merge the block descriptors for two separate portions
 * of a supernode
 */
void SnodeBlockMerge(src, buf, size, pid)
    snode_t *src;
    caddr_t buf;
    int size;
    int pid;
{
  caddr_t adr;
  snode_t dummy;
  block_t *newptr;
  int newcnt, bcnt;
  int i,j,k;

  /* First let the pointers point to valid addresses */
  dummy.buffer = buf;
  dummy.size   = size;
  dummy.sid    = src->sid;
  SetUpValuePtrs(&dummy, pid, DSK_NEXT);

  /* allocate storage of new block descriptor */
  if (!(newcnt = src->blkcnt + dummy.blkcnt))
    {
      Free(buf);
      return;
    }
  newptr = (block_t*) Calloc(newcnt, sizeof(block_t));

  /* simple merge algorithm should work */
  for (i=j=k=0; (i<src->blkcnt) && (j<dummy.blkcnt); k++)
    {
      if (src->blkPtr[i].urow < dummy.blkPtr[j].urow)
	newptr[k] = src->blkPtr[i++];
      else
	newptr[k] = dummy.blkPtr[j++];
    }
  for ( ; i<src->blkcnt; i++, k++)
    newptr[k] = src->blkPtr[i];
  for ( ; j<dummy.blkcnt; j++, k++)
    newptr[k] = dummy.blkPtr[j];    

  /* setup correct pointers upon return */
  src->blkPtr = newptr;
  src->blkcnt = newcnt;
  
  /* Add the new address into the queue of reclaimables */
  QueueAdd(src->bufq, newptr);
  QueueAdd(src->bufq, buf);
}




/* Routine to query the disk offset where a supernode is located */
int SnodeOffset( id )
    int id;
{
  return Offset[id];
}



/* Routine to query the size of a supernode on disk */
int SnodeSize(id)
    int id;
{
  return Size[id];
}



/* Routine to adjust the offset & size of a supernode */
void  SnodeLocatorAdjust( id, diff )
    int id, diff;
{
  Offset[id] += diff;
  Size[id] -= diff;

  dbg_if(Size[id]<0, panic(("Locator Error size:(%d) for %d, diff=%d", 
			    Size[id], id, diff)));
}



/* Routine to service supernode requests from memory if possible */
snode_t *SnodeRequest(id)
    int id;
{
  dbg_if(id<0||id>snCount, panic(("Invalid Snode Request for:%d", id)));
  return snStore[id];
}




void SnodeExit()
{
  int i;

  for (i=0; i<snCount; i++)
    if (snStore[i])
      {
	message(("\n\tAlive [sn=%d sz=%d st=%d]", i, snStore[i]->size, snStore[i]->status));
	snStore[i]->status = SN_INUSE;
	SnodeFreeForced(snStore[i]);
      }

  MemPut(snStore, snCount * sizeof(snode_t**));
}




void SnodeCompact(id, size)
    int id, size;
{
  int i, bfs;

#if 0
  for (bfs=i=0, size*=2; i<id && bfs<size; i++)
    if (snStore[i] && snStore[i]->status == SN_CACHED)
      {
	bfs += snStore[i]->size;
	SnodeFreeForced(snStore[i]);
      }
#endif
}





/* Routine to reclaim memory allocated by a supernode */
void SnodeFreeForced( ptr )
    snode_t *ptr;
{
  caddr_t adr;
  int bcnt;

  dbg_if(!ptr, panic(("NULL-SN ??")));
  dbg_if(!(ptr->status==SN_INUSE || ptr->status==SN_WRITING),
	 panic(("\nError: status=%d (!= %d) of %d", ptr->status, SN_INUSE, ptr->sid)));
  dbg_if(ptr!=snStore[ptr->sid],
	 panic(("\nDangling Ptr of Sn:%d", ptr->sid)));

  /* take the supernode off-line */
  snStore[ptr->sid] = (snode_t*) NULL;

  /* go through the allocated buffers */
  while (!QueueEmpty(ptr->bufq))
    {
      adr = QueueDelete(ptr->bufq);
      Free(adr);
    }
  QueueDestroy(ptr->bufq);

  /* reclaim main buffer */  
  MemPut(ptr, sizeof(snode_t)+ptr->size);
}




int SnodeAdjustSize(sptr, pvt, plim)
    snode_t *sptr;
    int pvt, plim;
{
  int j, cnt;

  cnt = 0;
  for (j=0; j<sptr->blkcnt && sptr->blkPtr[j].ycoor < pvt; j++) 
    ;
  
  plim += pvt;
  for ( ; (j<sptr->blkcnt) && (sptr->blkPtr[j].ycoor < plim) ; j++)
    cnt += BlockSizeByte(&sptr->blkPtr[j]);

  return cnt;
}



void SnodeStructPrint( sptr )
    snode_t *sptr;
{
  block_t *bptr;
  int j;
  
  for (j=0; j<sptr->blkcnt; j++)
    {
      bptr = &sptr->blkPtr[j];
      message(("\n\tBlk[%d of %d] is [%d %d @ %d %d %d %d] sz:%d", j, sptr->sid,
	       bptr->ycoor, bptr->xcoor, bptr->lcol, bptr->rcol,
	       bptr->urow, bptr->lrow, BlockSizeByte(bptr)));
    }
}
