/*
 * disk.c
 *
 * $Source: /fs/nene/uysal/pf/Part/RCS/disk.c,v $
 *
 * $Header: /fs/nene/uysal/pf/Part/RCS/disk.c,v 1.2 1995/10/09 19:27:01 uysal Exp $
 *
 * Routines to handle the disk I/O requests for the block partitioner
 *
 * Applicable Compiler Directives :
 * DUMP_METADATA   The metadata of the matrix is dumped in a human
 *                 readble form to stdout
 *
 * Author : Mustafa Uysal <uysal@cs.umd.edu>
 *
 */
#include <stdio.h>
#include <fcntl.h>
#include <malloc.h>

#include "world.h"
#include "mppworld.h"
#include "disk.h"
#include "queue.h"
#include "io.h"
#include "part.h"
#include "mtags.h"


#define NDX_CHUNK_SZ  (5*MEGABYTE)
#define VAL_CHUNK_SZ  (5*MEGABYTE)

extern nonzero_t ZERO;

/*--------------- Local Variables ---------------*/

static int orgFile;      /* File handler for original matrix */
static int symFile;      /* File handler for symbolic factor structure */
static int atrFile;      /* File handler to store the attribute data */
static int blkFile;      /* File handler to store the blocks */

static int glbCol;       /* Number of columns in the sparse matrix */
static int lclCol;       /* Local column count assigned to inpPid */

static nonzero_t *dataVect;
static int *indxVect;
static int *snOffs, *snSize, snCnt;
static int localOffs;
static int localNnz;
static int localAtr;

static double t0, t1;      /* timing variables */
double seek_time, read_time, write_time;
double bytes_read, bytes_written;
int seek_cntr;

/* Timers for supernode.c */
extern double sn_time, sm_time, sn_rt, sn_br;

/* Timers for decompose module */
double dec_rt, dec_wt, dec_br, dec_bw;



/*--------------- Local Function Protoypes ---------------*/

/* Routine to attach a string to an established string */
char *stratt ARGS((char *s1, char *s2));

/* Routine to read in the structure of a certain column (sequential access) */
int  *ColumnStructRead ARGS((int fid, int colno));

/* Routine to read in the original column (sequential access) */
int  *ColumnValueRead ARGS((int fid, int colno));

/* Routine to read a column in a multiprocessor fashion */
void ColumnRead ARGS((int map, int colno, column_t *ptr));


/*--------------- Implementation ---------------*/					
			

/* Routine to initialize timing vars */
void DiskTimingInit()
{
  seek_time = read_time = write_time = bytes_written = bytes_read = 0.0;
  seek_cntr = 0;
}
								


											
/*											
 * Function to initialize the disk I/O module for block					
 * partitioning phase									
 */											
int DiskBlockifyInit(scnt, outfn, sfid, orgfn, nc)
    int scnt, sfid;
    char *outfn, *orgfn;
    int nc;
{
  char *tfn;
  int nfid;
  int i;

  /* Sanity Checks */
  EXIT_IF(VAL_CHUNK_SZ%sizeof(int), 
	  (msg, "VAL_CHUNK_SZ[%d] should be divisible by %d", VAL_CHUNK_SZ, sizeof(int)));
  EXIT_IF(NDX_CHUNK_SZ%sizeof(int), 
	  (msg, "NDX_CHUNK_SZ[%d] should be divisible by %d", NDX_CHUNK_SZ, sizeof(int)));

  /* Initialize file handles */									
  tfn = stratt(orgfn, ".tmp");
  orgFile = open(tfn, O_RDONLY);
  free(tfn);
  symFile = sfid;

  tfn = stratt(outfn, ".atr");
  atrFile = open(tfn, O_CREAT | O_TRUNC | O_WRONLY, 0644);				
  free(tfn);

  tfn = stratt(outfn, ".blk");
  blkFile = open(tfn, O_CREAT | O_TRUNC | O_WRONLY, 0644);				
  free(tfn);

#if 0
  tfn = stratt(outfn, ".ndx");
  nfid = open(tfn, O_CREAT | O_TRUNC | O_WRONLY, 0644);
  free(tfn);
#else
  nfid = 0;
#endif
  
  EXIT_IF((orgFile<0)||(symFile<0)||(atrFile<0)||(blkFile<0)||(nfid<0),		
	  (msg, "DiskBlockifyInit() - Error in opening files [%d %d %d %d %d]",
	   orgFile, symFile, atrFile, blkFile, nfid));			

  /* Allocate and initialize space for local value scattering */			
  glbCol = nc;
  dataVect = (nonzero_t*) calloc(glbCol+1, sizeof(nonzero_t));
  CHKNULL(dataVect);									
  indxVect = (int*) calloc(glbCol+1, sizeof(int));
  CHKNULL(indxVect);									
  for (i=0; i<=glbCol; i++)								
    dataVect[i] = ZERO;									
											
  /* Set up the data file offset counters */						
  snOffs = (int*) calloc(scnt, sizeof(int));						
  CHKNULL(snOffs);									
  snSize = (int*) calloc(scnt, sizeof(int));						
  CHKNULL(snSize);									
  snCnt = scnt;
  localAtr = localNnz = localOffs = 0;

  return nfid;
}





/*
 * Routine to partition the columns of the original matrix
 * among processors based on the supernode assignment : ie.
 * all the columns of a supernode is sent to the processor
 * that will blockify that supernode
 */
void DecomposeOriginalMatrix(fn, sc, sl)
    char *fn;
    int sc;
    int *sl;
{
  queue_t *bufq;
  int fid, ifid;
  int nc, tp, bufszbyte;
  int *header, *buffer;
  int i, colno, sz;
  char *buf, *bufb, *bufa;
  int me, np;
  int ibuf;
  int tg, fm, nsz;

  nc = sl[sc]-1;

  /* Initialize timing vars */
  dec_rt = dec_wt = dec_br = dec_br = 0.0;

  buf = stratt(fn, ".tmp");
  fid = open(buf, O_CREAT | O_TRUNC | O_WRONLY, 0644);
  EXIT_IF(fid<0, (msg, "Cannot open tmp file : %s", buf));
  free(buf);

  bufq = QueueNew();
  me = MyProc();
  np = NumProcs();

  if (!me)
    {
      ifid = open(fn, O_RDONLY);

      buffer = (int*) calloc(2*nc+1, sizeof(int));
      CHKNULL(buffer);

      /* Read the header record */
      TimedRead(ifid, buffer, (2*nc+1)*sizeof(int), &dec_rt, &dec_br);
      header = buffer+1;

      for (i=0; i<sc; i++)
	{
	  tp = i % np;
	  for (colno=sl[i]; colno<sl[i+1]; colno++)
	    {
	      /* buffer to hold the original data */
	      sz = header[2*(colno-1)+1];
	      bufszbyte = sz*(sizeof(int)+sizeof(nonzero_t))+sizeof(int);
	      buf = (char*) malloc(bufszbyte);
	      MALCHK(buf, bufszbyte);

	      /*
	      if (colno<10)
		MESSAGE((msg, "COL[%d] mapped to:%d sz:%d", colno, tp, sz));
	      */

	      /* Read the column structure + nonzeros */
	      TimedRead(ifid, buf, bufszbyte, &dec_rt, &dec_br);
	      memcpy(&ibuf, buf, sizeof(int));
	      EXIT_IF(sz!=(ibuf-1), (msg, "Column format error @%d [%d != %d]",
				 colno, sz, ibuf));
	      
	      /* Ship data to where it belongs */
	      if (tp == me)
		TimedWrite(fid, buf, bufszbyte, &dec_wt, &dec_bw);
	      else
		MsgSend(MSG_COL+colno, tp, buf, bufszbyte);

	      free(buf);
	    }
	}

      close(ifid);
    }
  else
    {
      sz = (nc+1)*(sizeof(int)+sizeof(nonzero_t));
      bufa = (char*) malloc(sz);
      CHKNULL(bufa);
      bufb = (char*) malloc(sz);
      CHKNULL(bufb);

      for (i=0; i<sc; i++)
	{
	  tp = i % np;
	  if (tp!=me) continue;
	  buf = sl[i] % 2 ? bufa : bufb;
	  CommAsyncRecv(buf, 0, MSG_COL+sl[i], sz);
	  for (colno=sl[i]; colno<sl[i+1]; colno++)
	    {
	      /* Post Recv for the other buffer */
	      if ((colno+1)<sl[i+1])
		{
		  buf = (colno+1) % 2 ? bufa : bufb;
		  CommAsyncRecv(buf, 0, MSG_COL+colno+1, sz);
		}

	      /* Make sure that the current buffer is settled */
	      buffer = (int*) RecvQueuePoll();

	      bufszbyte = (buffer[0]-1)*(sizeof(int)+sizeof(nonzero_t))+sizeof(int);
	      buf = (char*) malloc(bufszbyte);
	      CHKNULL(buf);
	      QueueAdd(bufq, buf);

	      memcpy(buf, buffer, bufszbyte);
	    }

	  while (!QueueEmpty(bufq))
	    {
	      buffer = (int*) QueueDelete(bufq);
	      bufszbyte = (buffer[0]-1)*(sizeof(int)+sizeof(nonzero_t))+sizeof(int);
	      TimedWrite(fid, buffer, bufszbyte, &dec_wt, &dec_bw);
	      free(buffer);
	    }
	}

      free(bufa);
      free(bufb);
    }


  QueueDestroy(bufq);
  close(fid);

  /*CommAsyncStatus();*/

}
											
											
											
											
void DiskBlockifyExit(fn)
    char *fn;
{											
  char tfn[500];

  /* local offset write to ATR file */
  ASSERT(write(atrFile, &snCnt, sizeof(int)) == sizeof(int));
  ASSERT(write(atrFile, snOffs, snCnt*sizeof(int)) == snCnt*sizeof(int));
  ASSERT(write(atrFile, snSize, snCnt*sizeof(int)) == snCnt*sizeof(int));

  close(symFile);									
  close(orgFile);									
  close(atrFile);									
  close(blkFile);									

  free(dataVect);
  free(indxVect);

  sprintf(tfn, "/usr/bin/rm -f %s.tmp", fn);
  system(tfn);
}											
											


/* Routine to report disk I/O performance */
void DiskIOReport( void )
{
  MESSAGE((msg, "[INFO] %d Nonzeros, %d bytes of index data - Total:%d bytes", 
	   localNnz, localAtr, localAtr + localNnz * sizeof(nonzero_t)));

  /* Supernode Module */
  MESSAGE((msg, "[SN] Tcomp:[%.3lf secs] Tmerge:[%.3lf secs]", sn_time, sm_time));
  MESSAGE((msg, "[SN] Read:[%.3lf secs], Size:[%.3lf MB], Rate:[%.3lf MB/secs] [%.0lf Bytes]",
	   sn_rt, (sn_br/MEGABYTE), sn_br/(MEGABYTE*sn_rt), sn_br));

  /* Decompose Module */
  MESSAGE((msg, "[DECOMP] Write:[%.3lf secs], Size:[%.3lf MB], Rate:[%.3lf MB/secs]",
           dec_wt, (dec_bw/MEGABYTE), dec_bw/(MEGABYTE*dec_wt))); 
  MESSAGE((msg, "[DECOMP] Read:[%.3lf secs], Size:[%.3lf MB], Rate:[%.3lf MB/secs]",
	   dec_rt, (dec_br/MEGABYTE), dec_br/(MEGABYTE*dec_rt)));

  /* Partition Module */
  MESSAGE((msg, "[PART] Read:[%.3lf secs], Size:[%.3lf MB], Rate:[%.3lf MB/secs]",
	   read_time, (bytes_read/MEGABYTE), bytes_read/(MEGABYTE*read_time)));
  MESSAGE((msg, "[PART] Write:[%.3lf secs], Size:[%.3lf MB], Rate:[%.3lf MB/secs]",
	   write_time, (bytes_written/MEGABYTE), bytes_written/(MEGABYTE*write_time))); 

  /* Grand Totals */
  read_time += (dec_rt + sn_rt);
  bytes_read += (dec_br + sn_br);

  write_time += dec_wt;
  bytes_written += dec_bw;

  MESSAGE((msg, "[TOTAL] Read:[%.3lf secs], Size:[%.3lf MB], Rate:[%.3lf MB/secs]",
	   read_time, (bytes_read/MEGABYTE), bytes_read/(MEGABYTE*read_time)));
  MESSAGE((msg, "[TOTAL] Write:[%.3lf secs], Size:[%.3lf MB], Rate:[%.3lf MB/secs]",
	   write_time, (bytes_written/MEGABYTE), bytes_written/(MEGABYTE*write_time))); 
  MESSAGE((msg, "[TOTAL] Seek:[%.3lf secs], Count:[%d]", seek_time, seek_cntr));
}




/*
 * Read in all local columns within the specified range
 */
column_t *ReadColumnRange(int map, int sta, int end, int *ncol)
{
  column_t *cols;
  int i;

  if (MyProc() == map)
    {
      cols = (column_t*) calloc(end-sta, sizeof(column_t));
      CALCHK(cols, end-sta, sizeof(column_t));
      *ncol = end - sta;
    }
  else 
    {
      cols = NULL;
      *ncol = 0;
    }
  
  for (i=sta; i<end; i++)
    ColumnRead(map, i, cols+i-sta);
  
  return cols;
}






/*											
 * Reads the specified column from the scratch files					
 */											
void ColumnRead(map, colno, ptr)								
    int map;
    int colno;										
    column_t *ptr;									
{											
  int *buffer;
  int sz, bufszbyte;
  int *oldset, oldcnt;									
  int *newset, newcnt;									
  nonzero_t *valset;									
  int i;										

  if (MyProc() == map)  
    {
      buffer = ColumnValueRead(orgFile, colno);
      EXIT_IF(!buffer, (msg, "Decomposition error for col:%d", colno));

      oldcnt = buffer[0]-1;
      oldset = buffer+1;
      valset = (nonzero_t*) (buffer+oldcnt+1);

      /* scatter original value set to dataVect */						
      for (i=0; i<oldcnt; i++)								
	dataVect[oldset[i]] = valset[i];

      /* Either read locally or recv the new column structure */
      if (!(oldset=ColumnStructRead(symFile, colno)))
	{
	  MsgRecv(MSG_COL+colno, indxVect, (glbCol+1)*sizeof(int));
	  oldset = indxVect;
	}
      
      /* Copy the new index structure */
      newcnt = oldset[0]-1;
      newset = (int*) calloc(newcnt, sizeof(int));
      CALCHK(newset, newcnt, sizeof(int));
      for (i=0; i<newcnt; i++)
	newset[i] = oldset[i+1];
      
      /* Gather the nonzeros of the Cholesky Factor */					
      valset = (nonzero_t*) calloc(newcnt, sizeof(nonzero_t));
      CALCHK(valset, newcnt, sizeof(nonzero_t));
      for (i=0; i<newcnt; i++)
	{
	  valset[i] = dataVect[newset[i]];							
	  dataVect[newset[i]] = ZERO;							
	}

      /* Return the data to proc 0 in column_t */
      ptr->colno = colno;									
      ptr->size  = newcnt;									
      ptr->rowIndex = newset;								
      ptr->value = valset;									
    }
  else
    {
      newset = ColumnStructRead(symFile, colno);
      if (newset) MsgSend(MSG_COL+colno, map, newset, newset[0]*sizeof(int));
    }
}											




 /*
  * STRing ATTach for attaching two strings together and making a 
  * new string 
  */
 char *stratt(s1, s2)
      char *s1, *s2;
 {
   char *new;

   new = (char *) malloc(strlen(s1)+strlen(s2)+1);
   CHKNULL(new);

   strcpy(new, s1);
   strcat(new, s2);

   return(new);
 }







int *ColumnStructRead(fid, colno)
    int fid;
    int colno;
{
  static int firsttime = 1;
  static int *buf;
  static int ndx, tsize;
  int offset, i, j, limit;
  
  if (firsttime)
    {
      firsttime = 0;
      TimedSeek(fid, 3*sizeof(int), SEEK_SET, &seek_time, &seek_cntr);
      
      buf = (int*) malloc(NDX_CHUNK_SZ);
      CHKNULL(buf);

      tsize = TimedRead(fid, buf, NDX_CHUNK_SZ, &read_time, &bytes_read);
      ndx = 0;
    }  

  if (colno<0)
    {
      firsttime = 1;
      free(buf);
      buf = NULL;
      return buf;
      ndx = tsize = 0;
    }

  if ((ndx+buf[ndx])*sizeof(int)>=tsize)
    {
      limit = NDX_CHUNK_SZ / sizeof(int);
      for (j=0, i=ndx; i<limit; j++, i++)
	buf[j] = buf[i];
      tsize = TimedRead(fid, buf+j, NDX_CHUNK_SZ-j*sizeof(int), &read_time, &bytes_read);
      tsize += (j*sizeof(int));
      ndx = 0;
    }

  if (buf[ndx+1] == colno)
    {
      offset = ndx;
      ndx += buf[ndx];
      return &buf[offset];
    }
  else
    {
      return NULL;
    }
}





/*
 * History sensitive column value read function
 */
int *ColumnValueRead(fid, colno)
    int fid;
    int colno;
{
  static int firsttime = 1;
  static int *buf;
  static int ndx, tsize;
  int offset, i, j, limit;

  if (firsttime)
    {
      /* initialize and storage alloc */
      ndx = firsttime = 0;
      buf = (int*) malloc(VAL_CHUNK_SZ);
      MALCHK(buf, VAL_CHUNK_SZ);

      /* initial data read */
      tsize = TimedRead(fid, buf, VAL_CHUNK_SZ, &read_time, &bytes_read);
      
      if (!tsize)  return NULL;

      /* sanity check */
      offset = (buf[ndx]-1)*(sizeof(int)+sizeof(nonzero_t)) + sizeof(int);
      EXIT_IF(offset>=tsize, (msg,"Low Memory for VAL-CHUNK: %d < %d",tsize, offset));
    }  

  if (colno<0) /* equivalent of if (lasttime) */
    {
      firsttime = 1;
      free(buf);
      buf = NULL;
      return buf;
    }

  offset = (buf[ndx]-1)*(sizeof(int)+sizeof(nonzero_t))+sizeof(int);

  /* if finished processing of the current buffer */
  if ((ndx*sizeof(int)+offset)>=tsize)
    {
      /* Shift whatever is remaining to the beginning of the buffer */
      limit = VAL_CHUNK_SZ / sizeof(int);
      for (j=0, i=ndx; i<limit; j++, i++)
	buf[j] = buf[i];

      /* fill the remaining buffer */
      tsize = TimedRead(fid, buf+j, VAL_CHUNK_SZ-j*sizeof(int),
			&read_time, &bytes_read);

      tsize += (j*sizeof(int));

      /* if I don't have nothing - what can you do? */
      if (!tsize) return NULL;

      /* reinitialize and sanity check */
      ndx = 0;
      offset = (buf[ndx]-1)*(sizeof(int)+sizeof(nonzero_t)) + sizeof(int);
      EXIT_IF(offset>=tsize, (msg,"Low Memory for VAL-CHUNK: %d < %d",tsize,offset));
    }

  if (buf[ndx+1] == colno)
    {
      /*MESSAGE((msg, "ValRead(c:%d) approved", colno));*/

      i = ndx;
      EXIT_IF(offset%sizeof(int), (msg, "offset[%d] must be div by [%d]",
				   offset, sizeof(int)));
      ndx += (offset/sizeof(int));

      return &buf[i];
    }
  else
    {
      /*MESSAGE((msg, "ValRead(c:%d b:%d) denied @C", colno, buf[ndx+1]));*/
      return NULL;
    }
}







/* Routine to write the local portions of the supernode */
void DiskSNwrite(id, blkq, dl, dc)
    int id;
    queue_t *blkq;
    int *dl, dc;
{                                                                                       
  queue_t *tmpq;
  block_t *ptr;
  int size, len;
  int i, cnt;

  snOffs[id] = localOffs;                                                               
  snSize[id] = 0;                                                                       
  len = sizeof(block_t) - sizeof(int);   /* block_t.size is NOT written */

#if   defined(DUMP_METADATA)
  for (i=0; i<dc; i++)
    MESSAGE((msg, "\tDep[%d/%d of %d]=%d", i, dc, id, dl[i]));
#endif
  
  /* write out the dependency info for the supernode */
  TimedWrite(blkFile, &dc, sizeof(int), &write_time, &bytes_written);
  TimedWrite(blkFile, dl, dc*sizeof(int), &write_time, &bytes_written);
  snSize[id] += ((dc+1) * sizeof(int));
  
  /* block data is written next */
  tmpq = QueueNew();
  while (!QueueEmpty(blkq))
    {
      ptr = (block_t*) QueueDelete(blkq);
      size = ptr->size * sizeof(nonzero_t);
      snSize[id] += size;
      localNnz += ptr->size;
      TimedWrite(blkFile, ptr->value, size, &write_time, &bytes_written);
      QueueAdd(tmpq, ptr);
    }

  /* block descriptors */
  cnt = 0;
  while (!QueueEmpty(tmpq))
    {
      ptr = (block_t*) QueueDelete(tmpq);
#if   defined(DUMP_METADATA)      
      MESSAGE((msg, "\tBlk[%d of %d] is [%d %d @ %d %d %d %d]", cnt, id,
	       ptr->ycoor, ptr->xcoor, ptr->lcol, ptr->rcol, ptr->hrow, ptr->lrow));
#endif
      TimedWrite(blkFile, ptr, len, &write_time, &bytes_written);
      snSize[id] += len;
      localAtr += len;
      free(ptr);
      cnt++;
    }

  /* terminated by # of blocks record */
  TimedWrite(blkFile, &cnt, sizeof(int), &write_time, &bytes_written); 
  snSize[id] += sizeof(int);

  localOffs += snSize[id];
  localAtr += (2+dc)*sizeof(int);

  QueueDestroy(tmpq);
}

