/*
 * part.c
 *
 * $Source: /fs/nene/uysal/pf/Part/RCS/part.c,v $
 *
 * $Header: /fs/nene/uysal/pf/Part/RCS/part.c,v 1.2 1995/10/09 19:27:01 uysal Exp $
 *
 * Sparse Matrix block decomposition and data distribution
 * Parallel/Out-of-core Implementation
 *
 * Author : Mustafa Uysal <uysal@cs.umd.edu>
 *
 */
#include <stdio.h>
#include <fcntl.h>
#include <malloc.h>
#include <stdlib.h>

#include "world.h"
#include "mppworld.h"
#include "part.h"
#include "disk.h"
#include "queue.h"
#include "mtags.h"
#include "dep.h"

extern int errno;

#if defined(DBGLOG)
#undef DBGLOG
#endif

#define BLK_MSG_BUF_SZ     (5*MEGABYTE)



/*--------------- Static Global Variables ---------------*/
static int snLimit;    /* for error checking */

static queue_t *blkq;  /* local block queue */
static queue_t *rcbq;  /* remote block queue */
static queue_t *outq;  /* blocks to be sent out */
static queue_t *ndxq;  /* indices to be stored in the dependencies */

static caddr_t blkmsgbuf;



/*--------------- Function Prototypes ---------------*/


block_t *BlockReceive();
void BlockifyProcess();
void BlockifyWait();
void BlockRegister();
void BlockifySyncPoint();
void BlockifySyncMark();
void RectGenerate();
void DiagGenerate();

int  int_cmp();


/*--------------- Implementation ---------------*/


void BlockPartition(sndCnt, sndList, sfid, origfile, outfile)
    int sndCnt;     /* # of supernodes */
    int *sndList;   /* list of supernode separators */
    int sfid;
    char *origfile, *outfile;
{
  column_t *colptr, *tmpptr;
  int np, ccnt, colcnt, me;
  int i, j, k;

  /* Decomose the matrix based on the supernode structure */
  DecomposeOriginalMatrix(origfile, sndCnt, sndList);

  printf("\nMatrix decompose finished");

  /* Local Initialization */
  snLimit = sndCnt;
  np = NumProcs();
  me = MyProc();
  blkmsgbuf = (caddr_t) malloc(BLK_MSG_BUF_SZ);
  MALCHK(blkmsgbuf, BLK_MSG_BUF_SZ);
  
  /* Initialize file I/O module */
  DiskBlockifyInit(sndCnt, outfile, sfid, origfile, sndList[sndCnt]-1);

  /* Initialize communication module */
  CommBlockifyInit();

  /* Initialize Dependency Module */
  DepModuleInit();

  /* Initialize temporary queues */
  blkq = QueueNew(); 
  rcbq = QueueNew();
  outq = QueueNew();
  ndxq = QueueNew();

  for (i=0; i<sndCnt; i+=np)
    {
      if (!(i%1000)) printf("\n\tSn:%d in part", i);

      /* Parallel Read of local supernodes*/
      for (j=0; j<np && (j+i)<sndCnt; j++)
	{
	  tmpptr = ReadColumnRange(j, sndList[i+j], sndList[i+j+1], &ccnt);
	  if (j == me)
	    {
	      colptr = tmpptr;
	      colcnt = ccnt;
	    }
	  else
	    {
	      EXIT_IF(tmpptr||ccnt, (msg, "Error in ReadColumnRange(i:%d %d,%d,%d,%d)",
				     i, j, sndList[i+j], sndList[i+j+1], ccnt));
	    }
	}
      
      /* Each processor blockifies the local supernode */
      if ((me+i)<sndCnt)
	BlockifyProcess(colcnt, colptr, me+i, sndCnt, sndList);

      /* Synchronize for each round of block transfer */
      for (j=0; j<np && (j+i)<sndCnt; j++)
	BlockifySyncPoint(i+j, j);
    }

  /* housekeeping */
  DiskBlockifyExit(origfile);
  CommBlockifyExit();

  QueueDestroy(blkq);
  QueueDestroy(rcbq);
  QueueDestroy(outq);
  QueueDestroy(ndxq);
  free(blkmsgbuf);

  DiskIOReport();
}







/* 
 * Routine to partition a column-major supernode into row-major dense
 * blocks
 */
void BlockifyProcess(cnt, colptr, snid, scnt, stru)
    int cnt;
    column_t *colptr;
    int snid, scnt;
    int *stru;
{
  int k, colno, x;
  int hi, hindx, londx;
  int i;
  
  /* Sanity Check */
  EXIT_IF(stru[snid]!=colptr->colno, 
	  (msg, "***ERROR*** (%d != %d) @ %d", stru[snid], colptr->colno, snid));

#ifdef DBGLOG
  /* debugging aid : print column indices */
  printf("\n\tCol:%d : ", colptr->colno);
  for (i=0; i<colptr->size; i++)
    printf(" %d", colptr->rowIndex[i]);
#endif
  
  /* Generate the diagonal block */
  x = k = snid;
  DiagGenerate(cnt, colptr, x);
  
  /* Rest of the current supernode is chopped off on the supernode
   * boundaries
   */
  for (i=cnt; i<colptr->size; )
    {
      /* find the snode index of the block header */
      hindx = i;
      for (hi=colptr->rowIndex[i]; (k<scnt) && (stru[k]<hi); k++)
	;
      if (stru[k]>hi) k--;
      
      if (k>=scnt) { assert(k==scnt); k = scnt-1;}
      
      /* flesh out the blocks based on proximity and snode boundaries */
      for (++i;(i<colptr->size) && ((colptr->rowIndex[i] - colptr->rowIndex[i-1])==1) &&
	   (colptr->rowIndex[i] < stru[k+1]); i++)
	;
      londx = i - 1;

      RectGenerate(cnt, colptr, hindx, londx, x, k);
    }

  /* housekeeping */
  for (i=0; i<cnt; i++)
    {
      free(colptr[i].rowIndex);
      free(colptr[i].value);
    }
  free(colptr);
}






void BlockifySyncPoint(id, map)
    int id;
    int map;
{
  caddr_t buf;
  queue_t *locq, *tmpq;
  block_t *ptr;
  int *deplist, depcnt;
  int i;

  if (map == MyProc())
    {
      /* Inject my blocks into network */
      while (!QueueEmpty(outq))
	{
	  ptr = (block_t*) QueueDelete(outq);
	  CommAsyncSend(ptr, BlockMapping(ptr->xcoor, ptr->ycoor),
			MSG_BLK+id, ptr->size*sizeof(nonzero_t)+sizeof(block_t));
	}

      /* Terminate communication round */
      BlockifySyncMark(id);
      locq = blkq;
      tmpq = ndxq;
    }
  else
    {
      /* wait for the blocks to arrive */
      BlockifyWait(id);
      locq = rcbq;
      tmpq = NULL;
    }

  /* Extract supernodal dependencies */
  deplist = DepProcessing(id, map, &depcnt, tmpq);
  qsort(deplist, depcnt, sizeof(int), int_cmp);

  /* Write Out */
  DiskSNwrite(id, locq, deplist, depcnt);
  if (deplist) free(deplist);

  /* Poll until all my messages are gone */
  while (buf = SendQueuePoll())
    free(buf);
}





/* Routine to generate blocks along the diagonal */
void DiagGenerate(cnt, colptr, coor)
    int cnt;          /* size of the diagonal (in columns) */
    column_t *colptr; /* array of columns from which we extract nonzeros */
    int coor;         /* block coordinate in the matrix */
{
  caddr_t buf;
  block_t blk;
  int t,i,d,j,sz;
  
  /* error control */
  EXIT_IF(coor>=snLimit,
	  (msg, "Error in DiagGenerate x=y=%d l=%d", coor,snLimit));

  /* Metadata generate */
  blk.xcoor = blk.ycoor = coor;
  blk.lcol  = blk.hrow  = colptr->colno;
  blk.rcol  = blk.lrow  = colptr->rowIndex[cnt-1] + 1;
  ASSERT(colptr->colno == colptr->rowIndex[0]);
  
  /* Data portion - alloc space */
  t = blk.rcol - blk.lcol;
  blk.size = t * (t + 1) / 2;

  sz = blk.size*sizeof(nonzero_t) + sizeof(block_t);
  buf = (caddr_t) malloc(sz);
  MALCHK(buf, sz);
  blk.value = (nonzero_t*) (buf+sizeof(block_t));

  /* column major --> row major conversion */
  for (i=0; i<cnt; i++)
    {
      t = i * (i + 1) / 2 + i;
      d = i;

      for (j=0; j<(cnt-i); j++)
	{
	  blk.value[t] = colptr[i].value[j];
	  t += (++d);
	}
    }

#ifdef DBGLOG
  MESSAGE((msg, "\tDiagGen(%d %d)", blk.lcol, blk.rcol));
#endif

  BlockProduce(&blk, buf, coor);
}





void RectGenerate(cnt, colptr, hindx, londx, x, y)
    int cnt;
    column_t *colptr;
    int hindx, londx;
    int x,y;
{
  caddr_t buf;
  block_t blk;
  int t,i,d,j, sz;

  /* error control */
  DBG_MSG((x>=snLimit)||(y>=snLimit),
	  (msg, "in RectGenerate x=%d y=%d l=%d hr=%d lr=%d lx:%d hx:%d sz:%d",
	   x,y,snLimit,colptr->rowIndex[hindx],colptr->rowIndex[londx],
	   londx, hindx, colptr->size));
  if (y>=snLimit) y = snLimit-1;
  
  /* Metadata generate */
  blk.xcoor = x;
  blk.ycoor = y;
  blk.lcol  = colptr->colno;
  blk.rcol  = colptr[cnt-1].colno + 1;
  blk.hrow  = colptr->rowIndex[hindx];
  blk.lrow  = colptr->rowIndex[londx] + 1;
  
  /* Data portion - alloc space */
  d = blk.rcol - blk.lcol;
  t = blk.lrow - blk.hrow;
  blk.size = t * d;
  EXIT_IF(blk.size<=0, (msg, "RectGen - block with size 0, t=%d d=%d", t, d));
  
  sz = blk.size*sizeof(nonzero_t) + sizeof(block_t);
  buf = (caddr_t) malloc(sz);
  MALCHK(buf, sz);
  blk.value = (nonzero_t*) (buf+sizeof(block_t));

  /* column major --> row major conversion */
  for (i=0; i<cnt; i++)
    {
      for (t=i, j=hindx; j<=londx; j++, t+=d)
	blk.value[t] = colptr[i].value[j];

      hindx--;
      londx--;
    }

#ifdef DBGLOG
  MESSAGE((msg, "RectGen(%d %d %d %d)@(%d %d)", blk.lcol, blk.rcol, blk.hrow, blk.lrow,
	   blk.xcoor, blk.ycoor));
#endif
  
  IndexRegister(ndxq, blk.ycoor);
  BlockProduce(&blk, buf, x);
}





/* Routine to produce a block */
int BlockProduce(blk, buf, round)
    block_t *blk;
    caddr_t buf;
    int round;
{
  int m;

  /* Prepend the block metadata to the front */
  memcpy(buf, blk, sizeof(block_t));

  m = BlockMapping(blk->xcoor, blk->ycoor);

  if (m != MyProc())
    QueueAdd(outq, buf);
  else
    QueueAdd(blkq, buf);
}






/* Routine to recv all incoming blocks for round snid */
void BlockifyWait(snid)
    int snid;
{
  block_t *blk;

  while (blk = BlockReceive(snid))
    QueueAdd(rcbq, blk);
}





/* Routine to recv a block from the network (blocking) */
block_t *BlockReceive(round)
    int round;
{
  block_t *blk;
  caddr_t new;
  int len;

  len = MsgRecvFrom(MSG_BLK+round, blkmsgbuf, BLK_MSG_BUF_SZ, DONTCARE);
  blk = (block_t*) blkmsgbuf;

  if (blk->xcoor == -1)
    return (block_t *) NULL;

  new = (caddr_t) malloc(len);
  MALCHK(new, len);

  memcpy(new, blkmsgbuf, len);
  blk = (block_t*) new;
  blk->value = (nonzero_t*) new+sizeof(block_t);

  return new;
}





/* Routine to terminate the current round of block sends from this
 * particular processor */
void BlockifySyncMark(round)
    int round;
{
  block_t *blk;
  int i;

  for (i=0; i<NumProcs(); i++)
    if (i != MyProc())
      {
	blk = (block_t*) malloc(sizeof(block_t));
	MALCHK(blk, sizeof(block_t));

	blk->xcoor = -1;
	CommAsyncSend(blk, i, MSG_BLK+round, sizeof(block_t));
      }
}




int int_cmp(a, b)
    int *a, *b;
{
  if (*a < *b) return -1;
  if (*a > *b) return +1;
  return 0;
}
