/*
 * sp2.c
 *
 * $Source: /fs/nene/uysal/pf/newFact/RCS/sp2.c,v $
 *
 * $Header: /fs/nene/uysal/pf/newFact/RCS/sp2.c,v 1.1 1995/10/16 03:42:25 uysal Exp $
 *
 * Implementation of configuration and message passing functions for
 * the IBM SP-2 platform.
 *
 * Author : Mustafa Uysal <uysal@cs.umd.edu>
 *
 * Notes : 
 *
 */
#include <mpproto.h>

#if    !defined(SP2)
#define SP2
#endif !defined(SP2)

#include "nworld.h"
#include "nmppworld.h"
#include "queue.h"
#include "list.h"
#include "ndebug.h"


#define MSG_CONCAT  40000
#define POLL_INT     1000

#define RAW_COMM    1


/*-------------- Type Definitions ----------------------------*/

/* sendq member item type */
typedef struct {
  int      msgid;    /* msg identifier for the async send */
  caddr_t  bufadr;   /* ptr to the message buffer */
} send_t;



/* recvq member item type */
typedef struct {
  caddr_t  buffer;
  int      msgid;
  int      source;
  int      tag;
  int      len;
} recv_t;


/* message type */
typedef struct {
  caddr_t buf;
  int tag;
  int src;
  int len;
} mesg_t;




/*-------------- Global Variable Declarations ----------------*/

static queue_t *sendq; /* queue of outstanding asynchrounous send requests */
static queue_t *recvq; /* queue of outstanding asynchrounous recv requests */
static list_t *waitl;  /* list of messages that we are waiting */
static list_t *arrvl;  /* list of arrived but not used messages */
static int me, np;

/* Statistics Variables */
static double c_snd, c_rcv;
static double c_drnw, c_drnwnum;
static double c_sflush, c_drain, c_poll;
static long c_svl, c_rvl;



/*-------------- Local Function Prototypes ---------------------*/
double w_seconds ARGS((void));

#if defined(DRAINER)
int _MsgCompare ARGS((mesg_t *ap, mesg_t *bp));
void _MsgPrint ARGS((mesg_t *ap));
#endif defined(DRAINER)


/*--------------- Implementation ---------------*/


int NumProcs( VOID )
{
  int numtask, taskid;

  mpc_environ(&numtask, &taskid);
  return(numtask);
}




int MyProc( VOID )
{
  int numtask, taskid;

  mpc_environ(&numtask, &taskid);
  return(taskid);  
}



void MPPinit()
{
  /* bare essentials */
  me = MyProc();
  np = NumProcs();

  /* initialize async send/recv message queues */
  sendq = QueueNew();
  recvq = QueueNew();
  waitl = ListNew();
  arrvl = ListNew();

  /* product version tracking */
#if defined(SP2)
  print_if(!me, ("\nSP-2 custom version for Communication Software"));
#else
  print_if(!me, ("\nGeneric communication software"));
#endif

#if defined(PRODUCTION)
  print_if(!me, ("\nProduction level Communication Software, no checks for blocking"));
#else
  print_if(!me, ("\nDevelopment level Communication Software, checks for blocking"));
#endif

#if defined(DRAINER)
  print_if(!me, ("\nNetwork drainer installed in Communication Subsystem"));
#endif

  c_snd = c_rcv = c_drnw = c_drnwnum = c_sflush = c_drain = c_poll = 0.0;
  c_svl = c_rvl = 0;
} 



/* Routine to report the state of the MPP routine */
void MPPstatus()
{
  message(("\n---------- Status of MPP module: ----------"));
  MsgRecvDiscardAll();
  ListPrintAll(waitl, _MsgPrint, "waitl");
  ListPrintAll(arrvl, _MsgPrint, "arrvl");
}




void MPPreport()
{
  double total;
  
  message(("\nCOM-VOLUME snd:[%ld Bytes] rcv:[%ld Bytes]", c_svl, c_rvl));
  message(("\nCOM-TIME   snd:[%.3lf secs] rcv:[%.3lf secs]", c_snd, c_rcv));
  message(("\nCOM-RATE   snd:[%.3lf MB/s] rcv:[%.3lf MB/s]", 
	   (double)c_svl/((c_snd+c_sflush)*(double)MEGABYTE),
	   (double)c_rvl/((c_rcv+c_poll)*(double)MEGABYTE)))
  message(("\nCOM-DRAIN  snd:[%.3lf secs] rcv:[%.3lf secs] pol:[%.3lf secs]", 
	   c_sflush, c_drain, c_poll));

  total = c_sflush + c_drain + c_snd + c_rcv;
  message(("\nCOM-TOTAL-TIME [%.3lf secs]", total));
}







/* Communication module status report */
void CommAsyncStatus()
{
  queue_t *tmpq;
  recv_t *rcv;

  message(("\tComm: sendq=%d recvq=%d", sendq->load, recvq->load));
  if (!QueueEmpty(recvq))
    {
      tmpq = QueueNew();
      while (!QueueEmpty(recvq))
	{
	  rcv = (recv_t*) QueueDelete(recvq);
	  message(("[RCV - %d %d %d]",  rcv->tag, rcv->source, rcv->msgid));
	  QueueAdd(tmpq, rcv);
	}
      QueueDestroy(recvq);
      recvq = tmpq;
    }
}



/* Routine to send an asynchronous message */
void CommAsyncSend(adr, to, tag, size)
    caddr_t adr;
    int     to,tag,size;
{
  double t0, t1;
  send_t *sqitem;

  t0 = CLOCK();
  sqitem = (send_t*) Malloc(sizeof(send_t));
  if (mpc_send(adr, size, to, tag, &(sqitem->msgid))!=0)
    panic(("Send error : %d [sz:%d to:%d tg:%d]", mperrno, size, to, tag));

  sqitem->bufadr = adr;
  QueueAdd(sendq, sqitem);
  t1 = CLOCK();

  c_snd += (t1-t0);
  c_svl += size;

  dbg_msg(RAW_COMM, ("\n\tasnd(%d,%d,%d)", tag, to, size));
}





/* Routine to Probe the send queue */
int SendQueueProbe ARGS((void))
{
  send_t *snd;

  /* Send queue is empty */
  if (QueueEmpty(sendq))
    return 0;

  /* Probe the head of the queue for completion */
  snd = (send_t*) QueueHead(sendq);
  if (mpc_status(snd->msgid)<0)
    return 0;

  /* Operation completed, discard the queue entry (Semantics) */
  snd = (send_t*) QueueDelete(sendq);
  Free(snd);
  return 1;
}




/* Routine to poll the outgoing message queue for a completion */
caddr_t SendQueuePoll()
{
  caddr_t adr;
  send_t *snd;
  int cnt;

  if (QueueEmpty(sendq))
    return (caddr_t) NULL;

  snd = (send_t*) QueueDelete(sendq);

  for (cnt=1; mpc_status(snd->msgid)<0; cnt++)
    {
#if defined(DRAINER)
      if (!(cnt%POLL_INT)) NetworkDrain();
#endif
#if !defined(PRODUCTION)
      exit_if(cnt>TOOMANY, ("SendQueuePoll() blocked."));
#endif
    }

  adr = snd->bufadr;
  Free(snd);

  return adr;
}





/* Routine to drain the send queue */
void SendQueueDrain()
{
  double t0, t1;

  t0 = CLOCK();
  while (SendQueueProbe())
    ;
  c_sflush += (CLOCK()-t0);
}





/* Routine to flush the send queue */
void SendQueueFlush()
{
  double t0;

  t0 = CLOCK();
  while (SendQueuePoll())
    ;
  c_sflush += (CLOCK()-t0);
}





/* Broadcast of a message from a designated processor to all */
void MsgBroadcast(origin, message, size, tag)
    int origin;
    void *message;
    int size, tag;
{
  int i;

  if (me != origin)
    {
      MsgRecv(tag, message, size);
      return;
    }
  
  for (i=0; i<np; i++)
    if (i != me)
      MsgSend(tag, i, message, size);
}




void MsgSend(tag, to, buf, len)
     int tag;
     int to;
     char *buf;
     int len;
{
  double t0;

  t0 = CLOCK();
  if (mpc_bsend(buf, len, to, tag) < 0)
    panic(("Error in msg send code=%d tag:%d to:%d len:%d", mperrno, tag, to, len));

  c_snd += (CLOCK() - t0);
  c_svl += len;

  dbg_msg(RAW_COMM, ("\n\tbsnd(%d,%d,%d)", tag, to, len));
}










/* Routine to probe the incoming message queue */
int MsgProbe(tag, from)
    int tag, from;
{
  int size;

  if (mpc_probe(&from, &tag, &size)<0)
    panic(("mpc_probe(%d,%d,%d) failed, error:%d.", from, tag, size, mperrno));

  return size;
}








void MsgRecvDiscardAll()
{
  char *tbuf;
  int llen, lsrc, ltag, cnt;
  int newlen;
  
  ltag = lsrc = DONTCARE;
  while(mpc_probe(&lsrc, &ltag, &newlen)==0)
    {
      if (newlen<0) break;
      message(("\nMsgDiscard tg:%d fm:%d ln:%d", ltag, lsrc, newlen));
      tbuf = (char*) Malloc(newlen+1);
      mpc_brecv(tbuf, newlen, &lsrc, &ltag, &llen);
      Free(tbuf);
      ltag = lsrc = DONTCARE;
    }
}




double CLOCK()
{
  return w_seconds();
}




void Sync( void )
{
  mpc_sync(ALLGRP);
}


int LeftProc( proc )
    int proc;
{
  return (!proc) ? (np-1) : (proc-1);
}



int RightProc( proc )
    int proc;
{
  return (proc==(np-1)) ? 0 : proc+1;
}


int LastProc( proc )
    int proc;
{
  return (proc==(np-1)) ? 1 : 0;
}




/* Routine to concatenate local buffers in processor order */
void MsgConcat(inbuf, inlen, outbuf, outlen)
    char *inbuf, **outbuf;
    int inlen, *outlen;
{
  char *lclbuf;
  int *szvct, *gsvct, tsize;
  int *srcvct, *typvct, *idvct;
  int i,j;

  /* default return values */
  *outbuf = NULL;
  *outlen = 0;

  /* prepare comm buffers */
  szvct = (int*) Calloc(np, sizeof(int));
  gsvct = (int*) Calloc(np, sizeof(int));
  for (i=0; i<np; i++)
    gsvct[i] = szvct[i] = 0;
  szvct[me] = inlen;

  /* obtain global size vector */
  mpc_combine(szvct, gsvct, np*sizeof(int), i_vadd, ALLGRP);

  /* allocate memory for output */
  for (tsize=i=0; i<np; i++)
    tsize += gsvct[i];

  if (!tsize) return;

  lclbuf = (char*) Malloc(tsize);
  *outlen = tsize;
  
  /* post local rcvs from each processor - crude implementation : ALL-to-ALL */
  srcvct = (int*) Calloc(np, sizeof(int));
  typvct = (int*) Calloc(np, sizeof(int));
  for (tsize=i=0; i<np; i++)
    if (i!=me)
      {	
	srcvct[i] = i;
	typvct[i] = MSG_CONCAT+i;
	mpc_recv(lclbuf+tsize, gsvct[i], &srcvct[i], &typvct[i], &szvct[i]);
	tsize += gsvct[i];
	mpc_send(inbuf, inlen, i, MSG_CONCAT+me, &gsvct[i]);
      }
    else
      {
	memcpy(lclbuf+tsize, inbuf, inlen);
	tsize += gsvct[i];
      }

  /* Wait for messages to complete */
  for (i=0; i<np; i++)
    if (i!=me)
      {
	mpc_wait(&szvct[i], &tsize);
	mpc_wait(&gsvct[i], &tsize);
      }

  Free(szvct);
  Free(gsvct);
  Free(srcvct);
  Free(typvct);

  *outbuf = lclbuf;
}






#if !defined(DRAINER)



/* Routine to recv a message from the network */
int MsgRecv(tag, buf, len)
    int tag;
    char *buf;
    int len;
{
  int newlen, lsrc, ltag, cnt, lid;
  double t0;
  
  t0 = CLOCK();

  /* Post an ASYNC Recv */
  lsrc = DONTCARE;
  ltag = tag;
  if (mpc_recv(buf, len, &lsrc, &ltag, &lid)!=0)
    panic(("Recv error : %d [sz:%d fm:%d tg:%d]", mperrno, len, lsrc, tag));

  /* Block for it */
#if defined(PRODUCTION)
  while ((newlen = mpc_status(lid))<0)
    ;
#else
  for (cnt=0; (newlen = mpc_status(lid))<0; cnt++)
    exit_if(cnt>TOOMANY,("MsgRecv() blocked for tag:%d frm:%d siz:%d", tag, lsrc, len));
#endif

  /* after op checks */
  dbg_if (newlen>len, panic(("ShortBuffer tag=%d len=%d nlen=%d src=%d",tag,len,newlen,lsrc)));
  c_rcv += (CLOCK()-t0);

  return newlen;
}




/* Routine to recv a message from the network */
int MsgRecvFrom(tag, buf, len, src)
    int tag;
    char *buf;
    int len;
    int src;
{
  int newlen, lsrc, ltag, cnt, lid;
  double t0;

  t0 = CLOCK();

  /* Post an ASYNC Recv */
  lsrc = src;
  ltag = tag;
  if (mpc_recv(buf, len, &lsrc, &ltag, &lid)!=0)
    panic(("Recv error : %d [sz:%d fm:%d tg:%d]", mperrno, len, lsrc, tag));


  /* Block for it */
#if defined(PRODUCTION)
  while ((newlen = mpc_status(lid))<0)
    ;
#else
  for (cnt=0; (newlen = mpc_status(lid))<0; cnt++)
    exit_if(cnt>TOOMANY,("MsgRecvFrom(%d,%d,%d) blocked.", tag, lsrc, len));
#endif

  /* after op checks */
  dbg_if(newlen>len, panic(("ShortBuffer tag=%d len=%d nlen=%d src=%d",tag,len,newlen,lsrc)));
  c_rcv += (CLOCK() - t0);
  
  return newlen;
}




/* Routine to recv an asynchronous message */
void CommAsyncRecv(adr, from, tag, size)
    caddr_t adr;
    int     from,tag,size;
{
  recv_t *rqitem;
  double t0;

  t0 = CLOCK();
  rqitem = (recv_t*) Malloc(sizeof(recv_t));
  rqitem->buffer = adr ? adr : (caddr_t) Malloc(size);

  rqitem->source = from;
  rqitem->tag    = tag;
  if (mpc_recv(rqitem->buffer, size, &(rqitem->source), &(rqitem->tag), &(rqitem->msgid))!=0)
    panic(("Recv error : %d [sz:%d fm:%d tg:%d]", mperrno, size, from, tag));

  QueueAdd(recvq, rqitem);
  dbg_msg(1, ("\n\taRcv(%d,%d,%d)", tag, from, size));

  c_rcv += (CLOCK() - t0);
}




/* Routine to poll the incoming message queue for completion of a single message */
caddr_t RecvQueuePoll()
{
  int sz, fm, tg;
  caddr_t adr;
  recv_t *rcv;
  int cnt, lsz;
  double t0;

  if (QueueEmpty(recvq)) return NULL;
  t0 = CLOCK();

  rcv = (recv_t*) QueueDelete(recvq);

#if   defined(PRODUCTION)
  while ((lsz=mpc_status(rcv->msgid))<0)
    ;

#else /* Non-production */
  for (cnt=0; (lsz=mpc_status(rcv->msgid))<0; cnt++)
    {
      if (cnt>TOOMANY)
	panic(("RecvQueuePoll() blocked - t:%d fm:%d\n",rcv->tag,rcv->source));
    }
#endif

  adr = rcv->buffer;
  *tg = rcv->tag;
  *fm = rcv->source;
  *sz = lsz;
  Free(rcv);
  c_rcv += (CLOCK() - t0);
  
  return adr;
}




/* Routine to probe the incoming message queue for completions */
caddr_t RecvQueueProbe(sz, fm, tg)
    int *sz;
    int *fm;
    int *tg;
{
  caddr_t adr;
  recv_t *rcv;
  int cnt = 0;
  
  /* check if there is a pending message */
  *tg = *fm = *sz = -1;
  if (QueueEmpty(recvq))  return NULL;

  /* check if the pending message has completed or not */
  *sz = 0;
  rcv = (recv_t*) QueueHead(recvq);
  cnt = mpc_status(rcv->msgid);
  if (cnt<0)   return NULL;

  /* delete a completed message from queue, returning the contents */
  rcv = (recv_t*) QueueDelete(recvq);
  adr = rcv->buffer;
  *fm = rcv->source;
  *tg = rcv->tag;
  Free(rcv);

  *sz = cnt;
  return adr;
}






void MsgDiscard(tag)
    int tag;
{
  caddr_t buffer;
  int size;

  /* Wait for the message to arrive */
  while ((size = MsgProbe(tag, DONTCARE))<0)
    ;
  
  /* Receive and discard the message */
  buffer = (caddr_t) Malloc(size);
  MsgRecv(tag, buffer, size);
  Free(buffer);
}






#elif defined(DRAINER)



/* Network drainer for avoiding deadlocks */
int NetworkDrain()
{
  mesg_t msg, *mp;
  int count, retval, retlen;
  double t0;

  t0 = CLOCK();
  for (count=0;;count++) 
    {
      msg.src = DONTCARE;
      msg.tag = DONTCARE;
      
      if (mpc_probe(&msg.src, &msg.tag, &msg.len))
	panic(("mpc_probe(%d,%d,%d) failed with error=%d",msg.src,msg.tag,msg.len,mperrno));

      /* no more messages in the incoming queue */
      if (msg.len == -1)
	{
	  c_drain += (CLOCK() - t0);
	  return count;
	}
      
      /* try to match the message with an already posted recv */
      if (!(mp = (mesg_t*) ListFindDelete(waitl, &msg, _MsgCompare)))
	{
	  /* Try to allocate buffer, if fails, bail out */
	  mp = (mesg_t*) Malloc(sizeof(mesg_t));
	  *mp = msg;
	  mp->buf = Malloc(msg.len+1);
	  dbg_msg(1, ("\ndrain-new(%d,%d,%d)", mp->tag, mp->src, mp->len));
	}
      else
	{
	  if (!mp->buf)
	    {
	      mp->buf = (caddr_t) Malloc(msg.len+1);
	    }
	  dbg_msg(1, ("\ndrain-wait(%d,%d,%d)", mp->tag, mp->src, mp->len));
	}

      /* recv the message */
      if (mpc_brecv(mp->buf, msg.len, &mp->src, &mp->tag, &mp->len))
	panic(("Recv error : %d [sz:%d fm:%d tg:%d]", mperrno, mp->len, mp->src, mp->tag));
      c_rvl += mp->len;

      ListInsertTail(arrvl, mp);
    }
}





/* Network drainer used for catching a specific message, while draining the 
 * network
 */
caddr_t DrainNetworkWhileWait(buf, tag, src, len)
    caddr_t buf;
    int *tag, *src, *len;
{
  mesg_t msg, *mp, tmp;
  int count, retval, retlen;
  double t0;

  t0 = CLOCK();

  tmp.tag = *tag;
  tmp.src = *src;

  for (count=0;;count++)
    {
      msg.src = DONTCARE;
      msg.tag = DONTCARE;
      
      if (mpc_probe(&msg.src, &msg.tag, &msg.len))
	panic(("mpc_probe(%d,%d,%d) failed with error=%d",msg.src,msg.tag,msg.len,mperrno));

      /* no more messages in the incoming queue */
      if (msg.len == -1) 
	{
	  SendQueueDrain();
#if !defined(PRODUCTION)
	  if (count>TOOMANY)
	    {
	      MPPstatus();
	      panic(("DrainNetwork(%d,%d,%d) did not arrive.", *tag, *src, *len));
	    }
	  else
#endif 
	    continue;
	}

      /* Got the message that I am waiting for */
      if (_MsgCompare(&tmp, &msg)==0) 
	{
	  if (!buf)
	    {
	      tmp.buf = Malloc(msg.len+1);
	    }
	  else
	    {
	      tmp.buf = buf;
	      dbg_if (*len<msg.len,
		      {
			_MsgPrint(&tmp, "tmpm");
			MPPstatus();
			panic(("\nShort buffer (%d<%d) t:(%d,%d) f:(%d,%d)",*len, 
			       msg.len, tmp.tag, msg.tag,tmp.src, msg.src));
		      });
	    }
	  
	  if (mpc_brecv(tmp.buf, msg.len, src, tag, len))
	    panic(("Recv error : %d [sz:%d fm:%d tg:%d]",mperrno,tmp.len,tmp.src,tmp.tag));
	  c_drain += (CLOCK() - t0);
	  c_rvl   += *len;
	  
	  return tmp.buf;
	}

      /* try to match the message with an already posted recv */
      mp = (mesg_t*) ListFindDelete(waitl, &msg, _MsgCompare);
      if (!mp)
	{
	  mp = (mesg_t*) Malloc(sizeof(mesg_t));
	  *mp = msg;
	  mp->buf = (caddr_t) Malloc(msg.len+1);
	}

      if (!mp->buf)
	mp->buf = (caddr_t) Malloc(msg.len+1);

      /* recv the message */
      if (mpc_brecv(mp->buf, msg.len, &mp->src, &mp->tag, &mp->len))
	panic(("\nRecv error : %d [sz:[%d]%d fm:%d tg:%d]", mperrno, msg.len, mp->len, 
	       mp->src, mp->tag));
      c_rvl += mp->len;

      ListInsertTail(arrvl, mp);
    }
}









/* Routine to recv an asynchronous message */
void CommAsyncRecv(adr, from, tag, size)
    caddr_t adr;
    int     from,tag,size;
{
  recv_t *rqitem;
  mesg_t msg, *mp;
  int i;
  double t0;

  t0 = CLOCK();
  message(("\n\tarcv-in(%d, %d, %d)", tag, from, size));
  
  rqitem = (recv_t*) Malloc(sizeof(recv_t));

  /* Initialize mesg structs */
  msg.src = rqitem->source = from;
  msg.tag = rqitem->tag    = tag;
  msg.len = size;

  /* check if we have already received it */
  if (mp = (mesg_t*) ListLookup(arrvl, &msg, _MsgCompare))
    {
      dbg_if(size>=0 && size<mp->len, 
	     panic(("\narcv-short buffer (%d<%d), msg(%d[%d],%d[%d])", size, mp->len,
		    tag, mp->tag, from, mp->src)));
      dbg_msg(1, ("\n\tarcv-in(%d(%d), %d(%d), %d(%d))", tag, mp->tag, from, mp->src, 
		  size, mp->len));

      /* Copy message contents into new buf */
      if (adr)
	{
	  memcpy(adr, mp->buf, mp->len);
	  Free(mp->buf);
	  mp->buf = adr;
	}

      c_rcv += (CLOCK() - t0);
      return;
    }

  msg.buf = adr;

  mp = (mesg_t*) Malloc(sizeof(mesg_t));
  *mp = msg;
  ListInsertTail(waitl, mp);
  QueueAdd(recvq, rqitem);

  c_rcv += (CLOCK() - t0);
  dbg_msg(1, ("\n\tarcv-wait(%d(%d), %d(%d), %d(%d))", tag, mp->tag, from, mp->src, 
	      size, mp->len));
}








/* Routine to poll the incoming message queue for completion of a single message */
caddr_t RecvQueuePoll()
{
  mesg_t msg, *mp;
  recv_t *rcv;
  caddr_t adr;
  double t0;
  int tg, fm, sz;
  
  t0 = CLOCK();

  /* get the identity of the message we are looking for */
  if (QueueEmpty(recvq)) return NULL;
  rcv = (recv_t*) QueueDelete(recvq);
  tg = msg.tag = rcv->tag;
  fm = msg.src = rcv->source;
  
  /* do we have it on proc? if not wait for it, while draining the network */
  if (!(mp = (mesg_t*)ListFindDelete(arrvl, &msg, _MsgCompare)))
    {
      /* we must have the signature in wait list */
      if (!(mp = (mesg_t*) ListFindDelete(waitl, &msg, _MsgCompare)))
	{
	  MPPstatus();
	  _MsgPrint(&msg, "searched");
	  panic(("Corrupt waitl"));
	}

      adr = DrainNetworkWhileWait(mp->buf, &tg, &fm, &sz);
      message(("\n\t poll(%d-%d,%d-%d,%d-%d) waits", mp->tag, tg, mp->src, fm, mp->len, sz));
    }
  else
    {
      adr = mp->buf;
      message(("\npoll(%d,%d,%d) returns", mp->tag, mp->src, mp->len));
    }

  Free(mp);
  Free(rcv);
  c_poll += (CLOCK() - t0);
  return adr;
}






#if 0
/* Routine to probe the incoming message queue for completions */
caddr_t RecvQueueProbe(sz, fm, tg)
    int *sz;
    int *fm;
    int *tg;
{
  mesg_t msg, *mp;
  caddr_t adr;
  recv_t *rcv;
  int cnt = 0;
  
  /* check if there is a pending message */
  *tg = *fm = *sz = -1;
  if (QueueEmpty(recvq))  return NULL;

  rcv = (recv_t*) QueueHead(recvq);
  msg.tag = rcv->tag;
  msg.src = rcv->source;

  if (mp = (mesg_t*)ListFindDelete(arrvl, &msg, _MsgCompare))
    {
      cnt = mp->len;
    }
  else
    {
      /* check if the pending message has completed or not */
      if ((cnt = mpc_status(rcv->msgid))<0) return NULL;
      if (!(mp = (mesg_t*)ListFindDelete(waitl, &msg, _MsgCompare)))
	panic(("msg(%d,%d,%d) not in wait list", rcv->tag, rcv->source, rcv->len));
    }

  exit_if(mp->buf!=rcv->buffer, ("Structure error (%d, %d) != (%d, %d)",
				 mp->tag, mp->src, rcv->tag, rcv->source));

  /* delete a completed message from queue, returning the contents */
  rcv = (recv_t*) QueueDelete(recvq);
  adr = rcv->buffer;
  *fm = rcv->source;
  *tg = rcv->tag;

  Free(rcv);
  Free(mp);

  *sz = cnt;
  return adr;
}
#endif







/* Routine to recv a message from the network */
int MsgRecvFrom(tag, buf, len, src)
    int tag;
    char *buf;
    int len;
    int src;
{
  mesg_t msg, *mp;
  int i;
  double t0;
  
  dbg_msg(1, ("\n\tbrcv(%d,%d)", tag, len));

  t0 = CLOCK();

  /* if I have it already, fine return it */
  msg.tag = tag;
  msg.src = src;
  if (mp = (mesg_t*) ListFindDelete(arrvl, &msg, _MsgCompare))
    {
      dbg_if (len<mp->len, 
	      {
		MPPstatus();
		_MsgPrint(&msg, "searched");
		_MsgPrint(mp, "found");
		panic(("MsgRecv(%d,%d) - Short buffer (%d<%d) for msg(%d,%d)",
		       tag, len, len, mp->len, mp->tag, mp->src));
	      });
      
      memcpy(buf, mp->buf, mp->len);
      len = mp->len;

      Free(mp->buf);
      Free(mp);
    }
  else
    {
      /* otherwise, post an async msg and drain network while waiting */
      DrainNetworkWhileWait(buf, &tag, &src, &len);
    }

  c_rcv += (CLOCK() - t0);

  return len;
}








/* Routine to recv a message from the network */
int MsgRecv(tag, buf, len)
    int tag;
    char *buf;
    int len;
{
  return MsgRecvFrom(tag, buf, len, DONTCARE);
}








int _MsgCompare(ap, bp)
    mesg_t *ap, *bp;
{
  /* Limited Message Matching Capability */
  if (ap->tag==bp->tag && ap->src==bp->src)  return 0;
  if (ap->tag==bp->tag && ap->src==DONTCARE)  return 0;

#if 0
  if (ap->tag==DONTCARE && ap->src==bp->src)  return 0;
  if (ap->tag==DONTCARE && ap->src==DONTCARE) return 0;
#endif
  return 1;
}



void _MsgPrint(ap, str)
    mesg_t *ap;
    char *str;
{
  message(("\n\t%s t:%d s:%d l:%d", str, ap->tag, ap->src, ap->len));
}

#endif defined(DRAINER)





