//
// main.C
//
// $Header: /fs/nene/uysal/mambo/cholesky/part/RCS/main.C,v 1.1 1997/02/26 21:05:26 uysal Exp uysal $
//
// Sparse Matrix Partitioning for Cholesky Factorization
//
// Input  : 
//   1 - Symbolic factor (generated by 'symbolic')
//   2 - Supernode identification (generated by 'snode')
//   3 - Sparse Matrix File in Binary (generated by 'asc2bin')
//
// Output : 
//   Distributed sparse matrix ready for factorization
//
// Data Formats :
//
//    Output Matrix :    
//    1  int   Number of Supernodes (N)
//    N  int   Data size (in bytes) for each supernode in local file
//    N  long  Data offset (in bytes) for each supernode in local file
//   N+1 int   Column Index for each supernode
//    N  snod  Data for each supernode (see below)    
//  
//    Output Supernode Format :
//    1  int   Count of Updaters (U)
//    U  int   Index of supernode to update current
//    1  int   No. of rows stored locally (R)
//    R  int   Row Indices, one for each row
//    -  elem  Matrix Data (if any) stored locally in row major format  
//
// Author : Mustafa Uysal <uysal@cs.umd.edu>
//
//
//
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>

#include "sp.h"
#include "vector.h"
#include "buffer.h"
#include "datatype.h"
#include "mapping.h"


// Function Prototypes
void    Usage(char *prog);
int     StructOpen(char *sfile);
int    *SuperNodeMapping(int nsn, int *colIndex);
int     ColumnScan(int fid, int colno, int *stru);
char   *MatrixRead(int fid, int sta, int end, int &nbyte);
elem_t *ExtractRow(int row, char *vbuf, int rcol, int &nelem);

// Global Variables 
ibmsp  mpp;          // will run for SP-2
int    nval = 0;     // # of nonzeros in sparse matrix 
int    nval2 = 0;    // # of nonzeros in sparse matrix (test)
int    nval3 = 0;
int    nval4 = 0;

elem_t dzero = 0.0;

#if defined(TRACE)
extern "C" trace_init(char *name);
extern "C" trace_start(void);
extern "C" trace_end(void);
#endif

int main(int argc, char *argv[])
{
    // argument control 
    if (argc<4) Usage(argv[0]);

    // argument extraction
    int  argndx = 1;
    char *matfile = argv[argndx++];
    char *symfile = argv[argndx++];  
    char *sndfile = argv[argndx++];  
    char *outfile = argv[argndx++];
#if defined(TRACE)
    if (argndx < argc) trace_init(argv[argndx]);
    else Usage(argv[0]);
#endif

    // Output Logging
    mpp.init();
    mpp.printOnZero("\nParallel Sparse Matrix Partitioner");
    mpp.printOnZero("\tMatrix File          : ", matfile);
    mpp.printOnZero("\tSymbolic Factor      : ", symfile);
    mpp.printOnZero("\tSupernode File       : ", sndfile);
    mpp.printOnZero("\tOutput File          : ", outfile);

    // Trace collection starts here
#if defined(TRACE)  
    mpp.printOnZero("\tTrace File           : ", argv[argndx]);
    trace_start();
#endif


    // Read supernode identification
    int fid = open(sndfile, O_RDONLY);
    if (fid < 0)
      {
	cerr << "Error in opening supernode file : " << sndfile << ".\n";
	return -1;
      }
    int numsnode = 0;
    int bread = read(fid, &numsnode, sizeof(int));
    if (bread != sizeof(int))
      {
	cerr << "Error in reading supernode file : " << sndfile << ".\n";
	return -1;
      }

    int *snColIndex = new int[numsnode+1];
    bread = read(fid, snColIndex, sizeof(int)*(numsnode+1));
    if (bread != sizeof(int)*(numsnode+1))
      {
	cerr << "Error in reading supernode file : " << sndfile << ".\n";
	return -1;
      }
    close(fid);

    mpp.printOnZero("Supernode Descriptors read in.");

    // Initialize and write header of output file
    int ofid = open(outfile, O_RDWR | O_TRUNC | O_CREAT, 0644);
    if (ofid < 0)
      {
	cerr << "Error in opening output file : " << outfile << ".\n";
	return -1;
      }

    // temporarily dump header data into matrix
    int bwrite = write(ofid, &numsnode, sizeof(int));

    int *snodeSize = new int[numsnode];
    for (int i = 0; i < numsnode; i++) snodeSize[i] = 0;
    bwrite = write(ofid, snodeSize, sizeof(int)*numsnode);

    long *snodeOffs = new long[numsnode];
    for (i = 0; i < numsnode; i++) snodeOffs[i] = 0L;
    bwrite = write(ofid, snodeOffs, sizeof(long)*numsnode);

    // supernode indices
    bwrite = write(ofid, snColIndex, sizeof(int)*(numsnode+1));

    mpp.printOnZero("Output File Header written.\n");

    // Mapping of supernodes into clusters of segments
    int *colv = SuperNodeMapping(numsnode, snColIndex);

    // Following object manages of mapping supernodes onto processors
    TorusMap smap((mpapi&)mpp);
    if (mpp.myProc() == 0) smap.integrityCheck();

    // Open original matrix file (only procs who has it)
    int mfid = open(matfile, O_RDONLY), vsiz;

    // Initialize matrix state:
    char *vbuf = MatrixRead(mfid, 0, 0, vsiz);

    // Open symbolic factor file (only procs who has it)
    int sfid = StructOpen(symfile);

    // following is the temporary working space
    int  ncol  = snColIndex[numsnode]; 
    int *index = new int[ncol];
    assert( index != 0 );

    // produce a column to supernode mapping
    int k, kk;
    Vector col2snode(ncol, -1);
    for (k = 0; k < numsnode; k++)
      for (kk = snColIndex[k]; kk < snColIndex[k+1]; kk++)
	col2snode.setElem(kk, k);

    // create a local dependency list
    Buffer *depList = new Buffer[numsnode];
    mpp.printOnZero("Initialization complete, start processing...\n");

    // matrix partitoning happens here:
    for (k = 0; k < numsnode; k++)
      {
	int ndxcol   = snColIndex[k];
	int rightcol = snColIndex[k+1];
	int colmap   = colv[k];

	if (k>0) snodeOffs[k] = snodeOffs[k-1] + (long) snodeSize[k-1];
	else     snodeOffs[k] = ((numsnode+1)*2*sizeof(int)+(numsnode*sizeof(long)));

	if ((((k%100)==0) || (k==(numsnode-1))) && mpp.myProc()==0)
	  {
	    cout << ">supernode : " << k << " [" << ndxcol << " - " << rightcol << "]...\n";
	    cout.flush();
	  }

	// Read in original matrix data 
	vbuf = MatrixRead(mfid, ndxcol, rightcol, vsiz);

	// all processor on which sn(k) mapped:
	if (smap.isColMember(colmap))
	  {
	    Buffer ndxLocal(10);   // Local index struct of supernode

	    // retrieve column struct from the server
	    if (ColumnScan(sfid, ndxcol, index))
	      {
		Vector *vp = smap.procCol(colmap);
		mpp.multicast(vp, ndxcol, index[0]*sizeof(int), index);
		delete vp;
	      }
	    else
	      {
		int colsize = ncol * sizeof(int);
		int msgsrc  = DONTCARE;
		mpp.brecv(ndxcol, msgsrc, colsize, index);
	      }

	    // retrieve original matrix values
	    if (!vbuf)
	      {
		int msiz = sizeof(int);
		int msrc = DONTCARE;
		int mtag = ndxcol * 2;
		mpp.brecv(mtag, msrc, msiz, &vsiz);
		mtag++;
		vbuf = new char[vsiz];
		mpp.brecv(mtag, msrc, vsiz, vbuf);
	      }
	    else
	      {
		Vector *pv = smap.procCol(colmap);
		mpp.multicast(pv, 2*ndxcol, sizeof(int), &vsiz);
		mpp.multicast(pv, 2*ndxcol+1, vsiz, vbuf);
		delete pv;
	      }

	    // determine all rows to be stored locally:
	    int numElem  = 0;
	    int snWidth  = rightcol - ndxcol;
	    int psmap    = -1;
	    for (int j = 1; j < index[0]; j++)
	      {
		int snmap  = col2snode.getElem(index[j]);
		int rowmap = colv[snmap];
		
		nval2 += ((j < snWidth) ? j : snWidth);
		if (smap.isLocal(colmap,rowmap))
		  {
		    ndxLocal.append(index[j]);
		    numElem += ((j < snWidth) ? j : snWidth);
		  }
		if ((snmap != psmap)&&(snmap>k))
		  {
		    psmap = snmap;
		    depList[snmap].append(k);
		  }
	      }

	    // integrity checks:
	    int ccnt = index[0]-1;
	    int lcnt = 0;
	    for (int jj = snColIndex[k]; jj < snColIndex[k+1]; jj++)
	      {
		nval4 += ccnt;
		lcnt  += ccnt;
		ccnt--;
	      }

	    //cout << "\t--- " << lcnt << " elements --- " << snWidth << " cols --- ";
	    //cout << index[0] << " indices ---\n"; 

	    // arrange local nonzeros in row-major order:
	    elem_t *matVect = new elem_t[numElem];
	    for (numElem = j = 0; j < ndxLocal.numUsed(); j++)
	      {
		int nelem;
		elem_t *rowVect = ExtractRow(ndxLocal.getElem(j), vbuf, rightcol, nelem);
		for (int jj = 0; jj < nelem; jj++)
		  matVect[numElem++] = rowVect[jj];
		delete[] rowVect;
	      }

	    //char ms[20];
	    //sprintf(ms, "[-+]%3d", k);
	    //depList[k].print(15, ms);

	    // collect dependency structs for this snode
	    mpp.bufCollect(depList[k], k);

	    //sprintf(ms, "[++]%3d", k);
	    //depList[k].print(15, ms);
	    
	    // write all local data to my local file for this supernode:
	    int numItem = depList[k].numUsed();
	    bwrite = write(ofid, &numItem, sizeof(int));
	    snodeSize[k] += sizeof(int);
	    depList[k].append(ndxLocal.numUsed());
	    numItem++;
	    bwrite = write(ofid, depList[k].getData(), numItem * sizeof(int));
	    snodeSize[k] += (numItem * sizeof(int));
	    bwrite = write(ofid, ndxLocal.getData(), ndxLocal.numUsed() * sizeof(int));

	    if (k == 4)
	      {
		ndxLocal.print(10, "local data of sn 4");
		cout.flush();
	      }

	    snodeSize[k] += (ndxLocal.numUsed() * sizeof(int));
	    bwrite = write(ofid, matVect, numElem * sizeof(elem_t));
	    snodeSize[k] += (numElem * sizeof(elem_t));
	    nval  += numElem;

	    delete[] matVect;
	    delete[] vbuf;
	  }
	else 
	  {
	    Vector *pv = smap.procCol(colmap);

	    if (ColumnScan(sfid, ndxcol, index))
	      {
		mpp.multicast(pv, ndxcol, index[0]*sizeof(int), index);
	      }
	    if (vbuf)
	      {
		mpp.multicast(pv, 2*ndxcol, sizeof(int), &vsiz);
		mpp.multicast(pv, 2*ndxcol+1, vsiz, vbuf);
		delete[] vbuf;
	      }

	    delete pv;

	    //char ms[20];
	    //sprintf(ms, "[--]%3d", k);
	    //depList[k].print(15, ms);

	    mpp.bufCollect(depList[k], k);

	    //sprintf(ms, "[+-]%3d", k);
	    //depList[k].print(15, ms);


	    // write all local data to my local file for this supernode:
	    int numItem = depList[k].numUsed();
	    bwrite = write(ofid, &numItem, sizeof(int));
	    //cout << "\trecord<" << k << ">, <" << numItem << "> deps.\n";
	    snodeSize[k] += sizeof(int);
	    depList[k].append(0);
	    numItem++;
	    bwrite = write(ofid, depList[k].getData(), numItem * sizeof(int));
	    snodeSize[k] += (numItem * sizeof(int));
	  }

	depList[k].purge();
      }

    lseek(ofid, sizeof(int), SEEK_SET);
    bwrite = write(ofid, snodeSize, sizeof(int)*numsnode);
    bwrite = write(ofid, snodeOffs, sizeof(long)*numsnode);

    ColumnScan(sfid, snColIndex[numsnode], index);
    cerr << "\n\t\tPartition done, (" << nval << " nonzeros).";
    cerr << "\n\t\tIntegrity check(1): " << nval2 << " nonzeros.";
    cerr << "\n\t\tIntegrity check(2): " << nval3 << " nonzeros.";
    cerr << "\n\t\tIntegrity check(3): " << nval4 << " nonzeros.\n\n";

    fsync(ofid);
    close(ofid);
    close(sfid);
    close(mfid);

#if defined(TRACE)
    trace_end();
#endif

    delete[] index;
    delete[] snodeSize;
    delete[] snodeOffs;
    return 0;
}



// Early exit from the program with a display of program usage
void Usage(char *progname)
{
    cerr << "\n\nUsage : " << progname << " MatrixFile SymbolicFactor SupernodeFile  outputFile [traceFile] \n\n";
    exit(1);
}




int *SuperNodeMapping(int nsn, int *colIndex)
{
  int *mapv = new int[nsn];

  for (int k = 0; k < nsn; k++)
    mapv[k] = k;

  return mapv;
}


int StructOpen(char *sfile)
{
  int fid = open(sfile, O_RDONLY);

  if (fid < 0) return fid;

  int hbuf[3];
  if (read(fid, hbuf, 3 * sizeof(int)) != 3 * sizeof(int))
    {
      cerr << "\nError in reading file : " << sfile << ".\n";
      exit(1);
    }

  return fid;
}


int ColumnScan(int fid, int colno, int *stru)
{
  static int state[2] = {0, -1};  // colsize, colno

  if (fid<0||state[0]<0) return 0;

  //cout << "\tColumnScan(" << colno << "), start = (" << state[1] << "): ";
  
  while (state[1] < colno)
    {
      if (state[0]) lseek(fid, (state[0]-2)*sizeof(int), SEEK_CUR);
      int nbyt = read(fid, state, 2*sizeof(int));
      if ((nbyt == 0) || (state[0] == -1)) 
	{
	  state[0] = -1;
	  break;
	}
      else if (nbyt != 2*sizeof(int))
	{
	  cerr << "Error in reading column header (column = " << colno << ").";
	  cerr << "\tnbyte = " << nbyt << "\n";
	  break;
	}

      nval3 += (state[0]-1);
      //cout << "(" << state[1] << "," << state[0] << "), ";
      //cout.flush();
    }

  if (state[1] == colno)
    {
      stru[0] = state[0];
      stru[1] = state[1];
      if (read(fid, stru+2, (state[0]-2)*sizeof(int)) != (state[0]-2)*sizeof(int))
	{
	  cerr << "Error in reading column : " << colno << "\n";
	  return 0;
	}
      state[0] = 0;

      //cout << " success (" << stru[0] << " ints).\n";
      return 1;
    }

  //cout << " fails (current = " << state[1] << ")\n";
  return 0;
}



static int *header = NULL;

char *MatrixRead(int fid, int sta, int end, int &nbyte)
{
  if (fid<0) return NULL;

  if (!sta)
    {
      read(fid, &end, sizeof(int));
      cout << end << " columns in the matrix.\n";
      end *= 2;
      header = new int[end];
      read(fid, header, sizeof(int)*(end));
      return NULL;
    }

  nbyte = 0;
  for (int k = sta; k < end; k++)
    nbyte += (header[2*(k-1)+1]*(sizeof(int)+sizeof(elem_t)) + sizeof(int));

  //cout << "MatrixRead(" << sta << ", " << end << ") : " << nbyte << " bytes.\n";
  
  char *buf = new char[nbyte];
  if (read(fid, buf, nbyte) != nbyte)
    {
      cerr << "Error reading matrix file\n";
      return NULL;
    }

  return buf;
}


elem_t *ExtractRow(int row, char *vbuf, int rcol, int &nelem)
{
  int *index = (int*) vbuf;

  nelem = (row < rcol) ? (row - index[1] + 1) : (rcol - index[1]);
  assert( nelem > 0 );

  elem_t *buffer = new elem_t[nelem];
  int     offset = 0;

  for (int k = 0; k < nelem; k++)
    {
      char *ptr = vbuf + offset;
      index = (int*)ptr;

      for (int j = 1; j < index[0]; j++)
	if (index[j] == row)
	  {
	    char *p = ptr + index[0]*sizeof(int) + (j-1) * sizeof(elem_t);
	    buffer[k] = *((elem_t*)p);
	    break;
	  }
	else if (index[j]>row)
	  {
	    buffer[k] = dzero;
	    break;
	  }

      offset += (index[0]*sizeof(int) + (index[0]-1)*sizeof(elem_t));
    }

  return buffer;
}
