//
// main.C
//
// $Header: /fs/nene/uysal/mambo/cholesky/supernode/RCS/main.C,v 1.1 1997/03/02 06:14:26 uysal Exp uysal $
//
// Outofcore Supernode Identification Program
//
// Output (on each participating processor) :
//   A File containing supernode boundaries 
//
//      1    int    Number of supernodes (N)
//    (N+1)  int    Column numbers separating supernodes
//                  supernode k spans columns ndx(k) - ndx(k+1)-1 
//     
// 
// Author : Mustafa Uysal <uysal@cs.umd.edu>
//
//
//
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include "sp.h"

#include "column.h"
#include "vector.h"
#include "buffer.h"

#define MegaByte 0x100000
#define CHUNK    64

// Function Prototypes
void    usage(void);
Buffer *MatrixCluster( char *sfile );
Buffer *SnodeCollect(ibmsp &mpp, Buffer &input);
void    MapCollect(ibmsp &mpp, Vector &vct, int myid, int mymap);
int     ColumnRead(int fid, int *buffer);

// Global Variables 
ibmsp  mpp;          // will run for SP-2
char   *progname;    // name of the binary
int    numElem = 0;  // matrix size

int main(int argc, char *argv[])
{
    char      *symfile, *outfile;

    // get the executables' name
    progname = argv[0];

    // input control police
    if (argc<3) usage();

    // input file 
    int argndx = 1;
    symfile = argv[argndx++];  
    outfile = argv[argndx++];

    // Output Logging
    mpp.init();
    mpp.printOnZero("Supernode Identification Program.");
    mpp.printOnZero("\tSymbolic Factor      : ", symfile);
    mpp.printOnZero("\tOutput File          : ", outfile);;

    // Supernode Composition
    Buffer *snorig = MatrixCluster(symfile);

    // write output... 
    int ofid;

    if ((ofid = open(outfile, O_WRONLY | O_TRUNC | O_CREAT, 0644)) < 0)
      {
	cerr << "Error in opening output file : " << outfile << ".\n";
	return 0;
      }

    int  outsize = snorig->numUsed()+1;
    int *outbuf  = new int[outsize];

    outbuf[0]  = snorig->numUsed() - 1;
    for (int k = 0; k < snorig->numUsed(); k++)
      outbuf[k+1] = snorig->getElem(k);

    outsize *= sizeof(int);
    write(ofid, outbuf, outsize);

    close(ofid);
    delete outbuf;
    
    cout << "Finished, " << numElem << " nonzeros.\n";

    return 0;
}



// Early exit from the program with a display of program usage
void usage(void)
{
    cerr << "\n\nUsage : " << progname << " SymFactFile outputFile\n\n";
    exit(1);
}



//
// Function to read in the structure of a sparse matrix and cluster it 
// into supernodes. Returned sparse matrix has only supernode boundaries 
// valid.
//
Buffer *MatrixCluster(char *sfile)
{
  Vector procMap(mpp.numProcs(), -1);
  Buffer sncols;
  int    sfid;
  int    hbuf[3];

  if ((sfid = open(sfile, O_RDONLY)) < 0)
    {     // do not participate in the clustering
          // participate in the reverse map operation
      MapCollect(mpp, procMap, -1, -1);
    }
  else   // clustering happens here
    { 
      if (read(sfid, hbuf, 3 * sizeof(int)) != 3 * sizeof(int))
	{
	  cerr << "\nError in reading file : " << sfile << ".\n";
	  exit(1);
	}
      if (mpp.myProc() == 0) 
	cout << "Matrix Dimensions : [" << hbuf[0] << " x " << hbuf[0] << "].\n";

      // setup data for reading matrix struct 
      mpapi src_mpp(hbuf[2], hbuf[1]);
      int   *Abuf = new int[hbuf[0]];
      int   *Bbuf = new int[hbuf[0]];
      int   *Cbuf = new int[hbuf[0]];
      int   *Dbuf = new int[hbuf[0]];
      int   *cbuf = Abuf, *nbuf = Bbuf;
      int   *rbuf = Cbuf, *sbuf = Dbuf;
      int   rtag, rsrc;
      int    sndid, rcvid, sndon = 0, rcvon = 0;
      int    reading = 1;

      // find reverse mapping
      MapCollect(mpp, procMap, src_mpp.myProc(), mpp.myProc());

      if (src_mpp.myProc() == 0) sncols.append(1);

      // read and cluster matrix struct
      while (1)
	{
	  if (reading) 
	    if (!ColumnRead(sfid, cbuf)) break;

	  Column mycol(cbuf[1], cbuf[0], cbuf+1);
	  if (mycol.colid() == hbuf[0]) sncols.append(mycol.colid()+1);

	  //cout << "column = " << mycol.colid() << " size = " << cbuf[0] << ".\n";

	  if (reading)
	    {
	      int stride = mpp.colMapStride();
	      int siz = hbuf[0] * sizeof(int);
	      rtag = mycol.colid()-(mycol.colid()%stride)+stride;
	      if (rtag <= hbuf[0])
		{
		  rsrc  = procMap.getElem(src_mpp.colMap(rtag));
		  rcvid = mpp.arecv(rtag, rsrc, siz, rbuf);
		  rcvon = 1;
		}
	    }

	  // does mycol need to be sent away?
	  if (src_mpp.colMap(mycol.colid() - 1) != src_mpp.myProc())
	    {
	      if (sndon) mpp.mwait(sndid);
	      int nelem = cbuf[0]+1;
	      for (int k = 0; k < nelem; k++)
		sbuf[k] = cbuf[k];
	      int to   = procMap.getElem(src_mpp.colMap(mycol.colid() - 1));
	      sndid    = mpp.asend(to, mycol.colid(), nelem * sizeof(int), sbuf);
	      sndon    = 1;
	      //cout << "\tSending column = " << sbuf[1] << " to: " << to << "\n";
	      //cout.flush();
	    }

    	  // does nxcol need to be received?
	  if (src_mpp.colMap(mycol.colid()+1) != src_mpp.myProc())
	    {
	      if (rtag > hbuf[0]) break;

	      mpp.mwait(rcvid);
	      rcvon = 0;
	      for (int k = 0; k < rbuf[0]+1; k++)
		nbuf[k] = rbuf[k];
	      //cout << "\tReceived column : " << rbuf[1] << ".\n";
	      //cout.flush();
	    }
	  else // nxcol is one of the local
	    {
	      if (!ColumnRead(sfid, nbuf)) break;
	    }

	  Column nxcol(nbuf[1], nbuf[0], nbuf+1);

	  // structure comparison:
	  if (!mycol.sameStruct(nxcol))
	    sncols.append(mycol.colid()+1);

	  if (src_mpp.colMap(mycol.colid()+1) == src_mpp.myProc())
	    {
	      reading = 0;
	      int *tbuf = cbuf;
	      cbuf = nbuf;
	      nbuf = tbuf;
	    }
	  else reading = 1;
	}

      delete Abuf;
      delete Bbuf;
      delete Cbuf;
      delete Dbuf;
      close(sfid);
    }

  return SnodeCollect(mpp, sncols);
}




int ColumnRead(int fid, int *buffer)
{
  int rincol;

  if (read(fid, &rincol, sizeof(int)) != sizeof(int))
    return 0;

  // read local column struct 
  buffer[0] = rincol - 1;
  numElem += buffer[0];

  int nr, rsize = buffer[0] * sizeof(int);
  if ((nr = read(fid, buffer+1, rsize)) != rsize)
    {
      cerr << "\nError in reading column struct, size = " << rsize << " read = " << nr << ".\n";
      exit(1);
    }

  return nr;
}




void MapCollect(ibmsp &mpp, Vector &vct, int myid, int mymap)
{
  cout << "MapCollect(): myid = " << myid << " map = " << mymap << ".\n";
  cout.flush();

  for (int k = 0; k < mpp.numProcs(); k++)
    {
      int buf[2] = {myid, mymap};
      mpp.broadcast(k, k, 2 * sizeof(int), buf);
      if (buf[0] == -1) continue;
      vct.setElem(buf[0], buf[1]);
    }
}


Buffer *SnodeCollect(ibmsp &mpp, Buffer &input)
{
  Buffer *output;
  
  mpp.printOnZero("Supernode Collection Begins...");
  for (int k = 0; k < mpp.numProcs(); k++)
    {
      int nelem = input.numUsed();
      mpp.broadcast(k, k, sizeof(int), &nelem);

      int *data = new int[nelem];
      if (mpp.myProc() == k)
	for (int j = 0; j < nelem; j++)
	  data[j] = input.getElem(j);

      mpp.broadcast(k, mpp.numProcs()+k, nelem*sizeof(int), data);

      if (k == 0)
	{
	  output = new Buffer(nelem);
	  for (int j = 0; j < nelem; j++)
	    output->append(data[j]);
	}
      else
	{
	  output->merge(nelem, data);
	}
      delete data;
    }

  // output->sort(output->numUsed());
  if (mpp.myProc() == 0) output->print(10);
  return output;
}


