/*
 * io.c
 *
 * $Source: /fs/nene/uysal/pf/SymFact/RCS/io.c,v $
 *
 * $Header: /fs/nene/uysal/pf/SymFact/RCS/io.c,v 1.2 1995/09/09 18:51:39 uysal Exp $
 *
 * Disk I/O module for the parallel symbolic factorization. This module
 * is responsible for reading in the original input and managing the
 * asycnronous write operations for the partitioned columns.
 *
 * Author : Mustafa Uysal <uysal@cs.umd.edu>
 *
 */
#include <stdio.h>
#include <fcntl.h>
#include <malloc.h>

#include "world.h"
#include "mppworld.h"
#include "ipc.h"


/*--------------- Externally Defined Global Variables ---------------*/
extern int **Matrix;
extern int block_size;
extern int nz_size;



/*--------------- Function Prototypes ---------------*/
void TimedRead();
void TimedSeek();
void TimedWrite();
void ReceiveLoop ARGS((int *ncols, int *lcols));
int ColMap ARGS((int colno));
int ColNdx ARGS((int colno));


/*--------------- Implementation ---------------*/					
											
											
void ReadColIndices(file, ncols, lcols)
    char *file;										
    int  *ncols;									
    int  *lcols;  									
{
  int inFile;      /* File handle for the input file */

  int nc,lc;        /* number of columns in the matrix */				
  int *index;       /* Buffer for reading row indices */				
  int *header;      /* contents of the header */
  
  int i, to;
  int sz,nx;
  int colno;        /* Current column number */

  double t0,t1;     /* Variables for Benchmarking */
  double read_time, seek_time, bytes_read;
  int seek_cnt;
  
  seek_time = read_time = bytes_read = 0.0;
  seek_cnt  = 0;
  
  /* All processors other than master waits for the data */				
  if (MyProc())										
    {											
      ReceiveLoop(ncols, lcols);					
      return;										
    }											

  /* Open Files */
  inFile = open(file, O_RDONLY);
  EXIT_IF((inFile<0), (msg, "Error in opening input file [%s]", file));

  /* Read no. of columns */
  TimedRead(inFile, &nc, sizeof(int), &read_time, &bytes_read);

  /* Broadcast to all processors */
  *ncols  = nc;										
  MsgBroadcast(0, ncols, sizeof(int), MSG_MATRIX_SIZE);
  printf("\n %d x %d Matrix", nc, nc);

  /* Local data allocation */								
  lc = nc / NumProcs() + 5*block_size;
  Matrix = (int**) calloc(lc, sizeof(int*));						
  CHKNULL(Matrix);									
  for (*lcols=i=0; i<lc; i++)									
    Matrix[i] = (int*) NULL;								

  /* Allocate space for the matrix header */
  header = (int*) calloc(2*nc, sizeof(int));
  CHKNULL(header);

  /* Read and broadcast the header record */
  TimedRead(inFile, header, 2*nc*sizeof(int), &read_time, &bytes_read);
  MsgBroadcast(0, header, 2*nc*sizeof(int), MSG_MATRIX_HEADER);
  
  /* Read input column by column and write out */
  for (colno=1; colno<=nc; colno++)
    {
      sz = header[2*(colno-1)+1];
      index = (int*) calloc(sz+1, sizeof(int));
      CHKNULL(index);
      
      /* Read the column structure */
      TimedRead(inFile, index, (sz+1)*sizeof(int), &read_time, &bytes_read);
      TimedSeek(inFile, sz*nz_size, SEEK_CUR, &seek_time, &seek_cnt);
      assert(index[0] == sz+1);
      assert(index[1] == colno);
      /*printf("col:%d nnz:%d\n", index[1], sz);*/
      
      /* Send it out */
      to = ColMap(colno);
      if (to != MyProc())
	{
	  MsgSend(MSG_COL_NDX+colno, to, index, (header[2*(colno-1)+1]+1)*sizeof(int));
	  free(index);
	}
      else
	{
	  nx = ColNdx(colno);
	  Matrix[nx] = index;
	  if (nx>*lcols) *lcols=nx;
	}
    }
  
  close(inFile);

  printf("\n\tRead Time   : %.3lf secs", read_time);
  printf("\n\tSeek Time   : %.3lf secs [%d seeks]", seek_time, seek_cnt);
  printf("\n\tBytes Read  : %.3lf Mbytes [%.0lf bytes]", bytes_read/MEGABYTE, bytes_read);
  printf("\n\tRead Rate   : %.3lf MB/secs", bytes_read/(MEGABYTE*read_time));
  printf("\n\n");
  fflush(stdout);
}






/*
 * This function is called by all the slave processors during the
 * initial data read phase.
 */
void ReceiveLoop(ncol, lcol)
    int *ncol, *lcol;
{
  int *header;
  int *ptr;
  int nc, lc, sz, nx;
  int colno, i;
  
  /* Receive matrix size */
  MsgRecvFrom(MSG_MATRIX_SIZE, &nc, sizeof(int), 0);

  /* Receive matrix header */
  header = (int*) calloc(2*nc, sizeof(int));
  CHKNULL(header);
  MsgRecvFrom(MSG_MATRIX_HEADER, header, 2*nc*sizeof(int), 0);
  
  /* Local data allocation */
  lc = nc / NumProcs() + 5*block_size;
  Matrix = (int**) calloc(lc, sizeof(int*));
  CHKNULL(Matrix);
  for (i=0; i<lc; i++)
    Matrix[i] = (int*) NULL;

  for (*lcol=0, colno=1; colno<=nc; colno++)
    if (ColMap(colno) == MyProc())
      {
	sz = (header[2*(colno-1)+1]+1)*sizeof(int);
	ptr = (int*) malloc(sz);
	CHKNULL(ptr);

	MsgRecvFrom(MSG_COL_NDX+colno, ptr, sz, 0);
	assert(ptr[0] == (sz/sizeof(int)));
	assert(ptr[1] == colno);
	
	nx = ColNdx(ptr[1]);
	Matrix[nx] = ptr;
	if (*lcol<nx) *lcol=nx;
      }

  *ncol = nc;
}

											
											


/*
 * routine to yield the time taken by a read operation
 */
void TimedRead(fd, buf, size, timer, ctr)
    int fd;
    char *buf;
    int size;
    double *timer, *ctr;
{
  double t0, t1;
  int nb;
  
  t0 = CLOCK();
  nb = read(fd,buf,size);
  t1 = CLOCK();

  EXIT_IF(nb!=size, (msg, "Read Error - req:%d read:%d sofar:%.0lf", size, nb, *ctr));
  
  *timer += (t1 - t0);
  *ctr   += (double) size;
}



/*
 * routine to yield the time taken by a seek operation
 */
void TimedSeek(fd, offset, whence, timer, ctr)
    int fd;
    int offset;
    int whence;
    double *timer;
    int *ctr;
{
  double t0, t1;

  t0 = CLOCK();
  lseek(fd, offset, whence);
  t1 = CLOCK();

  *timer += (t1 - t0);
  (*ctr)++;
}




/*
 * routine to yield the time taken by a write operation
 */
void TimedWrite(fd, buf, size, timer, ctr)
    int fd;
    char *buf;
    int size;
    double *timer, *ctr;
{
  double t0, t1;

  t0 = CLOCK();
  assert(write(fd,buf,size)==size);
  t1 = CLOCK();

  *timer += (t1 - t0);
  *ctr   += (double) size;
}




