/*
 * sp2.c
 *
 * $Source: /fs/nene/uysal/mambo/cholesky/symbolic/RCS/sp2.c,v $
 *
 * $Header: /fs/nene/uysal/mambo/cholesky/symbolic/RCS/sp2.c,v 1.2 1995/09/08 09:40:42 uysal Exp $
 *
 * Implementation of configuration and message passing functions for
 * the IBM SP-2 platform.
 *
 * Author : Mustafa Uysal <uysal@cs.umd.edu>
 *
 */
#include <mpproto.h>

#include "world.h"
#include "mppworld.h"
#include "queue.h"


/*-------------- Type Definitions ----------------------------*/

/* sendq member item type */
typedef struct {
  int  msgid;    /* msg identifier for the async send */
  int  msgtag;    
} send_t;


/* recvq member item type */
typedef struct {
  caddr_t  buffer;
  int      msgid;
  int      source;
  int      tag;
} recv_t;


/*-------------- Global Variable Declarations ----------------*/

static queue_t *sendq; /* queue of outstanding asynchrounous send requests */
static queue_t *recvq; /* queue of outstanding asynchrounous recv requests */
static int me, np;



/*-------------- Local Function Prototypes ---------------------*/
void SendQueuePoll ARGS((void));
double w_seconds ARGS((void));

/* Routine to send an asynchronous message */
void CommAsyncSend ARGS((caddr_t adr, int to, int tag, int size));

/* Routine to recv an asynchronous message */
void CommAsyncRecv ARGS((caddr_t adr, int from, int tag, int size));

/* Routine to poll the network for completion of message passing activity */
void CommPollNetwork ARGS(( snode_t *src));




/*--------------- Implementation ---------------*/


int NumProcs( VOID )
{
  int numtask, taskid;

  mpc_environ(&numtask, &taskid);

  return(numtask);
}




int MyProc( VOID )
{
  int numtask, taskid;

  mpc_environ(&numtask, &taskid);

  return(taskid);  
}


void MPPinit()
{
  me = MyProc();
  np = NumProcs();

  /* initialize async send/recv message queues */
  sendq = QueueNew();
  recvq = QueueNew();
} 


/*
 * Cleanup function for the communications module
 */
void MPPexit()
{
  SendQueuePoll();
}



/*
 * Communication module status report
 */
void CommStatus()
{
  MESSAGE((msg, "\tComm: sendq=%d recvq=%d", sendq->load, recvq->load));
}





/* Routine to send an asynchronous message */
void CommAsyncSend(adr, to, tag, size)
    caddr_t adr;
    int     to,tag,size;
{
  send_t *sqitem;
  
  sqitem = (send_t*) malloc(sizeof(send_t));
  CHKNULL(sqitem);
  
  if (mpc_send(adr, size, to, tag, &(sqitem->msgid))!=0)
    PANIC((msg, "Send error : %d [sz:%d to:%d tg:%d]", mperrno, size, to, tag));
  sqitem->msgtag = tag;

  QueueAdd(sendq, sqitem);
}



/* Routine to recv an asynchronous message */
void CommAsyncRecv(adr, from, tag, size)
    caddr_t adr;
    int     from,tag,size;
{
  recv_t *rqitem;
  
  rqitem = (recv_t*) malloc(sizeof(recv_t));
  CHKNULL(rqitem);

  rqitem->buffer = adr ? adr : (caddr_t) malloc(size);
  CHKNULL(rqitem->buffer);

  rqitem->source = from;
  rqitem->tag    = tag;
  if (mpc_recv(rqitem->buffer, size, &(rqitem->source), &(rqitem->tag), &(rqitem->msgid))!=0)
    PANIC((msg, "Recv error : %d [sz:%d fm:%d tg:%d]", mperrno, size, from, tag));
  
  QueueAdd(recvq, rqitem);
}



/* Routine to poll the outgoing message queue for completions */
void SendQueuePoll()
{
  send_t *snd;

  while (!QueueEmpty(sendq))
    {
      snd = (send_t*) QueueHead(sendq);
      if (mpc_status(snd->msgid)<0)
	break;
      snd = (send_t*) QueueDelete(sendq);
      free(snd);
    }
}




/* Routine to drain the send queue */
void SendQueueDrain(col)
    int col;
{
  send_t *snd;
  int cnt = 0;

  while (!QueueEmpty(sendq))
    {
      snd = (send_t*) QueueDelete(sendq);
      while (mpc_status(snd->msgid)<0)
	{
	  if (++cnt>=TOOMANY)
	    PANIC((msg, "SendQueuePoll() blocked at col:%d for tag:%d", col, snd->msgtag));
	}
      free(snd);
    }
}



/* Routine to poll the incoming message queue for completions */
caddr_t RecvQueuePoll()
{
  caddr_t adr;
  recv_t *rcv;
  int cnt = 0;
  
  if (!QueueEmpty(recvq))
    {
      rcv = (recv_t*) QueueDelete(recvq);
      while (mpc_status(rcv->msgid)<0)
	{
	  if (++cnt>=TOOMANY)
	    PANIC((msg, "RecvQueuePoll() blocked for tag:%d", rcv->tag));
	}
      adr = rcv->buffer;
      free(rcv);
    }
  else
    adr = (caddr_t) NULL;

  return adr;
}




/*
 * Broadcast of a message from a designated processor to
 * all other peers
 */
void MsgBroadcast(origin, message, size, tag)
    int origin;
    void *message;
    int size, tag;
{
  int i;

  if (me != origin)
    {
      MsgRecv(tag, message, size);
      return;
    }
  
  for (i=0; i<np; i++)
    if (i != me)
      MsgSend(tag, i, message, size);
}




void MsgSend(tag, to, buf, len)
     int tag;
     int to;
     char *buf;
     int len;
{
  if (mpc_bsend(buf, len, to, tag) < 0)
    {
      PANIC((msg, "Error in msg send code=%d tag:%d to:%d len:%d", mperrno, tag, to, len));
    }
  /*
  else
    {
      MESSAGE((msg, "tg:%d to:%d sz:%d sent successfully", tag, to, len));
    }
  */
}





int MsgRecv(tag, buf, len)
     int tag;
     char *buf;
     int len;
{
  int src;
  int newlen;

  src = DONTCARE;
  mpc_brecv(buf, len+1000000, &src, &tag, &newlen);
  DBG_ACT(newlen>len, PANIC((msg, "ShortBuffer tag=%d len=%d nlen=%d src=%d",
			     tag, len, newlen, src)));

  return newlen;
}





int MsgRecvFrom(tag, buf, len, src)
     int tag;
     int src;
     char *buf;
     int len;
{
  int newlen, lsrc, ltag, cnt;

  lsrc = src;
  ltag = tag;

  cnt  = 0;
  while(mpc_probe(&lsrc, &ltag, &newlen)==0)
    {
      if (newlen>=0) break;
      if (++cnt>500000) 
	{
	  ltag = lsrc = DONTCARE;
	  mpc_probe(&lsrc, &ltag, &newlen);
	  MESSAGE((msg, "Result of WildCard probe t:%d s:%d l:%d", ltag, lsrc, newlen));
	  PANIC((msg, "RecvMsgFrom : Msg t:%d s:%d not arrived", tag, src));
	}
    }

  mpc_brecv(buf, len+1000000, &lsrc, &ltag, &newlen);
  DBG_ACT(newlen>len, PANIC((msg, "ShortBuffer tag=%d len=%d nlen=%d src=%d",
			     tag, len, newlen, src)));

  return newlen;
}





int RecvFromAnySrc(tag, buf, len)
     int tag;
     char *buf;
     int len;
{
  char *tbuf;
  int llen, lsrc, ltag, cnt;
  int newlen;
  
  lsrc = DONTCARE;
  ltag = tag;
  llen = len;

  cnt  = 0;
  while(mpc_probe(&lsrc, &ltag, &newlen)==0)
    {
      if (newlen>=0) break;
      if (++cnt>TOOMANY) 
	{
	  ltag = lsrc = DONTCARE;
	  while(mpc_probe(&lsrc, &ltag, &newlen)==0)
	    {
	      if (newlen<0) break;
	      MESSAGE((msg, "Flush Sequence t:%d s:%d l:%d", ltag, lsrc, newlen));
	      tbuf = (char*) malloc(newlen);
	      CHKNULL(tbuf);
	      mpc_brecv(tbuf, newlen, &lsrc, &ltag, &llen);
	      free(tbuf);
	      ltag = lsrc = DONTCARE;
	    }
	  PANIC((msg, "RecvMsgFrom : Msg t:%d not arrived", tag));
	}
    }

  mpc_brecv(buf, len+1000000, &lsrc, &ltag, &llen);
  DBG_ACT(llen>len, PANIC((msg, "ShortBuffer tag=%d len=%d nlen=%d src=%d",
			     tag, len, llen, lsrc)));

  return(lsrc);
}



double CLOCK()
{
  return w_seconds();
}




void Sync( void )
{
  mpc_sync(ALLGRP);
}








/*-------------- Function Implementations ---------------------*/








		  
